/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xmu/StdSel.h>
#include "Util.h"

#include "Compat.h"
#include "ArtTextP.h"

static XtResource resources[] = {
    {XtNshadowWidth, XtCShadowWidth, XtRDimension, sizeof(Dimension),
     XtOffsetOf(ArtTextRec, shadow.shadow_width), XtRImmediate, (XtPointer)1},
#define offset(field) XtOffsetOf(ArtTextRec, arttext.field)
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(foreground_pixel), XtRString, XtDefaultForeground},
    {XtNquoteColor, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(quote_pixel), XtRString, (XtPointer)XtDefaultForeground},
    {XtNheaderColor, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(header_pixel), XtRString, XtDefaultForeground},
    {XtNclickableColor, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(clickable_pixel), XtRString, XtDefaultForeground},
    {XtNhighlightBackground, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(highlight_bg), XtRString, XtDefaultForeground},
    {XtNhighlightForeground, XtCBackground, XtRPixel, sizeof(Pixel),
     offset(highlight_fg), XtRString, XtDefaultBackground},
    {XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(font), XtRString, XtDefaultFont},
    {XtNquoteFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(quote_font), XtRString, XtDefaultFont},
    {XtNheaderFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(header_font), XtRString, XtDefaultFont},
    {XtNclickableFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(clickable_font), XtRString, XtDefaultFont},
    {XtNfirst, XtCFirst, XtRLong, sizeof(long),
     offset(first), XtRImmediate, (XtPointer)0},
    {XtNshown, XtCShown, XtRLong, sizeof(long),
     offset(shown), XtRImmediate, (XtPointer)0},
    {XtNlines, XtCLines, XtRLong, sizeof(long),
     offset(lines), XtRImmediate, (XtPointer)0},
    {XtNnAlloc, XtCNAlloc, XtRLong, sizeof(long),
     offset(n_alloc), XtRImmediate, (XtPointer)0},
    {XtNresizeCallback, XtCCallback, XtRCallback, sizeof(XtCallbackList),
     offset(resize_callback), XtRImmediate, (XtPointer)NULL},
    {XtNimageMargin, XtCMargin, XtRDimension, sizeof(Dimension),
     offset(image_margin), XtRImmediate, (XtPointer)8},
    {XtNseparatorMargin, XtCMargin, XtRDimension, sizeof(Dimension),
     offset(separator_margin), XtRImmediate, (XtPointer)2},
    {XtNpreferredLines, XtCPreferredLines, XtRDimension, sizeof(Dimension),
     offset(preferred_lines), XtRImmediate, (XtPointer)32},
    {XtNpreferredColumns, XtCPreferredColumns, XtRDimension, sizeof(Dimension),
     offset(preferred_columns), XtRImmediate, (XtPointer)80},
#undef offset
};

static void Initialize(Widget, Widget, ArgList, Cardinal*);
static void Destroy(Widget);
static void Realize(Widget, XtValueMask*, XSetWindowAttributes*);
static void Resize(Widget);
static void Redisplay(Widget, XEvent*, Region);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static XtGeometryResult QueryGeometry(Widget, XtWidgetGeometry*,
				      XtWidgetGeometry*);

static void select_start(Widget, XEvent*, String*, Cardinal*);
static void select_extend(Widget, XEvent*, String*, Cardinal*);
static void select_extend_start(Widget, XEvent*, String*, Cardinal*);
static void select_end(Widget, XEvent*, String*, Cardinal*);
static void select_word(Widget, XEvent*, String*, Cardinal*);
static void click(Widget, XEvent*, String*, Cardinal*);

static XtActionsRec actions[] = {
    {"select-start",		select_start},
    {"select-extend",		select_extend},
    {"select-extend-start",	select_extend_start},
    {"select-end",		select_end},
    {"select-word",		select_word},
    {"click",			click},
};

static char translations[] =
"<Btn1Down>:	select-start() click() \n"
"<Btn1Motion>:	select-extend() \n"
"<Btn1Up>:	select-end(PRIMARY) \n"
"<Btn3Down>:	select-extend-start() \n"
"<Btn3Motion>:	select-extend() \n"
"<Btn3Up>:	select-end(PRIMARY) \n"
"<Btn2Down>:	select-word() \n"
"<Btn2Motion>:	select-word() \n";

ArtTextClassRec artTextClassRec = {
    { /* core fields */
        (WidgetClass) &shadowClassRec,	/* superclass			*/
        "ArtText",                      /* class_name                   */
        sizeof(ArtTextRec),	        /* widget_size                  */
        NULL,                           /* class_initialize             */
        NULL,                           /* class_part_initialize        */
        FALSE,                          /* class_inited                 */
        Initialize,                     /* initialize                   */
        NULL,                           /* initialize_hook              */
        Realize,        	        /* realize                      */
        actions,                        /* actions                      */
        XtNumber(actions),              /* num_actions                  */
        resources,                      /* resources                    */
        XtNumber(resources),            /* num_resources                */
        NULLQUARK,                      /* xrm_class                    */
        TRUE,                           /* compress_motion              */
#if (XtSpecificationRelease < 4)
        TRUE,                           /* compress_exposure            */
#elif (XtSpecificationRelease < 6)
	XtExposeCompressMaximal | XtExposeGraphicsExposeMerged,
					/* compress_exposure		*/
#else
	XtExposeCompressMaximal | XtExposeGraphicsExposeMerged |
	XtExposeNoRegion,		/* compress_exposure		*/
#endif
        TRUE,                           /* compress_enterleave          */
        FALSE,                          /* visible_interest             */
        Destroy,                        /* destroy                      */
        Resize,                         /* resize                       */
        Redisplay,                      /* expose                       */
        SetValues,                      /* set_values                   */
        NULL,                           /* set_values_hook              */
        XtInheritSetValuesAlmost,       /* set_values_almost            */
        NULL,                           /* get_values_hook              */
        NULL,                           /* accept_focus                 */
        XtVersion,                      /* version                      */
        NULL,                           /* callback_private             */
        translations,                   /* tm_table                     */
        QueryGeometry,                  /* query_geometry               */
        XtInheritDisplayAccelerator,    /* display_accelerator          */
        NULL                            /* extension                    */
    },
    { /* shadow fields */
	XtInheritPixelOffset,		/* pixel_offset			*/
	False,				/* use_arm_for_background	*/
	XtInheritAllocShadowColors,	/* alloc_shadow_colors		*/
	XtInheritAllocShadowPixmaps,	/* alloc_shadow_pixmaps		*/
	XtInheritAllocArmColor,		/* alloc_arm_color		*/
	XtInheritAllocArmPixmap,	/* alloc_arm_pixmap		*/
	XtInheritAllocGCs,		/* alloc_gcs			*/
	NULL,				/* extension			*/
    },
    { /* arttext fields */
        0                               /* extension                    */
    }
};

WidgetClass artTextWidgetClass = (WidgetClass)&artTextClassRec;

/*************************************************************************/

static void init_gcs(ArtTextWidget w)
{
    XGCValues	values;

    values.font = w->arttext.font->fid;

    values.foreground = w->arttext.foreground_pixel;
    values.background = w->core.background_pixel;
    w->arttext.ordinary_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.foreground = w->arttext.highlight_fg;
    values.background = w->arttext.highlight_bg;
    w->arttext.ordinary_h_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.font = w->arttext.quote_font->fid;

    values.foreground = w->arttext.quote_pixel;
    values.background = w->core.background_pixel;
    w->arttext.quote_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.foreground = w->arttext.highlight_fg;
    values.background = w->arttext.highlight_bg;
    w->arttext.quote_h_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.font = w->arttext.header_font->fid;

    values.foreground = w->arttext.header_pixel;
    values.background = w->core.background_pixel;
    w->arttext.header_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.foreground = w->arttext.highlight_fg;
    values.background = w->arttext.highlight_bg;
    w->arttext.header_h_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.font = w->arttext.clickable_font->fid;

    values.foreground = w->arttext.clickable_pixel;
    values.background = w->core.background_pixel;
    w->arttext.clickable_gc =
	XtGetGC((Widget)w, GCForeground | GCBackground | GCFont, &values);

    values.foreground = w->arttext.highlight_bg;
    w->arttext.highlight_gc = XtGetGC((Widget)w, GCForeground, &values);
}

static void free_gcs(ArtTextWidget w)
{
    XtReleaseGC((Widget)w, w->arttext.ordinary_gc);
    XtReleaseGC((Widget)w, w->arttext.ordinary_h_gc);
    XtReleaseGC((Widget)w, w->arttext.quote_gc);
    XtReleaseGC((Widget)w, w->arttext.quote_h_gc);
    XtReleaseGC((Widget)w, w->arttext.header_gc);
    XtReleaseGC((Widget)w, w->arttext.header_h_gc);
    XtReleaseGC((Widget)w, w->arttext.clickable_gc);
    XtReleaseGC((Widget)w, w->arttext.highlight_gc);
}

static Dimension preferred_width(ArtTextWidget w)
{
    return w->arttext.preferred_columns *
	w->arttext.font->max_bounds.width;
}

static Dimension preferred_height(ArtTextWidget w)
{
    return w->arttext.preferred_lines *
	(w->arttext.font->ascent + w->arttext.font->descent);
}

/*************************************************************************/

static void call_clickable(ArtTextWidget w,
			   struct line_table_item *item,
			   int click)
{
    if (item->type == LineTypeClickable &&
	item->u.cli.callback)
	item->u.cli.callback((Widget)w,
			     item->u.cli.client_data,
			     (XtPointer)click);
}

static char *process_and_dup_string(char *str)
{
    if (strchr(str, '\t')) { /* this won't happen very often */
	char	*res;
	int	i, j, n, n_alloced;
	
	n = strlen(str);
	n_alloced = n + 8;
	res = XtMalloc(n_alloced);

	for (i = 0, j = 0 ; i <= n ; i++) {
	    if (j + 10 > n_alloced) {
		n_alloced += 16;
		res = XtRealloc(res, n_alloced);
	    }

	    if (str[i] == '\t') {
		int	m = 8 - (j & 7);

		while (m-- > 0) res[j++] = ' ';
	    } else {
		res[j++] = str[i];
	    }
	}

	return res;
    } else {
	return XtNewString(str);
    }
}

static void alloc_line_table(ArtTextWidget w, long n)
{
    if (n > w->arttext.n_alloc - 2) {
	struct line_table_item	*loop;
	long			i = w->arttext.n_alloc;

	w->arttext.n_alloc = 2 * (i + 1);
	w->arttext.line_table =
	    (struct line_table_item *)XtRealloc((char *)w->arttext.line_table,
						w->arttext.n_alloc *
						sizeof(struct
						       line_table_item));
	loop = &w->arttext.line_table[i];
	if (i == 0)
	    loop->y = 0;
	for (i = w->arttext.n_alloc - i ; i > 0 ; loop++, i--) {
	    loop->u.str.string = NULL;
	    loop->u.str.no_offsets = 0;
	    loop->u.str.offsets = NULL;
	    loop->type = LineTypeOrdinary;
	}
    }
}


static void free_line_table(ArtTextWidget w)
{
    struct line_table_item	*loop = w->arttext.line_table;

    if (!loop) return;

    for (;;) {
	switch (loop->type) {
	case LineTypeOrdinary:
	case LineTypeHeader:
	case LineTypeQuoted:
	    if (!loop->u.str.string) {
		XtFree((char *)w->arttext.line_table);
		w->arttext.line_table = NULL;
		return;
	    }
	    XtFree(loop->u.str.string);
	    XtFree((char *)loop->u.str.offsets);
	    break;
	case LineTypeSeparator:
	case LineTypeSpace:
	    break;
	case LineTypeImage:
	    XFreePixmap(XtDisplay(w), loop->u.img.pixmap);
	    break;
	case LineTypeClickable:
	    call_clickable(w, loop, False);
	    XtFree(loop->u.cli.string);
	    break;
	}
	loop++;
    }
    /* not reached */
}

static void calc_shown(ArtTextWidget w)
{
    if (w->arttext.lines == 0)
	w->arttext.shown = 0;
    else {
	long	i, y;

	i = w->arttext.first;
	y = w->arttext.line_table[i].y + w->core.height;

	while (++i <= w->arttext.lines)
	    if (w->arttext.line_table[i].y >= y) {
		w->arttext.shown = i - w->arttext.first - 1;
		return;
	    }

	w->arttext.shown = i - w->arttext.first;
    }
}

static void update_line_table_item(ArtTextWidget w, long i)
{
    struct line_table_item	*item = &w->arttext.line_table[i];
    long			*offsets;
    long			no_alloc;
    char			*string;
    long			n, len;
    XFontStruct			*font;

    switch (item->type) {
    case LineTypeImage:
	(item + 1)->y = item->y + item->u.img.height;
	return;
    case LineTypeSeparator:
	(item + 1)->y = item->y + item->u.sep.size;
	return;
    case LineTypeClickable:
	font = w->arttext.clickable_font;
	(item + 1)->y = item->y + font->ascent + font->descent;
	return;
    case LineTypeSpace:
	(item + 1)->y = item->y + item->u.spc.size;
	return;
    case LineTypeOrdinary:
	font = w->arttext.font;
	break;
    case LineTypeQuoted:
	font = w->arttext.quote_font;
	break;
    case LineTypeHeader:
	font = w->arttext.header_font;
	break;
    default:
	return;
    }

    offsets = item->u.str.offsets;
    no_alloc = item->u.str.no_offsets;
    string  = item->u.str.string;

    if (no_alloc == 0) {
	no_alloc = 1;
	offsets = (long *)XtMalloc(sizeof(long));
    }

    if (string)
	len = strlen(string);
    else
	len = 0;

    if (len == 0) {
	offsets[0] = 0;
	n = 1;
    } else {
	long	offset;

	for (offset = 0, n = 0 ; len > 0 ; n++) {
	    int	temp;
	    
	    if (n > no_alloc - 2) {
		no_alloc *= 2;
		offsets =
		    (long *)XtRealloc((char *)offsets,
				      no_alloc * sizeof(long));
	    }
	    
	    temp = MyXWidthToChars(font, &string[offset],
				   len, w->core.width);
	    offset += temp;
	    len -= temp;
	    offsets[n] = offset;
	    if (temp == 0) {
		n++;
		break;
	    }
	}
    }

    item->u.str.offsets = offsets;
    item->u.str.no_offsets = n;
    (item + 1)->y = item->y + n * (font->ascent + font->descent);
}

static void update_line_table(ArtTextWidget w)
{
    long	n = w->arttext.lines;

    if (n > 0) {
	long	i;

	for (i = 0 ; i < n ; i++)
	    update_line_table_item(w, i);
    }

    calc_shown(w);
}

static void draw_item(ArtTextWidget w, long line, int y)
{
    struct line_table_item	*item = &w->arttext.line_table[line];
    Display			*disp = XtDisplay(w);
    Window			win = XtWindow(w);
    XFontStruct			*font;
    GC				gc, h_gc;
    int				x, dy, max_y, no;

    switch (item->type) {
    case LineTypeImage:
	if (item->u.img.pixmap != None) {
	    XCopyArea(disp, item->u.img.pixmap, win, w->arttext.ordinary_gc,
		      0, 0, item->u.img.width, item->u.img.height,
		      w->arttext.image_margin, y);
	    XClearArea(disp, win,
		       item->u.img.width + w->arttext.image_margin, y,
		       0, item->u.img.height, False);
	} else {
	    XClearArea(disp, win, 0, item->y, 0, item->u.img.height, False);
	}
	return;
    case LineTypeSeparator:
	{
	    int	width;

	    x = w->arttext.separator_margin;
	    width = w->core.width - 2 * x;
	    if (width > 0) {
		XClearArea(disp, win, 0, y,
			   width, item->u.sep.size, False);
		XClearArea(disp, win, x + width, y,
			   0, item->u.sep.size, False);
		ShadowDrawShadows((ShadowWidget)w, x, y,
				  width, item->u.sep.size, True);
	    } else {
		XClearArea(disp, win, 0, y, 0, item->u.sep.size, False);
	    }
	}
	return;
    case LineTypeClickable:
	{
	    int	width;

	    font = w->arttext.clickable_font;
	    no = strlen(item->u.cli.string);
	    width = XTextWidth(font, item->u.cli.string, no);
	    XDrawImageString(disp, win, w->arttext.clickable_gc,
			     0, y + font->ascent, item->u.cli.string, no);
	    XDrawLine(disp, win, w->arttext.clickable_gc,
		      0, y + font->ascent + 1,
		      width, y + font->ascent + 1);
	    XClearArea(disp, win, width, y,
		       0, font->ascent + font->descent, False);
	}
	return;
    case LineTypeSpace:
	XClearArea(disp, win, 0, y, 0, item->u.spc.size, False);
	return;
    case LineTypeOrdinary:
	font = w->arttext.font;
	gc = w->arttext.ordinary_gc;
	h_gc = w->arttext.ordinary_h_gc;
	break;
    case LineTypeQuoted:
	font = w->arttext.quote_font;
	gc = w->arttext.quote_gc;
	h_gc = w->arttext.quote_h_gc;
	break;
    case LineTypeHeader:
	font = w->arttext.header_font;
	gc = w->arttext.header_gc;
	h_gc = w->arttext.header_h_gc;
	break;
    default:
	return;
    }

    no = item->u.str.no_offsets;
    dy = font->ascent + font->descent;
    max_y = w->core.height - dy;

    if (no > 0) {
	if (w->arttext.sel_ok && w->arttext.sel_start_line <= line &&
	    line <= w->arttext.sel_stop_line) {
	    if (w->arttext.sel_start_line < line &&
		line < w->arttext.sel_stop_line) {
		long	i, offset = 0;

		for (i = 0 ; i < no ; i++) {
		    if (y >= max_y)
			goto clear_rest;
		    XDrawImageString(disp, win, h_gc, 0, y + font->ascent,
				     &item->u.str.string[offset],
				     item->u.str.offsets[i] - offset);
		    x = XTextWidth(font, &item->u.str.string[offset],
				   item->u.str.offsets[i] - offset);
		    XFillRectangle(disp, win, w->arttext.highlight_gc,
				   x, y, w->core.width - x, dy);
		    offset = item->u.str.offsets[i];
		    y += dy;
		}
	    } else {
		long	i, offset = 0;
		
		for (i = 0 ; i < no ; i++) {
		    char	*string = &item->u.str.string[offset];
		    long	len = item->u.str.offsets[i] - offset;
		    long	pos = offset;
		    long	n;
			
		    if (y >= max_y)
			goto clear_rest;
		    x = 0;
		    n = w->arttext.sel_start_offset - offset;
		    if (n > 0 && w->arttext.sel_start_line == line) {
			if (n > len) n = len;
			XDrawImageString(disp, win, gc,
					 x, y + font->ascent,
					 string, n);
			x += XTextWidth(font, string, n);
			string += n;
			pos += n;
			len -= n;
		    }
			
		    if (line == w->arttext.sel_stop_line) {
			n = w->arttext.sel_stop_offset - pos + 1;
			if (n == len) {
			    if (len > 0) {
				XDrawImageString(disp, win, h_gc,
						 x, y + font->ascent,
						 string, len);
				x += XTextWidth(font, string, len);
			    }
			    XFillRectangle(disp, win, w->arttext.highlight_gc,
					   x, y, w->core.width - x, dy);
			} else if (n > len) {
			    XDrawImageString(disp, win, h_gc,
					     x, y + font->ascent,
					     string, len);
			    x += XTextWidth(font, string, len);
			    XFillRectangle(disp, win, w->arttext.highlight_gc,
					   x, y, w->core.width - x, dy);
			} else {
			    if (n > 0) {
				XDrawImageString(disp, win, h_gc,
						 x, y + font->ascent,
						 string, n);
				x += XTextWidth(font, string, n);
				string += n;
				len -= n;
			    }

			    if (len > 0) {
				XDrawImageString(disp, win, gc,
						 x, y + font->ascent,
						 string, len);
				x += XTextWidth(font, string, len);
			    }

			    XClearArea(disp, win, x, y, 0, dy, False);
			}
		    } else {
			if (len > 0) {
			    XDrawImageString(disp, win, h_gc,
					     x, y + font->ascent,
					     string, len);
			    x += XTextWidth(font, string, len);
			}
			XFillRectangle(disp, win, w->arttext.highlight_gc,
				       x, y, w->core.width - x, dy);
		    }
		    offset = item->u.str.offsets[i];
		    y += dy;
		}
	    }
	} else {
	    long	i, offset = 0;
	    
	    for (i = 0 ; i < no ; i++) {
		int	len = item->u.str.offsets[i] - offset;
		char	*string = &item->u.str.string[offset];
		
		if (y >= max_y)
		    goto clear_rest;
		offset = item->u.str.offsets[i];
		if (len > 0) {
		    XDrawImageString(disp, win, gc,
				     0, y + font->ascent, string, len);
		    x = XTextWidth(font, string, len);
		    XClearArea(disp, win, x, y, 0, dy, False);
		} else XClearArea(disp, win, 0, y, 0, dy, False);
		y += dy;
	    }
	}
    } else {
    clear_rest:
	XClearArea(disp, win, 0, y, 0, 0, False);
    }

}

static void draw_items(ArtTextWidget w, long first, long last)
{
    long	i, y0;

    i = first;
    if (i < w->arttext.first)
	i = w->arttext.first;
    y0 = w->arttext.line_table[w->arttext.first].y;

    while (i <= last) {
	int	y = w->arttext.line_table[i].y - y0;

	if (y >= w->core.height)
	    break;
	draw_item(w, i, y);
	i++;
    }
}

/*************************************************************************/

static void coords_to_line_offset(ArtTextWidget w, int ex, int ey,
				  long *line, long *offset)
{
    struct line_table_item	*loop = w->arttext.line_table;
    XFontStruct			*font;
    long			y, i, n;

    *offset = 0;
    if (!loop || (loop->type == LineTypeOrdinary && !loop->u.str.string)) {
	*line = 0;
	return;
    } else if (ey < 0) {
	if (w->arttext.first > 0)
	    *line = w->arttext.first - 1;
	else
	    *line = 0;
	return;
    }

    i = w->arttext.first;
    n = i + w->arttext.shown;
    loop += i;
    y = ey + loop->y;

    while (loop[1].y <= y) {
	loop++;
	i++;
	if (loop->type == LineTypeOrdinary && !loop->u.str.string) {
	    *line = w->arttext.lines;
	    return;
	} else if (i >= n) {
	    *line = n;
	    return;
	}
    }

    *line = i;

    switch (loop->type) {
    case LineTypeOrdinary:	font = w->arttext.font;		break;
    case LineTypeQuoted:	font = w->arttext.quote_font;	break;
    case LineTypeHeader:	font = w->arttext.header_font;	break;
    case LineTypeSeparator:
    case LineTypeImage:
    case LineTypeClickable:
    case LineTypeSpace:
    default:			return;
    }

    y -= loop->y;
    i = y / (font->ascent + font->descent);

    if (i == 0)
	*offset = MyXWidthToChars(font, loop->u.str.string,
				  loop->u.str.offsets[0], ex);
    else
	*offset = loop->u.str.offsets[i-1] +
	    MyXWidthToChars(font,
			    &loop->u.str.string[loop->u.str.offsets[i-1]],
			    loop->u.str.offsets[i] - loop->u.str.offsets[i-1],
			    ex);
}

static void make_visible(ArtTextWidget w, int line, int offset)
{
    XtCallbackList	c_list = w->arttext.resize_callback;

    if (line < w->arttext.first)
	ArtTextSetFirst((Widget)w, line);
    else if (line >= w->arttext.first + w->arttext.shown)
	ArtTextSetFirst((Widget)w, line - w->arttext.shown + 1);
    else
	return; /* dont't send report */

    if (c_list) {
	ArtTextReport	report;

	report.canvas_length = w->arttext.lines;
	report.slider_length = w->arttext.shown;
	report.slider_position = w->arttext.first;
	XtCallCallbackList((Widget)w, c_list, (XtPointer)&report);
    }
}

/*************************************************************************/

static Boolean convert_selection(Widget gw, Atom *selection, Atom *target,
				 Atom *type_return, XtPointer *value_return,
				 unsigned long *length_return,
				 int *format_return)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    Display		*disp = XtDisplay(w);

    if (!w->arttext.sel_ok || *selection != w->arttext.curr_sel)
	return False;

    if (*target == XA_TARGETS(disp)) {
	Atom		*std_targets, *atom;
	unsigned long	std_length, n;

	if (!XmuConvertStandardSelection((Widget)w, w->arttext.sel_time,
					 selection, target, type_return,
					 (XPointer *)&std_targets, &std_length,
					 format_return))
	    return False;

	*value_return = (XtPointer)XtMalloc((std_length + 8) * sizeof(Atom));
	atom = *(Atom **)value_return;
	n = std_length;

	n++; *atom++ = XA_STRING;
	n++; *atom++ = XA_TEXT(disp);
	n++; *atom++ = XA_LENGTH(disp);
	n++; *atom++ = XA_LIST_LENGTH(disp);

	memcpy((char *)atom, (char *)std_targets, std_length * sizeof(Atom));
	XtFree((char *)std_targets);

	*length_return = n;
	*type_return = XA_ATOM;
	*format_return = 32;

	return True;
    } else if (*target == XA_LENGTH(disp)) {
	long	*length = (long *)XtMalloc(sizeof(long));
	char	*c;

	if (w->arttext.sel_start_line == w->arttext.sel_stop_line)
	    *length = w->arttext.sel_stop_offset -
		w->arttext.sel_start_offset + 1;
	else {
	    int		start = w->arttext.sel_start_line;
	    int		stop = w->arttext.sel_stop_line;
	    int		i;

	    switch (w->arttext.line_table[start].type) {
	    case LineTypeOrdinary:
	    case LineTypeHeader:
	    case LineTypeQuoted:
		*length = 1 +
		    strlen(&w->arttext.line_table[start].
			   u.str.string[w->arttext.sel_start_offset]);
		break;
	    case LineTypeSeparator:
	    case LineTypeImage:
	    case LineTypeClickable:
	    case LineTypeSpace:
		*length = 0;
		break;
	    }

	    for (i = start + 1 ; i < stop ; i++)
		switch (w->arttext.line_table[i].type) {
		case LineTypeOrdinary:
		case LineTypeHeader:
		case LineTypeQuoted:
		    *length +=
			strlen(w->arttext.line_table[i].u.str.string) + 1;
		    break;
		case LineTypeSeparator:
		case LineTypeImage:
		case LineTypeClickable:
		case LineTypeSpace:
		    break;
		}

	    *length += w->arttext.sel_stop_offset + 1;
	}

	switch (w->arttext.line_table[w->arttext.sel_stop_offset].type) {
	case LineTypeOrdinary:
	case LineTypeHeader:
	case LineTypeQuoted:
	    c = w->arttext.line_table[w->arttext.sel_stop_line].u.str.string;
	    if (c && strlen(c) == w->arttext.sel_stop_offset)
		(*length)++;
	    break;
	case LineTypeSeparator:
	case LineTypeImage:
	case LineTypeClickable:
	case LineTypeSpace:
	    break;
	}

	*value_return = (XtPointer)length;
	*type_return = XA_INTEGER;
	*length_return = 1;
	*format_return = 32;

	return True;
    } else if (*target == XA_LIST_LENGTH(disp)) {
	long	*length = (long *)XtMalloc(sizeof(long));

	*length = 1;
	*value_return = (XtPointer)length;
	*type_return = XA_INTEGER;
	*length_return = 1;
	*format_return = 32;

	return True;
    } else if (*target == XA_STRING || *target == XA_TEXT(disp)) {
	char	*buffer = NULL, *c;
	long	start = w->arttext.sel_start_line;
	long	stop = w->arttext.sel_stop_line;
	long	len = 0;

	if (start == stop) {
	    switch (w->arttext.line_table[start].type) {
	    case LineTypeOrdinary:
	    case LineTypeHeader:
	    case LineTypeQuoted:
		len = w->arttext.sel_stop_offset -
		    w->arttext.sel_start_offset + 1;
		buffer = XtCalloc(1, len + 2);
		strncpy(buffer, &
			w->arttext.line_table[start].
			u.str.string[w->arttext.sel_start_offset],
			len);
		buffer[len] = '\0';
		break;
	    case LineTypeSeparator:
	    case LineTypeImage:
	    case LineTypeClickable:
	    case LineTypeSpace:
		len = 0;
		buffer = XtMalloc(1);
		buffer[0] = '\0';
		break;
	    }
	} else {
	    long	i;

	    switch (w->arttext.line_table[start].type) {
	    case LineTypeOrdinary:
	    case LineTypeHeader:
	    case LineTypeQuoted:
		len = strlen(&w->arttext.line_table[start].
			     u.str.string[w->arttext.sel_start_offset]) + 1;
		break;
	    case LineTypeSeparator:
	    case LineTypeImage:
	    case LineTypeClickable:
	    case LineTypeSpace:
		len = 0;
		break;
	    }

	    for (i = start + 1 ; i < stop ; i++) {
		switch (w->arttext.line_table[i].type) {
		case LineTypeOrdinary:
		case LineTypeHeader:
		case LineTypeQuoted:
		    len += strlen(w->arttext.line_table[i].u.str.string) + 1;
		    break;
		case LineTypeSeparator:
		case LineTypeImage:
		case LineTypeClickable:
		case LineTypeSpace:
		    break;
		}
	    }
	    len += w->arttext.sel_stop_offset + 1;
	    
	    buffer = XtCalloc(1, len + 2);

	    switch (w->arttext.line_table[start].type) {
	    case LineTypeOrdinary:
	    case LineTypeHeader:
	    case LineTypeQuoted:
		strcpy(buffer,
		       &w->arttext.line_table[start].
		       u.str.string[w->arttext.sel_start_offset]);
		strcat(buffer, "\n");
		break;
	    case LineTypeSeparator:
	    case LineTypeImage:
	    case LineTypeClickable:
	    case LineTypeSpace:
		break;
	    }

	    for (i = start + 1 ; i < stop ; i++)
		switch (w->arttext.line_table[i].type) {
		case LineTypeOrdinary:
		case LineTypeHeader:
		case LineTypeQuoted:
		    strcat(buffer, w->arttext.line_table[i].u.str.string);
		    strcat(buffer, "\n");
		    break;
		case LineTypeSeparator:
		case LineTypeImage:
		case LineTypeClickable:
		case LineTypeSpace:
		    break;
		}

	    switch (w->arttext.line_table[stop].type) {
	    case LineTypeOrdinary:
	    case LineTypeHeader:
	    case LineTypeQuoted:
		if (w->arttext.line_table[stop].u.str.string)
		    strncat(buffer, w->arttext.line_table[stop].u.str.string,
			    w->arttext.sel_stop_offset + 1);
		break;
	    case LineTypeSeparator:
	    case LineTypeImage:
	    case LineTypeClickable:
	    case LineTypeSpace:
		break;
	    }
	}

	*value_return = (XtPointer)buffer;
	*length_return = buffer ? strlen(buffer) : 0;
	*type_return = XA_STRING;
	*format_return = 8;
	
	switch (w->arttext.line_table[stop].type) {
	case LineTypeOrdinary:
	case LineTypeHeader:
	case LineTypeQuoted:
	    c = w->arttext.line_table[stop].u.str.string;
	    if (c && strlen(c) == w->arttext.sel_stop_offset)
		buffer[(*length_return)++] = '\n';
	    break;
	case LineTypeSeparator:
	case LineTypeImage:
	case LineTypeClickable:
	case LineTypeSpace:
	    break;
	}

	return True;
    } else {
	return
	    XmuConvertStandardSelection((Widget)w, w->arttext.sel_time,
					selection, target, type_return,
					(XPointer *)value_return,
					length_return, format_return);
    }
}

static void lose_selection(Widget gw, Atom *selection)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (w->arttext.sel_ok) {
	w->arttext.sel_ok = False;
	Redisplay((Widget)w, NULL, 0);
    }
}

/*************************************************************************/

static void click(Widget gw, XEvent *event, String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		line, offset;
    int			x, y;
    struct line_table_item	*item;

    if (!get_event_xy(event, &x, &y)) {
	XBell(XtDisplay(w), 0);
	return;
    }

    coords_to_line_offset(w, x, y, &line, &offset);

    if (offset != 0 || line < 0 || line >= w->arttext.lines)
	return;

    item = &w->arttext.line_table[line];
    if (item->type == LineTypeClickable)
	call_clickable(w, item, True);
}

static void select_start(Widget gw, XEvent *event,
			 String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		line, offset;
    int			x, y;

    if (!get_event_xy(event, &x, &y)) {
	XBell(XtDisplay(w), 0);
	return;
    }

    if (w->arttext.sel_ok) {
	w->arttext.sel_ok = False;
	if (w->arttext.curr_sel) {
	    XtDisownSelection((Widget)w, w->arttext.curr_sel,
			      w->arttext.sel_time);
	    w->arttext.curr_sel = (Atom)0;
	}
    }

    coords_to_line_offset(w, x, y, &line, &offset);
    if (line < 0 || line >= w->arttext.lines)
	return;

    switch (w->arttext.line_table[line].type) {
    case LineTypeOrdinary:
    case LineTypeHeader:
    case LineTypeQuoted:
	break;
    case LineTypeImage:
    case LineTypeSeparator:
    case LineTypeClickable:
    case LineTypeSpace:
	return;
    }

    w->arttext.sel_start_line = line;
    w->arttext.sel_stop_line  = line;
    w->arttext.sel_start_offset = offset;
    w->arttext.sel_stop_offset  = offset;
    w->arttext.sel_ok = (w->arttext.lines > 0);
    w->arttext.extending = True;
    w->arttext.extend_end = True;
    Redisplay((Widget)w, NULL, 0);
}

static void select_extend_start(Widget gw, XEvent *event,
				String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		line, old, offset;
    int			x, y;
    Boolean		extend_start;

    if (!get_event_xy(event, &x, &y)) {
	XBell(XtDisplay(w), 0);
	return;
    }

    if (!w->arttext.sel_ok) {
	select_start(gw, event, params, no_params);
	return;
    }

    coords_to_line_offset(w, x, y, &line, &offset);
    if (line < 0 || line >= w->arttext.lines)
	return;

    if (line < w->arttext.sel_start_line)
	extend_start = True;
    else if (line > w->arttext.sel_stop_line)
	extend_start = False;
    else if (w->arttext.sel_start_line == w->arttext.sel_stop_line) {
	if (offset <= w->arttext.sel_start_offset)
	    extend_start = True;
	else if (offset >= w->arttext.sel_stop_line)
	    extend_start = False;
	else if (offset - w->arttext.sel_start_offset <=
		 w->arttext.sel_stop_offset - offset)
	    extend_start = True;
	else
	    extend_start = False;
    } else {
	if (line - w->arttext.sel_start_line <=
	    w->arttext.sel_stop_line - line)
	    extend_start = True;
	else
	    extend_start = False;
    }

    w->arttext.extend_end = !extend_start;
    w->arttext.extending = True;
    if (extend_start) {
	old = w->arttext.sel_start_line;
	w->arttext.sel_start_line = line;
	w->arttext.sel_start_offset = offset;
    } else {
	old = w->arttext.sel_stop_line;
	w->arttext.sel_stop_line = line;
	w->arttext.sel_stop_offset = offset;
    }

    if (old < line) {
	if (line >= w->arttext.lines)
	    line = w->arttext.lines - 1;
	draw_items(w, old, line);
    } else {
	if (old >= w->arttext.lines)
	    old = w->arttext.lines - 1;
	draw_items(w, line, old);
    }
}

static void select_extend(Widget gw, XEvent *event,
			  String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		line, old, offset;
    int			x, y;

    if (!w->arttext.sel_ok || !w->arttext.extending)
	return;

    if (!get_event_xy(event, &x, &y)) {
	XBell(XtDisplay(w), 0);
	return;
    }

    coords_to_line_offset(w, x, y, &line, &offset);
    if (line < 0 || line >= w->arttext.lines)
	return;

    if (w->arttext.extend_end) {
	old = w->arttext.sel_stop_line;
	w->arttext.sel_stop_line = line;
	w->arttext.sel_stop_offset = offset;
    } else {
	old = w->arttext.sel_start_line;
	w->arttext.sel_start_line = line;
	w->arttext.sel_start_offset = offset;
    }

    if (w->arttext.sel_start_line > w->arttext.sel_stop_line) {
	int		temp_line = w->arttext.sel_start_line;
	Cardinal	temp_offset = w->arttext.sel_start_offset;

	w->arttext.sel_start_line = w->arttext.sel_stop_line;
	w->arttext.sel_start_offset = w->arttext.sel_stop_offset;
	w->arttext.sel_stop_line = temp_line;
	w->arttext.sel_stop_offset = temp_offset;
	w->arttext.extend_end = !w->arttext.extend_end;
    } else if (w->arttext.sel_start_line == w->arttext.sel_stop_line &&
	       w->arttext.sel_start_offset > w->arttext.sel_stop_offset) {
	Cardinal	temp_offset = w->arttext.sel_start_offset;

	w->arttext.sel_start_offset = w->arttext.sel_stop_offset;
	w->arttext.sel_stop_offset = temp_offset;
	w->arttext.extend_end = !w->arttext.extend_end;
    }

    if (old < line) {
	if (line >= w->arttext.lines)
	    line = w->arttext.lines - 1;
	draw_items(w, old, line);
    } else {
	if (old >= w->arttext.lines)
	    old = w->arttext.lines - 1;
	draw_items(w, line, old);
    }

    make_visible(w, line, offset);
}

static void select_end(Widget gw, XEvent *event,
		       String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    w->arttext.extending = False;
    if (w->arttext.sel_ok) {
	Atom	atom = XA_PRIMARY;
	Time	time_stamp = get_event_time(event);

	if (*no_params > 0)
	    atom = XInternAtom(XtDisplay(w), params[0], False);

	if (XtOwnSelection((Widget)w, atom, time_stamp, convert_selection,
			   lose_selection, NULL)) {
	    w->arttext.curr_sel = atom;
	    w->arttext.sel_time = time_stamp;
	}
    }
}

static void select_word(Widget gw, XEvent *event,
			String *params, Cardinal *no_params)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		line, offset;
    int			x, y;

    if (!get_event_xy(event, &x, &y)) {
	XBell(XtDisplay(w), 0);
	return;
    }

    if (w->arttext.sel_ok) {
	w->arttext.sel_ok = False;
	if (w->arttext.curr_sel) {
	    XtDisownSelection((Widget)w, w->arttext.curr_sel,
			      w->arttext.sel_time);
	    w->arttext.curr_sel = (Atom)0;
	}
    }

    coords_to_line_offset(w, x, y, &line, &offset);

    if (0 <= line && line < w->arttext.lines &&
	(w->arttext.line_table[line].type == LineTypeOrdinary ||
	 w->arttext.line_table[line].type == LineTypeHeader ||
	 w->arttext.line_table[line].type == LineTypeQuoted)) {
	char	*string;

	w->arttext.sel_start_line = line;
	w->arttext.sel_stop_line  = line;

	string = w->arttext.line_table[line].u.str.string;
	if (string[offset] == '\0' || string[offset] == ' ')
	    w->arttext.sel_ok = False;
	else {
	    long	i, n;

	    n = strlen(string);

	    for (i = offset ; i >= 0 ; i--)
		if (string[i] == ' ')
		    break;
	    w->arttext.sel_start_offset = i + 1;

	    for (i = offset ; i < n ; i++)
		if (string[i] == ' ')
		    break;
	    w->arttext.sel_stop_offset = i - 1;

	    w->arttext.sel_ok = True;
	    w->arttext.extending = False;
	    w->arttext.extend_end = False;
	}
    } else {
	w->arttext.sel_ok = False;
    }

    Redisplay((Widget)w, NULL, 0);
    select_end((Widget)w, event, params, no_params); /* own the sel */
}

/*************************************************************************/

static void Initialize(Widget grequest, Widget gnew,
		       ArgList args, Cardinal *no_args)
{
    ArtTextWidget	new = (ArtTextWidget)gnew;

    init_gcs(new);
    new->arttext.sel_ok = False;
    new->arttext.extending = False;
    new->arttext.curr_sel = (Atom)0;
    new->arttext.line_table = NULL;
    new->arttext.n_alloc = 0;
    new->arttext.lines = 0;
    new->arttext.shown = 0;
    new->arttext.first = 0;
    if (new->core.width  == 0)
	new->core.width  = preferred_width(new);
    if (new->core.height == 0)
	new->core.height = preferred_height(new);
}

static void Destroy(Widget gw)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (w->arttext.sel_ok) {
	if (w->arttext.curr_sel) {
	    XtDisownSelection((Widget)w, w->arttext.curr_sel, CurrentTime);
	    w->arttext.curr_sel = (Atom)0;
	}
	w->arttext.sel_ok = False;
    }
    free_line_table(w);
    w->arttext.first = 0;
    w->arttext.shown = 0;
    w->arttext.lines = 0;
    w->arttext.line_table = NULL;
    w->arttext.n_alloc = 0;

    free_gcs(w);
}

static void Realize(Widget gw, XtValueMask *mask,
		    XSetWindowAttributes *attributes)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    update_line_table(w);

    shadowWidgetClass->core_class.realize((Widget)w, mask, attributes);
}

static void Resize(Widget gw)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (XtIsRealized((Widget)w)) {
	XtCallbackList	c_list = w->arttext.resize_callback;

	update_line_table(w);

	if (c_list) {
	    ArtTextReport	report;

	    report.slider_position = w->arttext.first;
	    report.slider_length = w->arttext.shown;
	    report.canvas_length = w->arttext.lines;
	    XtCallCallbackList((Widget)w, c_list, (XtPointer)&report);
	}
    }
}

static void Redisplay(Widget gw, XEvent *event, Region region)
{
    ArtTextWidget		w = (ArtTextWidget)gw;

    if (!XtIsRealized((Widget)w))
	return;

    if (w->arttext.shown > 0)
	draw_items(w, w->arttext.first, w->arttext.lines);
}

static Boolean SetValues(Widget gcurrent,
			 Widget grequest,
			 Widget gnew,
			 ArgList args,
			 Cardinal *num_args)
{
    ArtTextWidget	current = (ArtTextWidget)gcurrent;
    ArtTextWidget	new = (ArtTextWidget)gnew;
    Boolean		redisplay = False;

    new->arttext.shown = current->arttext.shown;
    new->arttext.lines = current->arttext.lines;

    if (new->arttext.foreground_pixel != current->arttext.foreground_pixel ||
	new->arttext.quote_pixel != current->arttext.quote_pixel ||
	new->arttext.header_pixel != current->arttext.header_pixel ||
	new->arttext.clickable_pixel != current->arttext.clickable_pixel ||
	new->arttext.highlight_fg != current->arttext.highlight_fg ||
	new->arttext.highlight_bg != current->arttext.highlight_bg ||
	new->arttext.font != current->arttext.font ||
	new->arttext.quote_font != current->arttext.quote_font ||
	new->arttext.header_font != current->arttext.header_font ||
	new->arttext.clickable_font != current->arttext.clickable_font) {
	free_gcs(current);
	init_gcs(new);
	redisplay = True;
    }

    if (new->arttext.n_alloc != current->arttext.n_alloc) {
	struct line_table_item	*loop;
	long			n, i = current->arttext.n_alloc;

	n = new->arttext.n_alloc;
	if (n <= new->arttext.lines)
	    n = new->arttext.lines + 1;
	new->arttext.line_table =
	    (struct line_table_item *)XtRealloc((char *)new->arttext.
						line_table,
						n * sizeof(struct
							   line_table_item));
	for (loop = &new->arttext.line_table[i] ; i < n ; i++, loop++) {
	    if (i == 0)
		loop->y = 0;
	    loop->u.str.string = NULL;
	    loop->u.str.no_offsets = 0;
	    loop->u.str.offsets = NULL;
	    loop->type = LineTypeOrdinary;
	}	    
    }

    if (new->arttext.first != current->arttext.first)
	redisplay = True;

    return redisplay;
}

static XtGeometryResult QueryGeometry(Widget gw,
				      XtWidgetGeometry *intended,
				      XtWidgetGeometry *preferred)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    Dimension		intended_width;
    Dimension		intended_height;

    preferred->request_mode = CWHeight | CWWidth;
    preferred->height = preferred_height(w);
    preferred->width = preferred_width(w);

    if (intended->request_mode | CWWidth)
	intended_width = intended->width;
    else
	intended_width = w->core.width;
    if (intended->request_mode | CWHeight)
	intended_height = intended->height;
    else
	intended_height = w->core.height;

    if (intended_width == preferred->width &&
        intended_height == preferred->height)
	return XtGeometryYes;
    else if (preferred->width == w->core.width &&
             preferred->height == w->core.height)
	return XtGeometryNo;
    else
	return XtGeometryAlmost;
}

/*************************************************************************/

void ArtTextRot13(Widget gw)
{
    ArtTextWidget		w = (ArtTextWidget)gw;
    struct line_table_item	*loop = w->arttext.line_table;

    if (!loop) return;

    for (;;) {
	char	*c;

	switch (loop->type) {
	case LineTypeOrdinary:
	case LineTypeHeader:
	case LineTypeQuoted:
	    c = loop->u.str.string;

	    if (!c) {
		update_line_table(w);
		if (XtIsRealized((Widget)w))
		    Redisplay((Widget)w, NULL, 0);
		return;
	    }

	    while (*c != '\0') {
		if ((*c >= 'a' && *c < 'a' + 13) ||
		    (*c >= 'A' && *c < 'A' + 13)) *c += 13;
		else if ((*c >= 'a' + 13 && *c <= 'z') ||
			 (*c >= 'A' + 13 && *c <= 'Z')) *c -= 13;
		c++;
	    }
	    break;
	case LineTypeSeparator:
	case LineTypeImage:
	case LineTypeClickable:
	case LineTypeSpace:
	    break;
	}
	loop++;
    }
    /* not reached */
}

void ArtTextClearLines(Widget gw)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (w->arttext.sel_ok) {
	if (w->arttext.curr_sel) {
	    XtDisownSelection((Widget)w, w->arttext.curr_sel, CurrentTime);
	    w->arttext.curr_sel = (Atom)0;
	}
	w->arttext.sel_ok = False;
    }
    free_line_table(w);
    w->arttext.first = 0;
    w->arttext.shown = 0;
    w->arttext.lines = 0;
    w->arttext.line_table = NULL;
    w->arttext.n_alloc = 0;
    XClearWindow(XtDisplay(w), XtWindow(w));
}

void ArtTextAddLine(Widget gw, char *string, LineType type)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    if (type != LineTypeQuoted && type != LineTypeHeader)
	type = LineTypeOrdinary;

    n = w->arttext.lines++;
    alloc_line_table(w, n);

    w->arttext.line_table[n].u.str.string = process_and_dup_string(string);
    w->arttext.line_table[n].type = type;
    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextAppendToLast(Widget gw, char *string)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n = w->arttext.lines - 1;

    if (n < 0) {
	ArtTextAddLine((Widget)w, string, LineTypeOrdinary);
	return;
    }

    switch (w->arttext.line_table[n].type) {
    case LineTypeSeparator:
    case LineTypeImage:
    case LineTypeClickable:
    case LineTypeSpace:
	ArtTextAddLine((Widget)w, string, LineTypeOrdinary);
	return;
    case LineTypeOrdinary:
    case LineTypeHeader:
    case LineTypeQuoted:
	break;
    }

    if (strchr(string, '\t')) {
	char	*base;
	long	src, dest, n_alloced;

	base = w->arttext.line_table[n].u.str.string;
	dest = n_alloced = strlen(base);
	src = 0;

	for (src = 0 ; string[src] != '\0' ; src++) {
	    if (dest + 10 > n_alloced) {
		n_alloced += 16;
		base = XtRealloc(base, n_alloced);
	    }

	    if (string[src] == '\t') {
		long	d = 8 - (dest % 8);

		while (d-- > 0)
		    base[dest++] = ' ';
	    } else {
		base[dest++] = string[src];
	    }
	}
	base[dest] = '\0';
	w->arttext.line_table[n].u.str.string = base;
    } else {
	long	len1, len2;

	len1 = strlen(w->arttext.line_table[n].u.str.string);
	len2 = strlen(string);
	w->arttext.line_table[n].u.str.string =
	    XtRealloc(w->arttext.line_table[n].u.str.string, len1 + len2 + 2);
	(void)strcat(w->arttext.line_table[n].u.str.string, string);
    }

    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextAddSelected(Widget gw, char *string, LineType type,
			long sel_start, long sel_stop)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    for (n = 0 ; string[n] != '\0' ; n++) {
	if (string[n] == '\t') {
	    long	i = 8 - (n % 8);

	    if (sel_start > n)
		sel_start += i;
	    if (sel_stop > n)
		sel_stop += i;
	}
    }

    n = w->arttext.lines++;
    alloc_line_table(w, n);

    w->arttext.line_table[n].u.str.string = process_and_dup_string(string);
    w->arttext.line_table[n].type = type;
    update_line_table_item(w, n);
    
    if (w->arttext.sel_ok) {
	if (w->arttext.curr_sel) {
	    XtDisownSelection((Widget)w, w->arttext.curr_sel, CurrentTime);
	    w->arttext.curr_sel = (Atom)0;
	}
	w->arttext.sel_ok = False;
	draw_items(w, w->arttext.sel_start_line, w->arttext.sel_stop_line);
    }

    if (sel_start < sel_stop) {
	w->arttext.sel_ok = True;
	w->arttext.sel_start_line = w->arttext.sel_stop_line =
	    w->arttext.lines - 1;
	w->arttext.sel_start_offset = sel_start;
	w->arttext.sel_stop_offset = sel_stop;
    }

    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    } else {
	XtCallbackList	c_list = w->arttext.resize_callback;

	ArtTextSetFirst((Widget)w, n);
	if (c_list) {
	    ArtTextReport	report;

	    report.canvas_length = w->arttext.lines;
	    report.slider_length = w->arttext.shown;
	    report.slider_position = w->arttext.first;
	    XtCallCallbackList((Widget)w, c_list, (XtPointer)&report);
	}
    }
}

void ArtTextAddSeparator(Widget gw, Dimension size)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    n = w->arttext.lines++;
    alloc_line_table(w, n);
    w->arttext.line_table[n].type = LineTypeSeparator;
    w->arttext.line_table[n].u.sep.size = size;
    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextAddImage(Widget gw, Pixmap pixmap,
		     Dimension width, Dimension height)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    n = w->arttext.lines++;
    alloc_line_table(w, n);
    w->arttext.line_table[n].type = LineTypeImage;
    w->arttext.line_table[n].u.img.pixmap = pixmap;
    w->arttext.line_table[n].u.img.width = width;
    w->arttext.line_table[n].u.img.height = height;
    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextAddClickable(Widget gw, String string,
			 XtCallbackProc callback, XtPointer client_data)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    n = w->arttext.lines++;
    alloc_line_table(w, n);
    w->arttext.line_table[n].type = LineTypeClickable;
    w->arttext.line_table[n].u.cli.string = process_and_dup_string(string);
    w->arttext.line_table[n].u.cli.callback = callback;
    w->arttext.line_table[n].u.cli.client_data = client_data;
    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextAddSpace(Widget gw, Dimension size)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		n;

    n = w->arttext.lines++;
    alloc_line_table(w, n);
    w->arttext.line_table[n].type = LineTypeSpace;
    w->arttext.line_table[n].u.spc.size = size;
    update_line_table_item(w, n);
    if (w->arttext.line_table[n + 1].y <=
	w->core.height + w->arttext.line_table[w->arttext.first].y) {
	calc_shown(w);
	draw_items(w, n, n);
    }
}

void ArtTextGetFirstShownSize(Widget gw, long *first,
			      long *shown, long *lines)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (first)
	*first = w->arttext.first;
    if (shown)
	*shown = w->arttext.shown;
    if (lines)
	*lines = w->arttext.lines;
}

void ArtTextSetFirst(Widget gw, long first)
{
    ArtTextWidget	w = (ArtTextWidget)gw;

    if (w->arttext.first != first) {
	if (first >= w->arttext.lines)
	    first = w->arttext.lines - 1;
	if (first < 0)
	    first = 0;
	w->arttext.first = first;
	calc_shown(w);
	Redisplay((Widget)w, NULL, 0);
    }
}

int ArtTextDumpToFile(Widget gw, FILE *file)
{
    ArtTextWidget	w = (ArtTextWidget)gw;
    long		i;
    char		*c = NULL;

    for (i = 0 ; i < w->arttext.lines ; i++) {
	switch (w->arttext.line_table[i].type) {
	case LineTypeOrdinary:
	case LineTypeHeader:
	case LineTypeQuoted:
	    c = w->arttext.line_table[i].u.str.string;
	    break;
	case LineTypeSeparator:
	case LineTypeImage:
	    continue;
	case LineTypeClickable:
	    c = w->arttext.line_table[i].u.cli.string;
	    break;
	case LineTypeSpace:
	    continue;
	}

	if (c && fprintf(file, "%s\n", c) == EOF)
	    return -1;
    }

    return 0;
}
