/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include "ahead.h"
#include "codes.h"
#include "hash.h"
#include "server.h"
#include "thread.h"
#include "viewer.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Message.h"
#include "../Widgets/ScrList.h"

typedef enum {
    AheadStateDisconnected,
    AheadStateConnWait,
    AheadStateSentModeReader,
    AheadStateSentGroup,
    AheadStateSentXover,
    AheadStateDoingXover,
    AheadStateSentHead,
    AheadStateDoingHead,
    AheadStateSentNext,
    AheadStateIdle
} AheadState;

static AheadState	ahead_state = AheadStateDisconnected;
static SERVER		ahead_server = {-1, NULL, 0, 0, NULL};
static THREAD_CONTEXT	*current = NULL;
static XtIntervalId	timeout_id = 0;
static XtInputId	input_id = 0;
static long		n_to_put = 0;
static long		n_pending = 0;
static long		hit_end;
static ARTICLE		*curr_art = NULL;

#define HEAD_NEXT	12
#define MAX_HEAD_NEXT	16
static char	head_next[] =
"HEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\n"
"HEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\n"
"HEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\n"
"HEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\nHEAD\r\nNEXT\r\n";

typedef struct schedule_node {
    GROUP_NODE			*group;
    struct schedule_node	*next;
} SCHEDULE_NODE;

static SCHEDULE_NODE	*schedule = NULL;

static void timeout_proc(XtPointer client_data, XtIntervalId *id)
{
    timeout_id = 0;

    if (ahead_state == AheadStateIdle) {
	set_message("Closing second connection to server.", False);
	ahead_state = AheadStateDisconnected;
	server_write_raw(&ahead_server, "\r\nQUIT\r\n");
	server_close(&ahead_server);
	if (input_id != 0)
	    XtRemoveInput(input_id);
	input_id = 0;
    }
}

static void set_ahead_timeout()
{
    if (timeout_id != 0)
	XtRemoveTimeOut(timeout_id);
    timeout_id =
	XtAppAddTimeOut(global.app_cont, global.ahead_timeout,
			timeout_proc, NULL);
}

static void schedule_group(GROUP_NODE *group)
{
    SCHEDULE_NODE	*temp, *last = schedule;

    temp = (SCHEDULE_NODE *)XtMalloc(sizeof(SCHEDULE_NODE));
    temp->group = group;
    temp->next = NULL;

    if (last) {
	while (last->next)
	    last = last->next;
	last->next = temp;
    } else
	schedule = temp;

    group->ahead = (THREAD_CONTEXT *)XtMalloc(sizeof(THREAD_CONTEXT));
    nullify_thread_context(group->ahead);
    group->ahead->flag = AheadFlagScheduled;
    group->ahead->group = group;
}

static void schedule_group_top(GROUP_NODE *group)
{
    SCHEDULE_NODE	*temp;

    temp = (SCHEDULE_NODE *)XtMalloc(sizeof(SCHEDULE_NODE));
    temp->group = group;
    temp->next = schedule;
    schedule = temp;

    group->ahead = (THREAD_CONTEXT *)XtMalloc(sizeof(THREAD_CONTEXT));
    nullify_thread_context(group->ahead);
    group->ahead->flag = AheadFlagScheduled;
    group->ahead->group = group;
}

static void unschedule_group(GROUP_NODE *group)
{
    SCHEDULE_NODE	*prev, *loop;

    for (prev = NULL, loop = schedule ;
	 loop ; prev = loop, loop = loop->next)
	if (loop->group == group) {
	    if (loop->group->ahead)
		clear_thread_context(loop->group->ahead);
	    loop->group->ahead = NULL;
	    loop->group = NULL;
	    if (prev)
		prev->next = loop->next;
	    else
		schedule = loop->next;
	    XtFree((char *)loop);
	    break;
	}
}

static void free_current_context()
{
    if (current) {
	GROUP_NODE	*group = current->group;

	clear_thread_context(current);
	XtFree((char *)current);
	current = NULL;
	if (group) {
	    group->ahead = NULL;
	    update_group_entry(group);
	}
    }
}

static GROUP_NODE *get_group_to_schedule()
{
    while (schedule) {
	GROUP_NODE	*group = schedule->group;
	SCHEDULE_NODE	*next = schedule->next;

	schedule->group = NULL;
	schedule->next = NULL;
	XtFree((char *)schedule);
	schedule = next;

	if ((global.mode == NewsModeGroup ||
	     global.mode == NewsModeThread) &&
	    global.thr.group == group) {
	    if (group->ahead) {
		clear_thread_context(group->ahead);
		XtFree((char *)group->ahead);
		group->ahead = NULL;
	    }
	    continue;
	}

	if (!group->ahead)
	    continue;

	return group;
    }

    return NULL;
}


static void ahead_read_callback(XtPointer, int*, XtInputId*);

static void thread_ahead_schedule()
{
    if (ahead_state == AheadStateDisconnected) {
	int	temp;

	set_message("Opening second connection to server...", False);
	if (global.nntp_server[0] != '#')
	    temp = server_open(&ahead_server, NULL, 0);
	else
	    temp = server_fork(&ahead_server, NULL, 0);

	if (temp < 0) {
	    set_message("Failed to open second connection "
			"to server!", True);
	    server_close(&ahead_server);
	    return;
	}

	input_id = XtAppAddInput(global.app_cont,
				 ahead_server.fd,
				 (XtPointer)XtInputReadMask,
				 ahead_read_callback,
				 NULL);
	ahead_state = AheadStateConnWait;
    } else if (ahead_state == AheadStateIdle) {
	char		command[512];
	GROUP_NODE	*group = get_group_to_schedule();

	if (!group) {
	    /* set timer to close ahead_server */
	    current = NULL;
	    set_ahead_timeout();
	    return;
	}

	if (timeout_id)
	    XtRemoveTimeOut(timeout_id);
	timeout_id = 0;

	n_to_put = 0;
	n_pending = 0;
	current = group->ahead;
	current->flag = AheadFlagThreading;

	sprintf(command, "GROUP %s\r\n", group->name);

	if (server_write(&ahead_server, command) == 0) {
	    ahead_state = AheadStateSentGroup;
	    sprintf(command, "Threading %s...", group->name);
	    set_message(command, False);
	    update_group_entry(group);
	} else {
	    free_current_context();
	    schedule_group_top(group); /* put it back */
	    server_close(&ahead_server);
	    ahead_state = AheadStateDisconnected;
	    thread_ahead_schedule(); /* reconnect */
	}
    }
}

static int put_head_next()
{
    long	n = n_to_put;

    if (hit_end || n_pending >= MAX_HEAD_NEXT)
	return True;

    if (n == 0)
	n = 1;
    else {
	if (n > MAX_HEAD_NEXT)
	    n = MAX_HEAD_NEXT;
	n_to_put -= n;
    }
    n_pending += n;

    if (server_write(&ahead_server,
		     head_next + HEAD_NEXT * (MAX_HEAD_NEXT - n)) == 0)
	return True;

    /* error */
    free_current_context();
    server_close(&ahead_server);
    ahead_state = AheadStateDisconnected;
    thread_ahead_schedule(); /* reconnect */
    return False;
}

static void thread_ahead_start_head()
{
    n_pending = 0;
    n_to_put = current->total;
    hit_end = False;
    curr_art = NULL;

    (void)put_head_next();
}

static void process_thread_ahead()
{
    char	command[512];
    char	*buffer;
    long	status;

    for (;;) {
	if (ahead_state == AheadStateDoingHead)
	    buffer = server_get_chunk(&ahead_server);
	else
	    buffer = server_get_line(&ahead_server);

	if (!buffer)
	    return;

	switch (ahead_state) {
	case AheadStateDisconnected:
	    /* not reached */
	    return;
	case AheadStateConnWait:
	    status = atoi(buffer);
	    if (status == OK_CANPOST || status == OK_NOPOST) {
		if (server_write(&ahead_server, "MODE READER\r\n") == 0) {
		    ahead_state = AheadStateSentModeReader;
		    return; /* OK */
		}
		set_message("Thread ahead failed!", True);
	    } else {
		if (strlen(buffer) > 80) buffer[80] = '\0';
		sprintf(command,
			"Thread ahead failed, message from server is: %s",
			buffer);
		set_message(command, True);
	    }

	    /* error */
	    if (input_id)
		XtRemoveInput(input_id);
	    input_id = 0;
	    server_close(&ahead_server);
	    ahead_state = AheadStateDisconnected;
	    return;
	case AheadStateSentModeReader:
	    /* should check the reply */
	    ahead_state = AheadStateIdle;
	    thread_ahead_schedule();
	    return;
	case AheadStateSentGroup:
	    if (atoi(buffer) == OK_GROUP) {
		long	first, last, total;

		(void)sscanf(buffer, "%*d%ld%ld%ld", &total, &first, &last);
		if (total <= 0)
		    total = 1;
		current->total = total;
		current->n_done = 0;
		if (global.xover_supported) {
		    sprintf(command, "XOVER %ld-%ld\r\n", first, last);
		    if (server_write(&ahead_server, command) == 0)
			ahead_state = AheadStateSentXover;
		    else { /* broken connection */
			GROUP_NODE	*group = current->group;

			server_close(&ahead_server);
			ahead_state = AheadStateDisconnected;
			free_current_context();
			schedule_group_top(group);
			thread_ahead_schedule(); /* reconnect */
		    }
		} else {
		    ahead_state = AheadStateSentHead;
		    thread_ahead_start_head();
		}
	    } else {
		if (strlen(buffer) > 80)
		    buffer[80] = '\0';
		sprintf(command, "Failed to thread ahead %s:  %s!",
			current->group->name, buffer);
		set_message(command, True);
		free_current_context();
		ahead_state = AheadStateIdle;
		thread_ahead_schedule(); /* do next */
	    }
	    return;
	case AheadStateSentXover:
	    if (atoi(buffer) == OK_XOVER) {
		ahead_state = AheadStateDoingXover;
		break;
	    } else {
		/* fall back to head */
		thread_ahead_start_head();
		return;
	    }
	case AheadStateDoingXover:
	    if (buffer[0] == '.' && buffer[1] == '\0') {
		ahead_state = AheadStateIdle;
		current->flag = AheadFlagDone;
		update_group_entry(current->group);
		current = NULL;
		thread_ahead_schedule();
		return;
	    } else {
		ARTICLE	*art;

		current->n_done++;
		current->refs = NULL;
		art = parse_xover_line(current, buffer);
		if (art) {
		    if (current->last_art)
			current->last_art->next = art;
		    else
			current->articles = art;
		    current->last_art = art;

		    thread_article(current, art);
		}
		current->refs = NULL;

		break;
	    }
	case AheadStateSentHead:
	    curr_art = NULL;
	    if (atoi(buffer) == OK_HEAD)
		ahead_state = AheadStateDoingHead;
	    else
		ahead_state = AheadStateSentNext;
	    if (!hit_end) {
		current->n_done++;
		if (n_pending < MAX_HEAD_NEXT)
		    if (!put_head_next())
			return;
		curr_art = parse_head_reply(current, buffer);
	    }
	    break;
	case AheadStateDoingHead:
	    current->refs = NULL;
	    if (curr_art && parse_head(current, curr_art, buffer)) {
		if (current->last_art)
		    current->last_art->next = curr_art;
		else
		    current->articles = curr_art;
		current->last_art = curr_art;

		thread_article(current, curr_art);
	    }
	    curr_art = NULL;
	    current->refs = NULL;
	    ahead_state = AheadStateSentNext;
	    break;
	case AheadStateSentNext:
	    n_pending--;

	    if (!hit_end) {
		status = atoi(buffer);
		if (status != OK_NOTEXT) /* maybe better checking */
		    hit_end = True;
	    }

	    if (hit_end && n_pending <= 0) {
		ahead_state = AheadStateIdle;
		current->flag = AheadFlagDone;
		update_group_entry(current->group);
		current = NULL;
		thread_ahead_schedule();
		return;
	    }

	    ahead_state = AheadStateSentHead;

	    break;
	case AheadStateIdle:
	    /* probably a timeout message from the server */
	    free_current_context(); /* will probably no-op */
	    server_close(&ahead_server);
	    ahead_state = AheadStateDisconnected;	    
	    return;
	}
    }
}

static void ahead_read_callback(XtPointer client_data,
				int *fd,
				XtInputId *id)
{
    long	len;
    long	i, n;

    n = ahead_server.curr_pos +
	strlen(ahead_server.buffer + ahead_server.curr_pos);
    len = ahead_server.len - n;

    if (len <= 0) {
	len += ahead_server.len;
	ahead_server.len *= 2;
	ahead_server.buffer =
	    XtRealloc(ahead_server.buffer, ahead_server.len + 3);
    }

    i = read(ahead_server.fd, ahead_server.buffer + n, len);

    if (i <= 0) {
	char	*msg = "Thread ahead: Connection broken!";

	if (i < 0) {
	    perror("knews: read");
	    switch (errno) {
	    case EAGAIN:
#if EAGAIN != EWOULDBLOCK
	    case EWOULDBLOCK:
#endif
	    case EINTR:
		return; /* this should never happen */
	    case ECONNREFUSED:
		msg = "Thread ahead: Connection refused!";
		break;
	    case ENETUNREACH:
		msg = "Thread ahead: Network unreachable!";
		break;
	    case ETIMEDOUT:
		msg = "Thread ahead: Connection timed out!";
		break;
	    case EADDRNOTAVAIL:
		msg = "Thread ahead: Address not available!";
		break;
	    }
	}

	set_message(msg, True);
	(void)server_close(&ahead_server);
	ahead_state = AheadStateDisconnected;
	if (input_id != 0)
	    XtRemoveInput(input_id);
	input_id = 0;
	free_current_context();

	while (schedule) {
	    SCHEDULE_NODE	*next = schedule->next;

	    if (schedule->group && schedule->group->ahead) {
		clear_thread_context(schedule->group->ahead);
		schedule->group->ahead = NULL;
		update_group_entry(schedule->group);
	    }
	    schedule->group = NULL;
	    schedule->next = NULL;
	    XtFree((char *)schedule);
	    schedule = next;
	}

	return;
    }

    ahead_server.buffer[n + i] = '\0';

    process_thread_ahead();
}

void action_schedule_thread_ahead(Widget w, XEvent *event,
				  String *params, Cardinal *no_params)
{
    GROUP_NODE	*group = NULL;

    switch (global.mode) {
    case NewsModeConnected:
	if (w == main_widgets.group_list) {
	    long	n = ScrListEventToIndex(w, event);

	    if (n >= 0) {
		long	i;

		for (i = 0 ; i < global.no_groups ; i++)
		    if (global.groups[i]->disp == n) {
			group = global.groups[i];
			break;
		    }
	    }
	} else {
	    group = global.thr.group;
	}

	break;
    case NewsModeAllgroups:
	if (w == main_widgets.group_list) {
	    long	n = ScrListEventToIndex(w, event);

	    if (n >= 0 && n < global.no_groups)
		group = global.groups[n];
	} else {
	    group = global.thr.group;
	}

	break;
    case NewsModeDisconnected:
    case NewsModeGroup:
    case NewsModeThread:
    case NewsModeNewgroups:
	if (global.bell)
	    XBell(XtDisplay(w), 0);
	return;
    }

    if (!group) {
	if (global.bell)
	    XBell(XtDisplay(w), 0);
	return;
    }

    if (group->ahead) {
	char	message[128];

	switch (group->ahead->flag) {
	case AheadFlagDone:
	    set_message("That group has already been threaded!", True);
	    break;
	case AheadFlagScheduled:
	    set_message("That group is already scheduled for "
			"thread ahead.", True);
	    break;
	case AheadFlagThreading:
	    if (current && current->group == group)
		sprintf(message,
			"That group is beeing threaded, %ld%% done!",
			(100 * current->n_done) / current->total);
	    else
		sprintf(message, "That group is beeing threaded!");
	    set_message(message, False);
	    break;
	}
    } else {
	group->ahead =
	    (THREAD_CONTEXT *)XtMalloc(sizeof(THREAD_CONTEXT));
	nullify_thread_context(group->ahead);
	group->ahead->group = group;
	schedule_group(group);
	update_group_entry(group);

	if (ahead_state == AheadStateDisconnected ||
	    ahead_state == AheadStateIdle)
	    thread_ahead_schedule();
    }
}

char thread_ahead_char(THREAD_CONTEXT *context)
{
    if (context)
	switch (context->flag) {
	case AheadFlagThreading:
	    return '*';
	case AheadFlagDone:
	    return '+';
	case AheadFlagScheduled:
	    return '-';
	}

    return ' ';
}

void thread_ahead_shutdown()
{
    long	n;

    if (ahead_server.fd >= 0)
	(void)server_write_raw(&ahead_server, "\r\nQUIT\r\n");
    server_close(&ahead_server);
    ahead_state = AheadStateDisconnected;
    if (input_id != 0)
	XtRemoveInput(input_id);
    input_id = 0;
    if (timeout_id != 0)
	XtRemoveTimeOut(timeout_id);
    timeout_id = 0;
    n_pending = 0;
    n_to_put = 0;
    curr_art = 0;

    for (n = 0 ; n < global.no_groups ; n++)
	if (global.groups[n]->ahead) {
	    THREAD_CONTEXT	*context = global.groups[n]->ahead;

	    global.groups[n]->ahead = NULL;
	    clear_thread_context(context);
	    if (context == current)
		current = NULL;
	    XtFree((char *)context);
	}

    if (current) {
	clear_thread_context(current);
	XtFree((char *)current);
	current = NULL;
    }

    while (schedule) {
	SCHEDULE_NODE	*next = schedule->next;

	XtFree((char *)schedule);
	schedule = next;
    }
}

int thread_ahead_check(GROUP_NODE *group)
{
    THREAD_CONTEXT	*context = group->ahead;
    char		message[128];

    if (context) {
	switch (context->flag) {
	case AheadFlagThreading:
	    sprintf(message,
		    "Thread ahead in progress %ld%% done, try later.",
		    (100 * context->n_done) / context->total);
	    set_message(message, False);
	    return -1;
	case AheadFlagScheduled:
	    unschedule_group(group);
	    update_group_entry(group);
	    break;
	case AheadFlagDone:
	    break;
	}

	global.thr = *context;
	group->ahead = NULL;
	nullify_thread_context(context);
	XtFree((char *)context);
    }

    return 0;
}

void thread_ahead_leave_group(GROUP_NODE *group)
{
    KeepType	keep = GroupObjectKeepThreadInfo(global.group_object);

    if (keep == KeepTypeYes ||
	(group->subscribed && keep == KeepTypeSubscribed)) {
	group->ahead = (THREAD_CONTEXT *)XtMalloc(sizeof(THREAD_CONTEXT));
	*group->ahead = global.thr;
	group->ahead->flag = AheadFlagDone;
	nullify_thread_context(&global.thr);
	global.thr.group = group;
    } else {
	clear_thread_context(&global.thr);
	group->ahead = NULL;
	global.thr.group = group;
    }
}

void thread_em_all()
{
    long	n;

    for (n = 0 ; n < global.no_groups ; n++)
	if (!global.groups[n]->subscribed)
	    break;
	else {
	    schedule_group(global.groups[n]);
	    update_group_entry(global.groups[n]);
	}

    if (ahead_state == AheadStateDisconnected ||
	ahead_state == AheadStateIdle)
	thread_ahead_schedule();
}

void close_ahead_fd()
{
    if (ahead_server.fd >= 0)
	(void)close(ahead_server.fd);
}
