/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "ahead.h"
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "kill.h"
#include "misc.h"
#include "newsrc.h"
#include "server.h"
#include "viewer.h"
#include "widgets.h"
#include "xutil.h"
#include "NntpObject.h"
#include "../Widgets/Dialogue.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"

static Widget connect_dialogue = NULL;

static int server_init()
{
    char	message[256];
    char	*buffer;
    int		status;

    buffer = server_read(&global.server);
    if (!buffer) {
	if (global.aborted)
	    set_message("Aborted.", True);
	else
	    set_message("Error: Connection to server broken!", True);
	return False;
    }
    status = atoi(buffer);
    if (status != OK_CANPOST && status != OK_NOPOST) {
	if (strlen(buffer) > 100) buffer[100] = '\0';
	sprintf(message, "Error: Failed message from server is: %s", buffer);
	set_message(message, True);
	return False;
    }
    global.posting_allowed = (status == OK_CANPOST);

    buffer = server_comm(&global.server, "MODE READER\r\n", False);
    if (!buffer) {
	if (global.aborted)
	    set_message("Aborted.", True);
	else
	    set_message("Error: Connection to server broken!", True);
	return False;
    }
    status = atoi(buffer);
    if (status == OK_CANPOST)
	global.posting_allowed = True;
    else if (status == OK_NOPOST)
	global.posting_allowed = False;
    else if (status == ERR_ACCESS) {
	if (strlen(buffer) > 100) buffer[100] = '\0';
	sprintf(message, "Error: Failed message from server is: %s", buffer);
	set_message(message, True);
	return False;
    }

    buffer = server_comm(&global.server, "XOVER\r\n", False);
    if (!buffer) {
	if (global.aborted)
	    set_message("Aborted.", True);
	else
	    set_message("Error: Connection to server broken!", True);
	return False;
    }
    status = atoi(buffer);
    if (status == ERR_COMMAND)
	global.xover_supported = False;
    else {
	global.xover_supported = True;
	if (status == OK_XOVER) {
	    while (buffer && (buffer[0] != '.' || buffer[1] != '\0'))
		buffer = server_read(&global.server);
	    if (!buffer) {
		if (global.aborted)
		    set_message("Aborted.", True);
		else
		    set_message("Error: Connection to server broken!", True);
		return False;
	    }
	}
    }

    if (global.head_debug)
	global.xover_supported = False;

    return True;
}

void connect_server(void)
{
    int		status = -1;
    Boolean	read_active;

    if (global.nntp_object)
	XtDestroyWidget((Widget)global.nntp_object);
    global.nntp_object =
	NntpObjectCreate(main_widgets.shell, global.nntp_server);
    global.group_object =
	GroupObjectCreate((Widget)global.nntp_object, "none");
    update_misc_menu();

    set_busy(True, True);
    suspend_child_contexts();

    if (global.nntp_server[0] != '#')
	status = server_open(&global.server, global.nntp_server, 2);
    else {
	char	*cmd = get_server_command(global.nntp_server);

	if (cmd)
	    status = server_fork(&global.server, cmd, 2);
	else
	    set_message("Error!  No server command specified!", True);
    }

    if (status < 0) {
	if (global.aborted)
	    set_message("Aborted!", True);
	unset_busy();
	resume_child_contexts();
	return;
    }

    status = server_init();
    if (!status) {
	server_write_raw(&global.server, "\r\nQUIT\r\n");
	server_close(&global.server);
	global_cleanup(False);
	unset_busy();
	resume_child_contexts();
	return;
    }

    read_active = NntpObjectReadActiveFile(global.nntp_object);
    if (read_active) {
	if (global.posting_allowed)
	    set_message("Connected to server; reading active file...", False);
	else
	    set_message("Connected to server, posting not allowed; "
			"reading active file", False);

	status = get_newsgroups();
    } else {
	if (global.posting_allowed)
	    set_message("Connected to server; "
			"retrieving groups from newsrc...", False);
	else
	    set_message("Connected to server, posting not allowed; "
			"retrieving groups from newsrc...", False);
	status = get_newsgroups_from_newsrc();
    }

    if (status == OK_GROUPS) {
	if (NntpObjectRetrieveDescriptions(global.nntp_object)) {
	    set_message("Retrieving group descriptions...", False);
	    status = get_descriptions();
	    if (status != OK_GROUPS && status >= 0) {
		/* LIST NEWSGROUPS failed be silent */
		status = OK_GROUPS;
	    }
	}
    }

    if (read_active && status == OK_GROUPS) {
	set_message("Reading newsrc file...", False);
	parse_newsrc();
    }

    if (status == OK_GROUPS) {
	set_message("Reading kill file...", False);
	parse_kill_file();

	if (NntpObjectCheckForNewGroups(global.nntp_object)) {
	    set_message("Checking for new groups...", False);
	    status = check_for_new_groups();
	    if (status < 0) {
		if (global.aborted)
		    set_message("Aborted.", True);
		else
		    set_message("Connection to server broken!", True);
		server_close(&global.server);
		global_cleanup(False);
		unset_busy();
		resume_child_contexts();
		return;
	    } else if (status != OK_NEWGROUPS) {
		/* NEWGROUPS failed, be silent */
	    }
	}

	if (global.new_groups)
	    setNewsModeNewgroups();
	else
	    setNewsModeConnected();

	if (NntpObjectThreadEmAll(global.nntp_object))
	    thread_em_all();
    } else {
	if (status < 0) {
	    if (global.aborted)
		set_message("Aborted.", True);
	    else
		set_message("Connection to server broken!", True);
	} else {
	    if (read_active)
		set_message("Failed to read active file!", True);
	    else
		set_message("Failed to read newsrc file!", True);
	}
	server_close(&global.server);
	global_cleanup(False);
    }

    unset_busy();
    resume_child_contexts();
    add_rescan_timeout();
}

int reconnect_server(int tell)
{
    int		status = -1;

    if (global.aborted)
	set_message("Aborted, reconnecting to server...", True);
    else
	set_message("Reconnecting to server...", False);
    global.aborted = False;
    suspend_child_contexts(); /* may happen twice */

    (void)server_close(&global.server);
    if (global.nntp_server[0] != '#')
	status = server_open(&global.server, global.nntp_server, 1);
    else {
	char	*cmd = get_server_command(global.nntp_server);

	if (cmd)
	    status = server_fork(&global.server, cmd, 2);
	else
	    set_message("Error!  No server command specified!", True);
    }

    if (status < 0) {
	if (global.aborted)
	    set_message("Aborted!", True);
	/* else server_open will have done it */
	global_cleanup(False);
	resume_child_contexts();
	return -1;
    }

    status = server_init();
    if (!status) {
	server_write_raw(&global.server, "\r\nQUIT\r\n");
	server_close(&global.server);
	global_cleanup(False);
	resume_child_contexts();
	/* server_init will have set_message */
	return -1;
    }

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	char	message[256];
	char	*buffer;

	sprintf(message, "GROUP %s\r\n", global.thr.group->name);
	buffer = server_comm(&global.server, message, False);
	if (!buffer) {
	    set_message("Failed to reconnect to server!", True);
	    server_close(&global.server);
	    resume_child_contexts();
	    global_cleanup(False);
	    return -1;
	}

	status = atoi(buffer);
	if (status != OK_GROUP)
	    setNewsModeConnected();
    }

    if (tell)
	set_standard_message();
    resume_child_contexts();

    return 0;
}

/*************************************************************************/

static void connect_dialogue_callback(Widget w,
				      XtPointer client_data,
				      XtPointer call_data)
{
    DialogueReport	*report = (DialogueReport *)call_data;
    Arg			arg;

    switch (report->reply) {
    case DialogueReplyMiddle:
	XtSetArg(arg, XtNbuffer, "");
	XtSetValues(w, &arg, 1);
	break;
    case DialogueReplyLeft:
    case DialogueReplyEnter:
	XtFree(global.nntp_server);
	global.nntp_server = NULL;
	XtPopdown(w);
	if (report->buffer && report->buffer[0] != '\0') {
	    global.nntp_server = XtNewString(report->buffer);
	    connect_server();
	}
	break;
    case DialogueReplyRight:
    case DialogueReplyClose:
	XtPopdown(w);
	break;
    case DialogueReplyTab:
	break;
    }
}

void popup_connect_dialogue()
{
    if (!connect_dialogue) {
	Arg		args[6];
	Position	x, y;

	XtTranslateCoords(main_widgets.knapp[1], -10, -10, &x, &y);
	XtSetArg(args[0], XtNleftLabel, "Connect");
	XtSetArg(args[1], XtNmiddleLabel, "Clear");
	XtSetArg(args[2], XtNrightLabel, "Cancel");
	XtSetArg(args[3], XtNmessage, "Connect to nntpserver:");
	XtSetArg(args[4], XtNx, x);
	XtSetArg(args[5], XtNy, y);
	connect_dialogue =
	    XtCreatePopupShell("connect", dialogueWidgetClass,
			       main_widgets.shell, args, 6);
	XtAddCallback(connect_dialogue, XtNcallback,
		      connect_dialogue_callback, NULL);
    }

    XtPopup(connect_dialogue, XtGrabNone);
}
