/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include "ahead.h"
#include "context.h"
#include "server.h"

#define STDERR_BUFFLEN	512

typedef struct child_context {
    struct child_context	*next;
    pid_t			pid;
    volatile int		exited; /* FIXME: should be sigatomic_t */
    volatile int		status;
    void			*data;
    child_context_callback	proc;
    char			*stderr_buf;
    int				stderr_fd;
    XtInputId			stderr_id;
} CHILD_CONTEXT;

static CHILD_CONTEXT	*contexts = NULL;

static int		pipes[2];
static XtInputId	pipe_id = 0;

static void sigchld_handler(int sig)
{
    int		done_some = False;

    for (;;) {
	pid_t		pid;
	int		status;
	CHILD_CONTEXT	*loop;

	pid = waitpid(-1, &status, WNOHANG);

	if (pid < 0) {
	    if (errno == EINTR)
		continue; /* this should never happen */
	    else
		break;
	}

	if (pid == 0)
	    break;

	for (loop = contexts ; loop ; loop = loop->next)
	    if (loop->pid == pid) {
		done_some = True;
		loop->exited = True;
		loop->status = status;
		break;
	    }
    }

    if (done_some)
	while (write(pipes[1], &done_some, 1) < 0 &&
	       errno == EINTR);
}

#ifdef _POSIX_SOURCE

static void install_sig_handlers()
{
    struct sigaction	sig_act;

    sig_act.sa_handler = sigchld_handler;
    sigfillset(&sig_act.sa_mask);
    sig_act.sa_flags = 0;
#ifdef SA_RESTART
    sig_act.sa_flags |= SA_RESTART;
#endif
#ifdef SA_NOCLDSTOP
    sig_act.sa_flags |= SA_NOCLDSTOP;
#endif
    if (sigaction(SIGCHLD, &sig_act, NULL) < 0)
	perror("knews: sigaction");
    /*
     *  The Layout widget sometimes generates a SIGFPE. This may be fixed
     *  now, thanks to leo@marco.de, but I might as well keeps this.
     */
    sig_act.sa_handler = SIG_IGN;
    sigemptyset(&sig_act.sa_mask);
    sig_act.sa_flags = 0;
    if (sigaction(SIGPIPE, &sig_act, NULL) < 0)
	perror("knews: sigaction(SIGPIPE)");
    sig_act.sa_handler = SIG_IGN;
    sigemptyset(&sig_act.sa_mask);
    sig_act.sa_flags = 0;
    if (sigaction(SIGFPE, &sig_act, NULL) < 0)
	perror("knews: sigaction(SIGFPE)");
}

static void unblock_sigchld()
{
    sigset_t	set;

    sigemptyset(&set);
    sigaddset(&set, SIGCHLD);
    (void)sigprocmask(SIG_UNBLOCK, &set, NULL);
}

static void block_sigchld()
{
    sigset_t	set;

    sigemptyset(&set);
    sigaddset(&set, SIGCHLD);
    (void)sigprocmask(SIG_BLOCK, &set, NULL);
}

#else /* !defined(_POSIX_SOURCE): potentially unreliable signals */

static void install_sig_handlers()
{
    if (signal(SIGCHLD, sigchld_handler) == SIG_ERR)
	perror("knews: signal(SIGCHLD)");
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	perror("knews: signal(SIGPIPE)");
    if (signal(SIGFPE,  SIG_IGN) == SIG_ERR)
	perror("knews: signal(SIGFPE)");
}

static void unblock_sigchld()
{
    if (signal(SIGCHLD, sigchld_handler) == SIG_ERR)
	perror("knews: signal(SIGCHLD)");
    sigchld_handler(0);
}

static void block_sigchld()
{
    if (signal(SIGCHLD, SIG_DFL) == SIG_ERR)
	perror("knews: signal(SIGCHLD)");
}

#endif

static void stderr_finalize(CHILD_CONTEXT *context)
{
    if (context->stderr_fd >= 0) {
	int	i, n;

	i = strlen(context->stderr_buf);
	n = STDERR_BUFFLEN - i - 1;
	if (n > 0) {
	    char	*c = context->stderr_buf + i;

	    i = read(context->stderr_fd, c, n);
	    if (i >= 0) c[i] = '\0';
	    else c[0] = '\0';
	}

	(void)close(context->stderr_fd);
	context->stderr_fd = -1;
    }

    if (context->stderr_id != 0) {
	XtRemoveInput(context->stderr_id);
	context->stderr_id = 0;
    }
}

static void pipe_read_callback(XtPointer client_data, int *fd, XtInputId *id)
{
    CHILD_CONTEXT	*loop, *prev;
    char		c;

    block_sigchld();

    (void)read(pipes[0], &c, 1);

    loop = contexts;
    prev = NULL;
    while (loop) {
	if (loop->exited) {
	    CHILD_CONTEXT	*next = loop->next;

	    stderr_finalize(loop);
	    if (loop->proc)
		loop->proc(loop->data,
			   loop->status,
			   loop->stderr_buf);
	    if (prev)
		prev->next = next;
	    else
		contexts = next;
	    XtFree(loop->stderr_buf);
	    XtFree((char *)loop);
	    loop = next;
	} else {
	    prev = loop;
	    loop = loop->next;
	}
    }

    unblock_sigchld();
}

void suspend_child_contexts()
{
    if (pipe_id != 0) {
	XtRemoveInput(pipe_id);
	pipe_id = 0;
    }
}

void resume_child_contexts()
{
    if (pipe_id == 0)
	pipe_id = XtAppAddInput(global.app_cont, pipes[0],
				(XtPointer)XtInputReadMask,
				pipe_read_callback, NULL);
}

void init_child_contexts()
{
    long	flags;

    if (pipe(pipes) < 0) {
	perror("knews: pipe");
	exit(1);
    }

    flags = fcntl(pipes[0], F_GETFL);
    if (flags < 0) {
	perror("knews: fcntl");
	exit(1);
    }
    flags |= O_NONBLOCK;
    if (fcntl(pipes[0], F_SETFL, flags) < 0) {
	perror("knews: fcntl");
	exit(1);
    }

    flags = fcntl(pipes[1], F_GETFL);
    if (flags < 0) {
	perror("knews: fcntl");
	exit(1);
    }
    flags |= O_NONBLOCK;
    if (fcntl(pipes[1], F_SETFL, flags) < 0) {
	perror("knews: fcntl");
	exit(1);
    }

    resume_child_contexts();
    install_sig_handlers();
}

static void stderr_read_callback(XtPointer client_data,
				 int *fd,
				 XtInputId *id)
{
    CHILD_CONTEXT	*context = (CHILD_CONTEXT *)client_data;
    int			n, i;

    i = strlen(context->stderr_buf);
    n = STDERR_BUFFLEN - i - 1;

    if (n > 0) {
	char	*c = context->stderr_buf + i;

	i = read(context->stderr_fd, c, n);
	if (i > 0) {
	    c[i] = '\0';
	    n -= i;
	    if (n > 0)
		return;
	} else {
	    c[0] = '\0';
	    if (i < 0) {
		perror("knews: read");
		if (errno == EAGAIN || errno == EWOULDBLOCK)
		    return;
	    }
	}
    }

    XtRemoveInput(context->stderr_id);
    context->stderr_id = 0;
    (void)close(context->stderr_fd);
    context->stderr_fd = -1;
}

static CHILD_CONTEXT *create_child_context(pid_t pid,
					  void *data,
					  child_context_callback proc)
{
    CHILD_CONTEXT	*temp;

    temp = (CHILD_CONTEXT *)XtMalloc(sizeof(CHILD_CONTEXT));
    temp->pid = pid;
    temp->exited = False;
    temp->status = 0;
    temp->data = data;
    temp->proc = proc;
    temp->stderr_buf = NULL;
    temp->stderr_fd = -1;
    temp->stderr_id = 0;
    temp->next = contexts;
    contexts = temp;

    return temp;
}

static void close_child_contexts()
{
    CHILD_CONTEXT	*loop;

    (void)close(pipes[0]);
    (void)close(pipes[1]);
    for (loop = contexts ; loop ; loop = loop->next)
	if (loop->stderr_fd >= 0)
	    (void)close(loop->stderr_fd);
}

pid_t fork_nicely(void *data, child_context_callback proc, int use_stderr)
{
    pid_t	pid;
    int		stderr_fd[2];

    if (!use_stderr || pipe(stderr_fd) < 0) {
	stderr_fd[0] = -1;
	stderr_fd[1] = -1;
    }

    block_sigchld();
    pid = fork();

    if (pid == 0) { /* child */
	if (stderr_fd[0] >= 0)
	    (void)close(stderr_fd[0]);
	if (global.server.cookie)
	    (void)fclose(global.server.cookie);
	close_child_contexts();
	close_ahead_fd();
	(void)close(disp_conn_num);
	(void)close(global.server.fd);

	(void)setsid(); /* can't use setpgrp: Sun is broken */

	if (stderr_fd[1] >= 0 && stderr_fd[1] != STDERR_FILENO) {
	    if (dup2(stderr_fd[1], STDERR_FILENO) < 0)
		perror("knews: dup2");
	}
    } else { /* parent or error */
	if (pid < 0) {
	    perror("knews: fork");
	    if (stderr_fd[0] >= 0)
		(void)close(stderr_fd[0]);
	} else {
	    CHILD_CONTEXT	*context;

	    context = create_child_context(pid, data, proc);
	    if (stderr_fd[0] >= 0) {
		long	flags;

		flags = fcntl(stderr_fd[0], F_GETFL);
		if (flags < 0)
		    perror("knews: fcntl");
		else {
		    flags |= O_NONBLOCK;
		    if (fcntl(stderr_fd[0], F_SETFL, flags) < 0)
			perror("knews: fcntl");
		}
		context->stderr_buf = XtMalloc(STDERR_BUFFLEN);
		context->stderr_buf[0] = '\0';
		context->stderr_fd = stderr_fd[0];
		context->stderr_id =
		    XtAppAddInput(global.app_cont, stderr_fd[0],
				  (XtPointer)XtInputReadMask,
				  stderr_read_callback, (XtPointer)context);
	    }
	}
	if (stderr_fd[1] >= 0)
	    (void)close(stderr_fd[1]);
	unblock_sigchld();
    }

    return pid;
}

pid_t wait_for_pid(pid_t pid, int *status, char *stderr_buf)
{
    CHILD_CONTEXT	*context;
    pid_t		result;

    for (context = contexts ; context ;
	 context = context->next)
	if (context->pid == pid) break;

    if (!context) {
	errno = ECHILD;
	return -1;
    }

    suspend_child_contexts();

    while (!context->exited) {
	(void)do_wait(&pipes[0], True, XtIMXEvent | XtIMAlternateInput);
	(void)read(pipes[0], &result, 1);
    }

    stderr_finalize(context);
    result = context->pid;
    if (stderr_buf)
	strcpy(stderr_buf, context->stderr_buf);
    if (status)
	*status = context->status;

    resume_child_contexts();
    pipe_read_callback(NULL, NULL, NULL);

    return result;
}

char *signal_string(int sig)
{
#define DO(x) if (sig == x) return #x
    DO(SIGHUP);
    DO(SIGINT);
    DO(SIGQUIT);
    DO(SIGABRT);
    DO(SIGBUS);
    DO(SIGFPE);
    DO(SIGKILL);
    DO(SIGSEGV);
    DO(SIGPIPE);
    DO(SIGALRM);
    DO(SIGTERM);
    DO(SIGCHLD);
    DO(SIGCONT);
    DO(SIGSTOP);
#ifdef SIGTSTP
    DO(SIGTSTP);
#endif
#ifdef SIGTTIN
    DO(SIGTTIN);
#endif
#ifdef SIGTTOU
    DO(SIGTTOU);
#endif
#ifdef SIGURG
    DO(SIGURG);
#endif
#ifdef SIGWINCH
    DO(SIGWINCH);
#endif
#ifdef SIGIO
    DO(SIGIO);
#endif
#ifdef SIGPOLL
    DO(SIGPOLL);
#endif

    return "unknown signal";
}
