/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "../configure.h"
#include "global.h"
#include <signal.h>
#include <X11/Shell.h>
#include <X11/Xmu/Xmu.h>
#include "actions.h"
#include "ahead.h"
#include "connect.h"
#include "context.h"
#include "domain.h"
#include "misc.h"
#include "pedit.h"
#include "procs.h"
#include "save.h"
#include "server.h"
#include "tag.h"
#include "uudecode.h"
#include "widgets.h"
#include "xutil.h"

struct Global global;

static int disp_io_error_handler(Display *disp)
{
    perror("knews: fatal I/O error or KillClient");
    delete_window_callback(NULL, NULL, NULL);
    return 0;
}

static int disp_error_handler(Display *disp, XErrorEvent *event)
{
    char	buffer[256];
    char	number[32];

    XGetErrorText(disp, event->error_code, buffer, sizeof(buffer));
    fprintf(stderr,
	    "X Error of failed request               : %s\n",
	    buffer);

    sprintf(number, "%d", event->request_code);
    XGetErrorDatabaseText(disp, "XRequest", number, "",
			  buffer, sizeof(buffer));
    fprintf(stderr,
	    "  Major opcode of failed request        :  %d (%s)\n",
	    event->request_code, buffer);

    switch (event->error_code) {
    case BadValue:
	fprintf(stderr, "  Resource id in failed request         :  ");
	break;
    case BadAtom:
	fprintf(stderr, "  Value in failed request               :  ");
	break;
    default:
	fprintf(stderr, "  Atom id in failed request             :  ");
	break;
    }
    fprintf(stderr, "0x%lx\n", event->resourceid);

    return 0;
}

static void init_timer(XtPointer client_data, XtIntervalId *id)
{
    if (main_widgets.second_shell)
	XtPopup(main_widgets.second_shell, XtGrabNone);
    unset_busy();
    if (global.nntp_server)
	connect_server();
    else {
	global.nntp_object =
	    NntpObjectCreate(main_widgets.shell, "none");
	global.group_object =
	    GroupObjectCreate((Widget)global.nntp_object, "none");
    }
}

int main(int argc, char *argv[])
{
#ifndef DEFAULT_NNTPSERVER
#define DEFAULT_NNTPSERVER 0
#endif
#ifndef DEFAULT_EDIT_COMMAND
#define DEFAULT_EDIT_COMMAND "xterm -e vi +%i %s"
#endif
    static XtResource resource_spec[] = {
#define offset(field) XtOffsetOf(struct Global, field)
	{"nntpServer", "NntpServer", XtRString, sizeof(String),
	 offset(nntp_server), XtRImmediate, (XtPointer)DEFAULT_NNTPSERVER},
	{"editCommand", "EditCommand", XtRString, sizeof(String),
	 offset(edit_command), XtRImmediate, (XtPointer)DEFAULT_EDIT_COMMAND},
	{"mailName", "MailName", XtRString, sizeof(String),
	 offset(mail_name), XtRImmediate, (XtPointer)NULL},
	{"busyCursor", "BusyCursor", XtRCursor, sizeof(Cursor),
	 offset(busy_cursor), XtRString, (XtPointer)"watch"},
	{"cursor", "Cursor", XtRCursor, sizeof(Cursor),
	 offset(cursor), XtRString, (XtPointer)"top_left_arrow"},
	{"stderrTimeout", "StderrTimeout", XtRLong, sizeof(long),
	 offset(stderr_timeout), XtRImmediate, (XtPointer)3000},
	{"aheadTimeout", "AheadTimeout", XtRLong, sizeof(long),
	 offset(ahead_timeout), XtRImmediate, (XtPointer)30000},
	{"chunkSize", "ChunkSize", XtRInt, sizeof(int),
	 offset(chunk_size), XtRImmediate, (XtPointer)16},
	{"knewsVersion", "KnewsVersion", XtRString, sizeof(String),
	 offset(version), XtRImmediate, (XtPointer)NULL},
	{"separateWindows", "SeparateWindows", XtRBoolean, sizeof(Boolean),
	 offset(separate_windows), XtRImmediate, (XtPointer)False},
	{"bell", "Bell", XtRBoolean, sizeof(Boolean),
	 offset(bell), XtRImmediate, (XtPointer)True},
	{"headDebug", "Debug", XtRBoolean, sizeof(Boolean),
	 offset(head_debug), XtRImmediate, (XtPointer)False},
#undef offset
    };
    static XtActionsRec actions[] = {
	{"tree-up",			action_tree_up},
	{"tree-down",			action_tree_down},
	{"tree-left",			action_tree_left},
	{"tree-right",			action_tree_right},
	{"tree-down-right",		action_tree_down_right},
	{"followup",			action_followup},
	{"reply",			action_reply},
	{"followup-and-reply",		action_followup_and_reply},
	{"post-new",			action_post_new},
	{"list-up",			action_list_up},
	{"list-down",			action_list_down},
	{"tree-or-list-up",		action_tree_or_list_up},
	{"tree-or-list-down",		action_tree_or_list_down},
	{"enter-mode",			action_enter_mode},
	{"exit-mode",			action_exit_mode},
	{"tree-left-or-exit-mode",	action_tree_left_or_exit_mode},
	{"tree-right-or-enter-mode",	action_tree_right_or_enter_mode},
	{"read-article",		action_read_article},
	{"view-thread",			action_view_thread},
	{"mark-read-article",		action_mark_read_article},
	{"mark-read-subject",		action_mark_read_subject},
	{"mark-read-thread",		action_mark_read_thread},
	{"mark-read-subthread",		action_mark_read_subthread},
	{"mark-read-tagged",		action_mark_read_tagged},
	{"mark-read-all",		action_mark_read_all},
	{"mark-read-to-current",	action_mark_read_to_current},
	{"mark-unread-article",		action_mark_unread_article},
	{"mark-unread-subject",		action_mark_unread_subject},
	{"mark-unread-thread",		action_mark_unread_thread},
	{"mark-unread-subthread",	action_mark_unread_subthread},
	{"mark-unread-tagged",		action_mark_unread_tagged},
	{"mark-unread-all",		action_mark_unread_all},
	{"mark-unread-killed",		action_mark_unread_killed},
	{"uudecode",			action_uudecode},
	{"clear-tagged",		action_clear_tagged},
	{"save",			action_save},
	{"pipe",			action_pipe},
	{"tag-thread",			action_tag_thread},
	{"tag-subject",			action_tag_subject},
	{"catchup",			action_catchup},
	{"unsubscribe",			action_unsubscribe},
	{"subscribe",			action_subscribe},
	{"realize-thread",		action_realize_thread},
	{"change-size",			action_change_size},
	{"schedule-thread-ahead",	action_schedule_thread_ahead},
	{"popup-find-group",		action_popup_find_group},
    };
    static String fallback_resources[] = {
	"*quoteRegexp:			^[ \t]*[:>|]",
	"*grouplist*PreferredLines:	14",
	"*killist*PreferredLines:	14",
	"*PreferredColumns:		84",
	"*Knapp.BorderWidth:		1",
	"*MenuKnapp.BorderWidth:	1",
	"*TextField.BorderWidth:	2",
	"*BorderWidth:			0",
	"*Toggle.shadowWidth:		2",
	"*textscrbar.allowOff:		True",
	"*Knapp.Justify:		center",
	"*MenuKnapp.Justify:		center",
	"*quotetoggle.set:		True",
	/* colors */
	"*Foreground:		Black",
	"*Background:		Light Steel Blue",
	"*rubberColor:		Red",
	"*quoteColor:		Medium Blue",
	"*headerColor:		#000090",
	"*innerColor:		Red",
	"*knapp9.Foreground:	Red",
	/* fonts */
	"*ArtText*quoteFont:	"
	"-b&h-lucidatypewriter-bold-r-normal-*-*-120-*-*-*-*-iso8859-1",
	"*ArtText*headerFont:	"
	"-b&h-lucidatypewriter-bold-r-normal-*-*-120-*-*-*-*-iso8859-1",
	"*ArtText*Font:		"
	"-b&h-lucidatypewriter-medium-r-normal-*-*-120-*-*-*-*-iso8859-1",
	"*ScrList*Font:		"
	"-b&h-lucidatypewriter-medium-r-normal-*-*-120-*-*-*-*-iso8859-1",
	"*Font:			"
	"-b&h-lucidatypewriter-bold-r-normal-*-*-120-*-*-*-*-iso8859-1",
	/* layouts */
#include "layouts.h"
	/* translations */
	"*arttree.baseTranslations:	#augment \\n"
	"	<Btn1Down>:	set-selected()\\n"
	"	<Btn3Down>:	toggle-outer()",
	"*grouplist.baseTranslations:	#augment \\n"
	"	<Btn2Down>:	dnd-start() \\n"
	"	<Btn2Motion>:	dnd-do() \\n"
	"	<Btn2Up>:	dnd-end()",
	"*killist.baseTranslations:	#augment \\n"
	"	<Btn2Down>:	dnd-start() \\n"
	"	<Btn2Motion>:	dnd-do() \\n"
	"	<Btn2Up>:	dnd-end()",
	"*threadlist.baseTranslations:	#augment \\n"
	"	~Ctrl <Btn2Down>:	select() mark-read-thread() \\n"
	"	~Ctrl <Btn2Motion>:	select() mark-read-thread() \\n"
	"	Ctrl <Btn2Down>:	select() mark-unread-thread() \\n"
	"	<Btn3Down>:		select() tag-thread()",
	NULL,
    };
    static XrmOptionDescRec options[] = {
	{"-nntpServer",		".nntpServer",
	 XrmoptionSepArg,	(XtPointer)NULL},
	{"-descriptions",	"*retrieveDescriptions",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+descriptions",	"*retrieveDescriptions",
	 XrmoptionNoArg,	(XtPointer)"True"},
	{"-active",		"*readActiveFile",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+active",		"*readActiveFile",
	 XrmoptionNoArg,	(XtPointer)"True"},
	{"-fill",		"*fillNewsrcFile",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+fill",		"*fillNewsrcFile",
	 XrmoptionNoArg,	(XtPointer)"True"},
	{"-separate",		".separateWindows",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+separate",		".separateWindows",
	 XrmoptionNoArg,	(XtPointer)"True"},
	{"-bell",		".bell",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+bell",		".bell",
	 XrmoptionNoArg,	(XtPointer)"True"},
	{"-onlyUnread",		"*onlyUnread",
	 XrmoptionNoArg,	(XtPointer)"False"},
	{"+onlyUnread",		"*onlyUnread",
	 XrmoptionNoArg,	(XtPointer)"True"},
    };
    Arg			arg = {XtNinput, True};
    regex_t		re;

    (void)freopen("/dev/null", "r", stdin);
    (void)freopen("/dev/null", "w", stdout);

#if (XtSpecificationRelease > 4)
    XtSetLanguageProc(NULL, NULL, NULL);
#endif

#if (XtSpecificationRelease < 6)
    main_widgets.shell =
	XtAppInitialize(&global.app_cont, "Knews",
			options, XtNumber(options),
			&argc, argv, fallback_resources,
			&arg, 1);
#else
    main_widgets.shell =
	XtOpenApplication(&global.app_cont, "Knews",
			  options, XtNumber(options),
			  &argc, argv, fallback_resources,
			  sessionShellWidgetClass, &arg, 1);
#endif
    if (!main_widgets.shell)
	exit(1);

    if (argc != 1)
	fprintf(stderr,
		"Warning: Bad command line argument.\n"
		"         See the man-page for usage.\n");

    if (regcomp(&re, "^[:>]", REGEXP_COMPILE_FLAGS) != 0)
	fprintf(stderr,
		"Warning:  regcomp() couldn't compile \"^[:>]\"\n"
		"          Chances are your system's regex "
		"implementaton sucks.\n"
		"          You might want to recompile knews "
		"with NEED_POSIX_REGEXPS.\n"
		"          See the file knews.tmpl for details.\n");
    regfree(&re);

    XSetErrorHandler(disp_error_handler);
    XSetIOErrorHandler(disp_io_error_handler);
    XtAddConverter(XtRString, XtRLong, XmuCvtStringToLong, NULL, 0);

    init_child_contexts();

    XtAppAddActions(global.app_cont, actions, XtNumber(actions));

    XtGetApplicationResources(main_widgets.shell, (XtPointer)&global,
			      resource_spec, XtNumber(resource_spec),
			      NULL, 0);
    fix_domain_stuff();

    if (global.ahead_timeout < 0)
	global.ahead_timeout = 0;

    if (global.nntp_server && global.nntp_server[0] != '\0')
	global.nntp_server = XtNewString(global.nntp_server);
    else {
	char	*env_var = getenv("NNTPSERVER");

	if (env_var && env_var[0] != '\0')
	    global.nntp_server = XtNewString(env_var);
	else
	    global.nntp_server = NULL;
    }

    if (global.chunk_size <= 0)
	global.chunk_size = 16;

    global.server.fd = -1;

    if (!global.version)
	fprintf(stderr,
		"Warning: Application defaults file "
		"not properly installed.\n");
    else if (strncmp(global.version, KNEWS_VERSION, 5) != 0)
	fprintf(stderr,
		"Warning: Incompatible version of "
		"application defaults file.\n");

    create_main_widgets(main_widgets.shell);
    disp_conn_num = ConnectionNumber(XtDisplay(main_widgets.shell));

    setNewsModeDisconnected();
    set_standard_message();
    set_busy(False, False);
    XtAppAddTimeOut(global.app_cont, 500, init_timer, NULL);

    XtAppMainLoop(global.app_cont);

    return 0;
}
