/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "actions.h"
#include "kill.h"
#include "misc.h"
#include "newsrc.h"
#include "procs.h"
#include "read.h"
#include "search.h"
#include "tag.h"
#include "util.h"
#include "uudecode.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtText.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/PopdownSh.h"
#include "../Widgets/Menu.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/PullRight.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "../Widgets/SeparatorG.h"
#include "../Widgets/StringG.h"
#include "../Widgets/ToggleG.h"

typedef enum {
    MarkScopeArticle,
    MarkScopeSubject,
    MarkScopeThread,
    MarkScopeSubthread,
    MarkScopeTagged,
    MarkScopeAllarticles,
    MarkScopeKilled,
    MarkScopeToCurrent
} MarkScope;

static Widget head_toggle;
static Widget keep_thr_toggle;
static Widget subs_unsubs;

void update_misc_menu()
{
    KeepType	keep = GroupObjectKeepThreadInfo(global.group_object);

    ToggleGadgetSet(head_toggle,
		    GroupObjectFullHeader(global.group_object));
    ToggleGadgetSet(keep_thr_toggle,
		    keep == KeepTypeYes ||
		    (global.thr.group && keep == KeepTypeSubscribed &&
		     global.thr.group->subscribed));

    if (global.thr.group) {
	Arg	arg;

	arg.name = XtNlabel;
	if (global.thr.group->subscribed)
	    arg.value = (XtArgVal)"Unsubscribe";
	else
	    arg.value = (XtArgVal)"Subscribe";
	XtSetValues(subs_unsubs, &arg, 1);
    }
}

static void toggle_subs_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    Boolean	temp;

    if (global.busy ||
	(global.mode != NewsModeGroup &&
	 global.mode != NewsModeThread) ||
	!global.thr.group) return;

    temp = global.thr.group->subscribed;
    global.thr.group->subscribed = !temp;
    sort_groups();
    if (temp)
	set_message("Unsubscribed.", False);
    else
	set_message("Subscribed.",   False);
    update_misc_menu();
}

static void full_header_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (global.busy) return;

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	if (global.curr_art)
	    (void)read_article(global.curr_art, True, NULL);
	else
	    set_message("No selected article!", True);
    }
}

static void rot13_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    if (global.busy) return;

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	long		first, shown, size;

	ArtTextRot13(main_widgets.text);

	ArtTextGetFirstShownSize(main_widgets.text, &first, &shown, &size);
	ScrBarSetLengthsAndPos(main_widgets.text_scrbar, size, shown, size);
    }
}

static void clear_tags_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    if (global.busy) return;

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	clear_tagged_articles();
	set_message("Cleared tagged articles.", False);
    }
}

static void toggle_head_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (global.busy) return;

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	if (call_data && global.group_object) {
	    Arg	arg;

	    arg.name = XtNfullHeader;
	    arg.value = (XtArgVal)(*(Boolean *)call_data);
	    XtSetValues((Widget)global.group_object, &arg, 1);
	}
    }
}

static void keep_thr_callback(Widget w,
			      XtPointer client_data,
			      XtPointer call_data)
{
    if (global.busy) return;

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	if (call_data && global.group_object) {
	    Arg	arg;

	    if (*(Boolean *)call_data)
		XtSetArg(arg, XtNkeepThreadInfo, KeepTypeYes);
	    else
		XtSetArg(arg, XtNkeepThreadInfo, KeepTypeNo);
	    XtSetValues((Widget)global.group_object, &arg, 1);
	}
    }
}

static void mark_read_callback(Widget w,
			       XtPointer client_data,
			       XtPointer call_data)
{
    MarkScope	scope = (MarkScope)client_data;
    ARTICLE	*art = global.curr_art;
    SUBJECT	*subj;
    long	n;
    int		xref = GroupObjectProcessXrefs(global.group_object);
    int		fixed_author = GroupObjectFixedAuthor(global.group_object);

    if (global.busy ||
	(global.mode != NewsModeGroup &&
	 global.mode != NewsModeThread)) return;

    if (art)
	subj = art->subject;
    else
	subj = global.curr_subj;

    set_busy(False, True);

    switch (scope) {
    case MarkScopeArticle:
	if (!art)
	    set_message("No selected article!", True);
	else {
	    if (art->from && !art->read) {
		art->read = True;
		subj->no_unread--;
		global.thr.group->no_unread--;
		if (xref)
		    process_xref(art);
		if (art->hot > 0)
		    update_subj_hot_value(subj);
		update_subj_entry(subj, fixed_author);
		if (global.mode == NewsModeThread) {
		    ArtTreeNodeSetInner(main_widgets.arttree,
					(ART_TREE_NODE *)art, False);
		    ArtTreeNodeSetPixmap(main_widgets.arttree,
					 (ART_TREE_NODE *)art, None);
		}
	    }
	    set_standard_message();
	}
	break;
    case MarkScopeSubject:
	if (!subj)
	    set_message("No selected subject!", True);
	else {
	    (void)mark_subject_read(subj, xref, False);
	    update_subj_hot_value(subj);
	    update_subj_entry(subj, fixed_author);
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeThread:
	if (!subj)
	    set_message("No selected subject!", True);
	else {
	    art = subj->thread;

	    while (subj->prev && subj->prev->thread == art)
		subj = subj->prev;
	    while (subj && subj->thread == art) {
		(void)mark_subject_read(subj, xref, False);
		update_subj_hot_value(subj);
		update_subj_entry(subj, fixed_author);
		subj = subj->next;
	    }
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeSubthread:
	if (!art) 
	    set_message("No selected article!", True);
	else {
	    SUBJECT	*loop = subj;

	    while (loop->prev && loop->prev->thread == subj->thread)
		loop = loop->prev;
	    while (loop && loop->thread == subj->thread) {
		(void)mark_sub_subject_read(art, loop, xref, False);
		update_subj_hot_value(loop);
		update_subj_entry(loop, fixed_author);
		loop = loop->next;
	    }
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeTagged:
	n = no_tagged_articles();
	if (n <= 0)
	    set_message("No tagged articles!", True);
	else {
	    ARTICLE	**arts = get_tagged_articles();
	    long	i;

	    for (i = 0 ; i < n ; i++) if (arts[i]->from && !arts[i]->read) {
		arts[i]->read = True;
		arts[i]->subject->no_unread--;
		global.thr.group->no_unread--;
		if (xref)
		    process_xref(arts[i]);
		if (arts[i]->hot > 0)
		    update_subj_hot_value(arts[i]->subject);
		update_subj_entry(arts[i]->subject, fixed_author);
		if (global.mode == NewsModeThread) {
		    ArtTreeNodeSetInner(main_widgets.arttree,
					(ART_TREE_NODE *)arts[i], False);
		    ArtTreeNodeSetPixmap(main_widgets.arttree,
					 (ART_TREE_NODE *)arts[i], None);
		}
	    }

	    clear_tagged_articles();
	    set_standard_message();
	}
	break;
    case MarkScopeAllarticles:
	for (art = global.thr.articles ; art ; art = art->next) {
	    if (xref)
		process_xref(art);
	    art->read = True;
	}

	for (subj = global.thr.subjects ; subj ; subj = subj->next) {
	    subj->no_unread = 0;
	    subj->hot = 0;
	}

	global.thr.group->no_unread = 0;
	global.curr_subj = NULL;
	global.curr_art = NULL;

	global.mode = NewsModeGroup; /* in case we're in NewsModeThread */
	setNewsModeGroup(False);

	set_standard_message();
	break;
    case MarkScopeKilled: /*notreached*/
	break;
    case MarkScopeToCurrent:
	if (!subj)
	    set_message("No selected subject!", True);
	else {
	    SUBJECT	*loop;

	    art = subj->thread;
	    while (subj && subj->thread == art)
		subj = subj->next;

	    for (loop = global.thr.subjects ;
		 loop && loop != subj ;
		 loop = loop->next) {
		(void)mark_subject_read(loop, xref, False);
		update_subj_hot_value(loop);
		update_subj_entry(loop, fixed_author);
	    }

	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    }

    unset_busy();
}

static void mark_unread_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    MarkScope	scope = (MarkScope)client_data;
    ARTICLE	*art = global.curr_art;
    SUBJECT	*subj = global.curr_subj;
    long	n;
    int		fixed_author = GroupObjectFixedAuthor(global.group_object);

    if (global.busy ||
	(global.mode != NewsModeGroup &&
	 global.mode != NewsModeThread)) return;

    if (art)
	subj = art->subject;

    set_busy(False, True);

    switch (scope) {
    case MarkScopeArticle:
	if (!art)
	    set_message("No selected article!", True);
	else {
	    if (art->from && art->read) {
		art->read = False;
		subj->no_unread++;
		global.thr.group->no_unread++;
		if (art->hot > 0)
		    update_subj_hot_value(subj);
		update_subj_entry(subj, fixed_author);
		if (global.mode == NewsModeThread) {
		    ArtTreeNodeSetInner(main_widgets.arttree,
					(ART_TREE_NODE *)art, True);
		    if (art->hot > 0) {
			ArtTreeNodeSetPixmap(main_widgets.arttree,
					     (ART_TREE_NODE *)art,
					     global.hot_pixmaps[art->hot]);
		    }
		}
	    }
	    set_standard_message();
	}
	break;
    case MarkScopeSubject:
	if (!subj)
	    set_message("No selected subject!", True);
	else {
	    (void)mark_subject_unread(subj);
	    update_subj_hot_value(subj);
	    update_subj_entry(subj, fixed_author);
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeThread:
	if (!subj)
	    set_message("No selected subject!", True);
	else {
	    art = subj->thread;

	    while (subj->prev && subj->prev->thread == art)
		subj = subj->prev;
	    while (subj && subj->thread == art) {
		(void)mark_subject_unread(subj);
		update_subj_hot_value(subj);
		update_subj_entry(subj, fixed_author);
		subj = subj->next;
	    }
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeSubthread:
	if (!art)
	    set_message("No selected article!", True);
	else {
	    SUBJECT	*loop = subj;

	    while (loop->prev && loop->prev->thread == subj->thread)
		loop = loop->prev;
	    while (loop && loop->thread == subj->thread) {
		(void)mark_sub_subject_unread(art, loop);
		update_subj_hot_value(loop);
		update_subj_entry(loop, fixed_author);
		loop = loop->next;
	    }
	    if (global.mode == NewsModeThread)
		setNewsModeThread();
	    set_standard_message();
	}
	break;
    case MarkScopeTagged:
	n = no_tagged_articles();
	if (n <= 0)
	    set_message("No tagged articles!", True);
	else {
	    ARTICLE	**arts = get_tagged_articles();
	    long	i;

	    for (i = 0 ; i < n ; i++) if (arts[i]->from && arts[i]->read) {
		arts[i]->read = False;
		arts[i]->subject->no_unread++;
		global.thr.group->no_unread++;
		if (arts[i]->hot > 0)
		    update_subj_hot_value(arts[i]->subject);
		update_subj_entry(arts[i]->subject, fixed_author);
		if (global.mode == NewsModeThread) {
		    ArtTreeNodeSetInner(main_widgets.arttree,
					(ART_TREE_NODE *)arts[i], True);
		    ArtTreeNodeSetPixmap(main_widgets.arttree,
					 (ART_TREE_NODE *)arts[i],
					 global.hot_pixmaps[arts[i]->hot]);
		}
	    }

	    clear_tagged_articles();
	    set_standard_message();
	}
	break;
    case MarkScopeAllarticles:
	for (subj = global.thr.subjects ; subj ; subj = subj->next) {
	    subj->no_unread = 0;
	    subj->hot = 0;
	}
	for (art = global.thr.articles ; art ; art = art->next)
	    art->read = True;

	for (n = 0, subj = global.thr.subjects ; subj ; subj = subj->next) {
	    if (subj->prev && subj->prev->thread == subj->thread)
		continue;
	    for (art = subj->thread ; art ; art = next_in_thread_wrap(art)) {
		if (art->from && art->read) {
		    art->read = False;
		    art->subject->no_unread++;
		    n++;
		    if (art->hot > 0 && art->subject->hot == 0)
			art->subject->hot = art->hot;
		}
	    }
	}

	global.thr.group->no_unread = n;
	global.curr_subj = NULL;
	global.curr_art = NULL;

	global.mode = NewsModeGroup; /* in case we're in NewsModeThread */
	setNewsModeGroup(False);

	set_standard_message();
	break;
    case MarkScopeKilled:
	for (n = 0, art = global.thr.articles ; art ; art = art->next) {
	    if (art->read && art->killed && art->from) {
		art->read = False;
		art->subject->no_unread++;
		n++;
	    }
	}

	global.thr.group->no_unread += n;
	global.curr_subj = NULL;
	global.curr_art = NULL;

	global.mode = NewsModeGroup; /* in case we're in NewsModeThread */
	setNewsModeGroup(False);

	set_standard_message();
	break;
    case MarkScopeToCurrent: /* not reached */
	break;
    }

    unset_busy();
}

static void unsubscribe_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    action_unsubscribe(w, NULL, NULL, NULL);
}

static void catchup_callback(Widget w,
			     XtPointer client_data,
			     XtPointer call_data)
{
    action_catchup(w, NULL, NULL, NULL);
}

static void findgroup_callback(Widget w,
			       XtPointer client_data,
			       XtPointer call_data)
{
    Position	x, y;

    if (!w)
	return;

    XtTranslateCoords(w, -10, -10, &x, &y);
    popup_find_group(x, y);
}

#if 0
static void realize_thread_callback(Widget w,
				    XtPointer client_data,
				    XtPointer call_data)
{
    action_realize_thread(w, NULL, NULL, NULL);
}
#endif

void create_misc_menu1(Widget parent)
{
    Widget	menu, subshell, submenu;
    Widget	temp;
    Arg		arg[3];

    main_widgets.misc_menu1 =
	XtCreatePopupShell("miscshell1", popdownShellWidgetClass,
			   parent, NULL, 0);
    menu = XtCreateManagedWidget("miscmenu1", menuWidgetClass,
				 main_widgets.misc_menu1, NULL, 0);

    arg[0].name = XtNlabel;
    arg[1].name = XtNmenuName;

    arg[0].value = (XtArgVal)"Full Header";
    temp = MenuCreateGadget("fullheader", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, full_header_callback, NULL);
#if 0
    arg[0].value = (XtArgVal)"Realize thread";
    temp = MenuCreateGadget("realize", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, realize_thread_callback, NULL);
#endif
    arg[0].value = (XtArgVal)"Rot13";
    temp = MenuCreateGadget("rot13", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, rot13_callback, NULL);
    arg[0].value = (XtArgVal)"Clear tags";
    temp = MenuCreateGadget("cleartags", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, clear_tags_callback, NULL);
    arg[0].value = (XtArgVal)"Uudecode";
    temp = MenuCreateGadget("uudecode", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, uudecode_callback, NULL);
    arg[0].value = (XtArgVal)"Unsubscribe";
    subs_unsubs = MenuCreateGadget("unsubs", stringGadgetClass, menu, arg, 1);
    XtAddCallback(subs_unsubs, XtNcallback, toggle_subs_callback, NULL);
    (void)MenuCreateGadget("separator1", separatorGadgetClass, menu, NULL, 0);
    arg[0].value = (XtArgVal)"Mark Read";
    arg[1].value = (XtArgVal)"markreadshell";
    (void)MenuCreateGadget("markread", pullRightGadgetClass, menu, arg, 2);
    arg[0].value = (XtArgVal)"Mark Unread";
    arg[1].value = (XtArgVal)"markunreadshell";
    (void)MenuCreateGadget("markunread", pullRightGadgetClass, menu, arg, 2);
    (void)MenuCreateGadget("separator1", separatorGadgetClass, menu, NULL, 0);
    arg[0].value = (XtArgVal)"Full Header";
    head_toggle =
	MenuCreateGadget("headertoggle", toggleGadgetClass,
			 menu, arg, 1);
    XtAddCallback(head_toggle, XtNcallback, toggle_head_callback, NULL);
    arg[0].value = (XtArgVal)"Keep Thread Info";
    keep_thr_toggle =
	MenuCreateGadget("thrinfo", toggleGadgetClass, menu, arg, 1);
    XtAddCallback(keep_thr_toggle, XtNcallback, keep_thr_callback, NULL);

    subshell =
	XtCreatePopupShell("markreadshell", popdownShellWidgetClass,
			   parent, NULL, 0);
    submenu =
	XtCreateManagedWidget("markreadmenu", menuWidgetClass,
			      subshell, NULL, 0);

    arg[0].value = (XtArgVal)"Article";
    temp = MenuCreateGadget("article", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeArticle);
    arg[0].value = (XtArgVal)"Subject";
    temp = MenuCreateGadget("subject", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeSubject);
    arg[0].value = (XtArgVal)"Thread";
    temp = MenuCreateGadget("thread", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeThread);
    arg[0].value = (XtArgVal)"Subthread";
    temp = MenuCreateGadget("subthread", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeSubthread);
    arg[0].value = (XtArgVal)"Tagged";
    temp = MenuCreateGadget("tagged", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeTagged);
    arg[0].value = (XtArgVal)"All articles";
    temp = MenuCreateGadget("allarticles", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeAllarticles);
    arg[0].value = (XtArgVal)"To current";
    temp = MenuCreateGadget("tocurrent", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_read_callback, (XtPointer)MarkScopeToCurrent);

    subshell =
	XtCreatePopupShell("markunreadshell", popdownShellWidgetClass,
			   parent, NULL, 0);
    submenu =
	XtCreateManagedWidget("markunreadmenu", menuWidgetClass,
			      subshell, NULL, 0);

    arg[0].value = (XtArgVal)"Article";
    temp = MenuCreateGadget("article", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeArticle);
    arg[0].value = (XtArgVal)"Subject";
    temp = MenuCreateGadget("subject", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeSubject);
    arg[0].value = (XtArgVal)"Thread";
    temp = MenuCreateGadget("thread", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeThread);
    arg[0].value = (XtArgVal)"Subthread";
    temp = MenuCreateGadget("subthread", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeSubthread);
    arg[0].value = (XtArgVal)"Tagged";
    temp = MenuCreateGadget("tagged", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeTagged);
    arg[0].value = (XtArgVal)"All articles";
    temp = MenuCreateGadget("allarticles", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeAllarticles);
    arg[0].value = (XtArgVal)"Killed";
    temp = MenuCreateGadget("killed", stringGadgetClass, submenu, arg, 1);
    XtAddCallback(temp, XtNcallback,
		  mark_unread_callback, (XtPointer)MarkScopeKilled);
}

void create_misc_menu2(Widget parent)
{
    Widget	menu, temp;
    Arg		arg[3];

    main_widgets.misc_menu2 =
	XtCreatePopupShell("miscshell2", popdownShellWidgetClass,
			   parent, NULL, 0);
    menu = XtCreateManagedWidget("miscmenu2", menuWidgetClass,
				 main_widgets.misc_menu2, NULL, 0);

    arg[0].name = XtNlabel;

    arg[0].value = (XtArgVal)"Unsubscribe";
    temp = MenuCreateGadget("unsubscribe", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, unsubscribe_callback, NULL);

    arg[0].value = (XtArgVal)"Catchup";
    temp = MenuCreateGadget("catchup", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, catchup_callback, NULL);

    arg[0].value = (XtArgVal)"Find group";
    temp = MenuCreateGadget("findgroup", stringGadgetClass, menu, arg, 1);
    XtAddCallback(temp, XtNcallback, findgroup_callback, NULL);
}

void action_mark_read_article(Widget w, XEvent *event,
			      String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeArticle, NULL);
}

void action_mark_read_subject(Widget w, XEvent *event,
			      String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeSubject, NULL);
}

void action_mark_read_thread(Widget w, XEvent *event,
			     String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeThread, NULL);
}

void action_mark_read_subthread(Widget w, XEvent *event,
				String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeSubthread, NULL);
}

void action_mark_read_tagged(Widget w, XEvent *event,
			     String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeTagged, NULL);
}

void action_mark_read_all(Widget w, XEvent *event,
			  String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeAllarticles, NULL);
}

void action_mark_read_to_current(Widget w, XEvent *event,
				 String *params, Cardinal *no_params)
{
    mark_read_callback(w, (XtPointer)MarkScopeToCurrent, NULL);
}

void action_mark_unread_article(Widget w, XEvent *event,
				String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeArticle, NULL);
}

void action_mark_unread_subject(Widget w, XEvent *event,
				String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeSubject, NULL);
}

void action_mark_unread_thread(Widget w, XEvent *event,
			       String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeThread, NULL);
}

void action_mark_unread_subthread(Widget w, XEvent *event,
				  String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeSubthread, NULL);
}

void action_mark_unread_tagged(Widget w, XEvent *event,
			       String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeTagged, NULL);
}

void action_mark_unread_all(Widget w, XEvent *event,
			    String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeAllarticles, NULL);
}

void action_mark_unread_killed(Widget w, XEvent *event,
			       String *params, Cardinal *no_params)
{
    mark_unread_callback(w, (XtPointer)MarkScopeKilled, NULL);
}

void action_clear_tagged(Widget w, XEvent *event,
			       String *params, Cardinal *no_params)
{
    clear_tags_callback(w, NULL, NULL);
}

void action_catchup(Widget w, XEvent *event,
		    String *params, Cardinal *no_params)
{
    GROUP_NODE	*g;
    long	n;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeDisconnected:
	break;
    case NewsModeConnected:
	set_curr_group();
	g = global.thr.group;
	if (!g) {
	    set_message ("No selected group!", True);
	    break;
	}

	free_read_arts_list(g);
	g->read_arts = NULL;
	if (g->last_art >= g->first_art &&
	    g->last_art > 0) {
	    g->read_arts =
		(ART_LIST_NODE *)XtMalloc(sizeof(ART_LIST_NODE));
	    g->read_arts->first = 1;
	    g->read_arts->last = g->last_art;
	    g->read_arts->next = NULL;
	}
	setNewsModeConnected();
	break;
    case NewsModeGroup:
    case NewsModeThread:
	knapp6_callback(w, NULL, NULL);
	break;
    case NewsModeAllgroups:
	n = ScrListGetFirstSelected(main_widgets.group_list);
	if (n < 0) {
	    set_message("No selected groups.", True);
	    return;
	}

	while (n >= 0) {
	    g = global.groups[n];
	    free_read_arts_list(g);
	    g->read_arts = NULL;
	    if (g->last_art >= g->first_art &&
		g->last_art > 0) {
		g->read_arts =
		    (ART_LIST_NODE*)XtMalloc(sizeof(ART_LIST_NODE));
		g->read_arts->first = 1;
		g->read_arts->last = g->last_art;
		g->read_arts->next = NULL;
	    }
	    n = ScrListGetNextSelected(main_widgets.group_list, n);
	}

	set_message("Catched up.", False);
	break;
    case NewsModeNewgroups:
	break;
    }
}

void action_unsubscribe(Widget w, XEvent *event,
			String *params, Cardinal *no_params)
{
    if (global.busy) return;

    switch (global.mode) {
    case NewsModeDisconnected:
	break;
    case NewsModeConnected:
	set_curr_group();
	if (!global.thr.group) {
	    set_message("No selected group!", True);
	    return;
	}

	global.thr.group->subscribed = False;
	sort_groups();
	setNewsModeConnected();

	break;
    case NewsModeGroup:
    case NewsModeThread:
	set_curr_group();
	if (global.thr.group && global.thr.group->subscribed)
	    toggle_subs_callback(w, NULL, NULL);
	break;
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	knapp2_callback(w, NULL, NULL);
	break;
    }
}

void action_subscribe(Widget w, XEvent *event,
		      String *params, Cardinal *no_params)
{
    if (global.busy) return;

    switch (global.mode) {
    case NewsModeDisconnected:
    case NewsModeConnected:
	break;
    case NewsModeGroup:
    case NewsModeThread:
	if (global.thr.group && !global.thr.group->subscribed)
	    toggle_subs_callback(w, NULL, NULL);
	break;
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	knapp1_callback(w, NULL, NULL);
	break;
    }
}
