/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <utime.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include "codes.h"
#include "newsrc.h"
#include "server.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Message.h"

#define NO_GROUPS	4096
#define NEW_GROUPS	 256

#define MAX_GROUP_LEN	1024

#define IS_SPACE(c)	((c) == ' ' || (c) == '\t')

static int comp_group_nodes(const void *a, const void *b)
{
    return strcmp( (*((GROUP_NODE **)a))->name, (*((GROUP_NODE **)b))->name);
}

int get_newsgroups()
{
    char	*buffer;
    long	status;

    buffer = server_comm(&global.server, "LIST\r\n", False);
    if (!buffer)
	return -1;
    sscanf(buffer, "%ld", &status);
    if (status != OK_GROUPS)
	return status;

    global.no_groups = 0;
    for (buffer = server_read(&global.server) ;
	 buffer && (buffer[0] != '.' || buffer[1] != '\0') ;
	 buffer = server_read(&global.server)) {
	char *c;
	long first, last;
	char mod;

	if (global.no_groups > global.max_groups - 2) {
	    long	i = global.max_groups;

	    global.max_groups += NO_GROUPS;
	    global.groups =
		(GROUP_NODE **)XtRealloc((char *)global.groups,
					 global.max_groups *
					 sizeof(GROUP_NODE *));

	    while (i < global.max_groups)
		global.groups[i++] = NULL;

	}

	c = strchr(buffer, ' ');
	if (c) {
	    GROUP_NODE	*group;
	    int		k;

	    *(c++) = '\0';
	    k = sscanf(c, "%ld%ld%*c%c", &last, &first, &mod);
	    if (k < 2)
		continue;
	    if (k == 2)
		mod = 'y';
	    if (mod == '=' || mod == 'x')
		continue;
	    group = (GROUP_NODE *)XtMalloc(sizeof(GROUP_NODE));
	    global.groups[global.no_groups++] = group;
	    group->description = NULL;
	    group->ahead = NULL;
	    group->read_arts = NULL;
	    group->subscribed = False;
	    group->found_in_newsrc = False;
	    group->moderated = (mod == 'm');
	    group->name = XtNewString(buffer);
	    group->first_art = first;
	    group->last_art = last;
	}
    }

    if (!buffer) {
	long	i;

	for (i = 0 ; i < global.no_groups ; i++)
	    XtFree(global.groups[i]->name);
	XtFree((char *)global.groups);

	global.groups = NULL;
	global.no_groups = 0;
	global.max_groups = 0;

	return -1;
    }

    qsort(global.groups, global.no_groups,
	  sizeof(GROUP_NODE *), comp_group_nodes);

    return OK_GROUPS;
}

/* the grouplist must be sorted here */
GROUP_NODE *find_group(char *name)
{
    long	first = 0;
    long	last = global.no_groups - 1;

    while (first <= last &&
	   global.groups[first]->subscribed) {
	if (strcmp(name, global.groups[first]->name) == 0)
	    return global.groups[first];
	first++;
    }

    while (first <= last) {
	long	mid = (first + last) / 2;
	int	comp = strcmp(name, global.groups[mid]->name);

	if (comp == 0)
	    return global.groups[mid];

	if (comp < 0)
	    last = mid - 1;
	else
	    first = mid + 1;
    }

    return NULL;
}

int check_for_new_groups()
{
    GROUP_NODE	**list = NULL;
    long	l_max = 0;
    struct tm	*tm;
    char	command[64];
    char	*buffer;
    long	status, n = 0;

    if (!NntpObjectCheckForNewGroups(global.nntp_object) ||
	global.last_time == 0)
	return True;

    global.new_groups = NULL;
    global.no_new_groups = 0;

    tm = gmtime(&global.last_time);
    sprintf(command, "NEWGROUPS %02d%02d%02d %02d%02d%02d GMT\r\n",
	    tm->tm_year, tm->tm_mon+1, tm->tm_mday,
	    tm->tm_hour, tm->tm_min, tm->tm_sec);
    buffer = server_comm(&global.server, command, False);
    if (!buffer)
	return -1;
    status = atoi(buffer);
    if (status != OK_NEWGROUPS)
	return status;

    l_max = NEW_GROUPS;
    list = (GROUP_NODE **)XtMalloc(l_max * sizeof(GROUP_NODE *));

    for (buffer = server_read(&global.server) ;
	 buffer && (buffer[0] != '.' || buffer[1] != '\0') ;
	 buffer = server_read(&global.server)) {
	char	mod;
	char	*c;

	if (n >= l_max) {
	    l_max += NEW_GROUPS;
	    list =
		(GROUP_NODE **)XtRealloc((char *)list,
					 l_max * sizeof(GROUP_NODE *));
	}
	if (!(c = strchr(buffer, ' ')))
	    continue;
	*(c++) = '\0';
	sscanf(c, "%*d%*d%*c%c", &mod);
	if (mod == '=' || mod == 'x')
	    continue;
	list[n] = find_group(buffer);
	if (!list[n])
	    list[n] = create_group(buffer);
	n++;
    }

    sort_groups();

    if (!buffer) {
	XtFree((char *)list);
	return False;
    }

    if (n <= 0)
	XtFree((char *)list);
    else {
	global.new_groups = list;
	global.no_new_groups = n;
    }

    return OK_NEWGROUPS;
}

static int create_new_newsrc(char *file_name)
{
    char	**auto_sub = NntpObjectAutoSubscribe(global.nntp_object);
    FILE	*fout;

    if (!auto_sub || !*auto_sub)
	return False;

    if (!(fout = fopen(file_name, "w")))
	return False;

    while (*auto_sub) {
	fprintf(fout, "%s:\n", *auto_sub);
	auto_sub++;
    }
    (void)fclose(fout);

    return True;
}

/* this is a weak point */
static void put_subscribed_groups_first(long n)
{
    GROUP_NODE	**subs;
    long	i, j;

    if (n == 0) return;

    subs = (GROUP_NODE **)XtMalloc(n * sizeof(GROUP_NODE *));
    for (i = j = global.no_groups - 1 ; i >= 0 ; i--) {
	if (global.groups[i]->subscribed)
	    subs[global.groups[i]->disp] = global.groups[i];
	else
	    global.groups[j--] = global.groups[i];
    }
    for (i = 0 ; i < n ; i++)
	global.groups[i] = subs[i];

    XtFree((char *)subs);
}

void sort_groups()
{
    long i, n;

    for (n = 0 ; n < global.no_groups ; n++)
	if (!global.groups[n]->subscribed) break;

    for (i = n ; i < global.no_groups ; i++)
	if (global.groups[i]->subscribed) {
	    GROUP_NODE *temp = global.groups[n];

	    global.groups[n] = global.groups[i];
	    global.groups[i] = temp;
	    n++;
	}

    if (n < global.no_groups)
	qsort(&global.groups[n], global.no_groups - n,
	      sizeof(GROUP_NODE *), comp_group_nodes);
}

static int get_char(FILE *in, FILE *out)
{
    int		c;

    c = getc(in);
    if (c != EOF && out)
	putc(c, out);

    return c;
}

static ART_LIST_NODE *parse_read_arts_list(FILE *fin, FILE *fout)
{
    ART_LIST_NODE	*result;
    int			c;
    long		first= 0, last = 0;

    do {
	c = get_char(fin, fout);
    } while (c != EOF && c != '\n' && IS_SPACE(c));
    if (c == EOF || c == '\n')
	return NULL;

    if (c < '0' || c > '9') {
	fprintf(stderr, "knews: Parse error in newsrc.\n");
	while (c != '\n' && c != EOF)
	    c = get_char(fin, fout);
	return NULL;
    }

    while (c >= '0' && c <= '9') {
	first *= 10;
	first += c - '0';
	c = get_char(fin, fout);
    }

    while (c != EOF && c != '\n' && IS_SPACE(c))
	c = get_char(fin, fout);

    if (c == '-') {
	c = get_char(fin, fout);
	if (c >= '0' || c <= '9') {
	    while (c >= '0' && c <= '9') {
		last *= 10;
		last += c - '0';
		c = get_char(fin, fout);
	    }
	} else {
	    c = 'b'; /* b for bogus */
	}
    } else {
	last = first;
    }

    while (c != EOF && c != '\n' && IS_SPACE(c))
	c = get_char(fin, fout);

    if (c != ',' && c != '\n') {
	fprintf(stderr, "knews: Parse error in newsrc.\n");
	while (c != EOF && c != '\n')
	    c = get_char(fin, fout);
	return NULL;
    }

    result = (ART_LIST_NODE *)XtMalloc(sizeof(ART_LIST_NODE));
    result->first = first;
    result->last = last;

    if (c == ',')
	result->next = parse_read_arts_list(fin, fout);
    else
	result->next = NULL;

    return result;
}

void check_read_arts_sanity(GROUP_NODE *temp)
{
    ART_LIST_NODE	*loop;

    for (loop = temp->read_arts ; loop ; loop = loop->next) {
	if (loop->first > loop->last ||
	    (loop->next && loop->last > loop->next->first)) {
	    fprintf(stderr,
		    "knews: Bad list of read articles in .newsrc file, "
		    "group %s.  Marking all articles unread.\n",
		    temp->name);
	    free_read_arts_list(temp);
	    temp->read_arts = NULL;
	    return;
	}
    }
}

/* there should be some sort of dynamic allocation here */
void parse_newsrc()
{
    FILE	*fin, *fout;
    char	buffer[MAX_GROUP_LEN + 1];
    char	*file_name;
    long	n = 0;
    int		c, i;
    struct stat	buf;

    file_name = NntpObjectNewsrcFile(global.nntp_object);
    if (!file_name)
	return;
    if (!(fin = fopen(file_name, "r")) &&
	(!create_new_newsrc(file_name) ||
	 !(fin = fopen(file_name, "r"))))
	return;

    file_name = NntpObjectOldNewsrcFile(global.nntp_object);
    global.last_time = 0;
    if (file_name) {
	if (stat(file_name, &buf) == 0 &&
	    S_ISREG(buf.st_mode)) global.last_time = buf.st_mtime;
	fout = fopen(file_name, "w");
    } else {
	fout = NULL;
    }

    c = get_char(fin, fout);
    while (c != EOF) {
	for (i = 0 ; i < MAX_GROUP_LEN ; i++) {
	    if (c == ':' || c == '!' ||
		IS_SPACE(c) || c == '\0' || c == EOF) break;
	    buffer[i] = c;
	    c = get_char(fin, fout);
	}
	buffer[i] = '\0';

	if (i < MAX_GROUP_LEN && (c == ':' || c == '!')) {
	    GROUP_NODE	*temp = find_group(buffer);

	    if (temp) {
		if (temp->found_in_newsrc)
		    fprintf(stderr, "knews: Group %s appeared "
			    "twice in newsrc file!\n",
			    buffer);
		else {		    
		    temp->subscribed = (c == ':');
		    temp->found_in_newsrc = True;
		    if (temp->subscribed)
			temp->disp = n++;
		    temp->read_arts = parse_read_arts_list(fin, fout);
		    c = '\n';
		}
	    } else {
		fprintf(stderr,
			"knews: Bogus group in newsrc: %s\n",
			buffer);
	    }
	} else {
	    fprintf(stderr,
		    "knews: Bogus entry in newsrc, skipping: \n  %s",
		    buffer);
	}

	while (c != EOF && c != '\n')
	    c = get_char(fin, fout);
	if (c != EOF)
	    c = get_char(fin, fout);
    }

    (void)fclose(fin);
    (void)fclose(fout);

    put_subscribed_groups_first(n);
}

static GROUP_NODE *add_group(char *name)
{
    long	i, n = global.no_groups;

    for (i = 0 ; i < global.no_groups ; i++)
	if (strcmp(global.groups[i]->name, name) == 0)
	    return global.groups[i];

    global.no_groups++;
    if (i > global.max_groups - 3) {
	i = n;
	global.max_groups++;
	global.max_groups *= 2;
	global.groups =
	    (GROUP_NODE **)XtRealloc((char *)global.groups,
				     global.max_groups *
				     sizeof(GROUP_NODE *));
	while (i < global.max_groups)
	    global.groups[i++] = NULL;
    }

    global.groups[n] = (GROUP_NODE *)XtMalloc(sizeof(GROUP_NODE));
    global.groups[n]->name = XtNewString(name);
    global.groups[n]->description = NULL;
    global.groups[n]->ahead = NULL;
    global.groups[n]->no_unread = 0;
    global.groups[n]->first_art = 0;
    global.groups[n]->last_art = 0;
    global.groups[n]->read_arts = NULL;
    global.groups[n]->disp = -1;
    global.groups[n]->subscribed = False;
    global.groups[n]->moderated = False;
    global.groups[n]->found_in_newsrc = False;

    return global.groups[n];
}

int get_newsgroups_from_newsrc()
{
    FILE	*fin, *fout;
    char	buffer[MAX_GROUP_LEN + 4];
    char	*file_name;
    long	n = 0;
    int		c, i;
    struct stat	buf;

    file_name = NntpObjectNewsrcFile(global.nntp_object);
    if (!file_name)
	return 0;
    if (!(fin = fopen(file_name, "r")) &&
	(!create_new_newsrc(file_name) ||
	 !(fin = fopen(file_name, "r"))))
	return 0;

    file_name = NntpObjectOldNewsrcFile(global.nntp_object);
    global.last_time = 0;
    if (file_name) {
	if (stat(file_name, &buf) == 0 &&
	    S_ISREG(buf.st_mode))
	    global.last_time = buf.st_mtime;
	fout = fopen(file_name, "w");
    } else {
	fout = NULL;
    }

    strcpy(buffer, "GROUP ");
    c = get_char(fin, fout);
    while (c != EOF) {
	for (i = 6 ; i < MAX_GROUP_LEN ; i++) {
	    if (c == ':' || c == '!' ||
		IS_SPACE(c) || c == '\0' || c == EOF) break;
	    buffer[i] = c;
	    c = get_char(fin, fout);
	}
	buffer[i] = '\0';

	if (i >= MAX_GROUP_LEN || (c != ':' && c != '!'))
	    fprintf(stderr,
		    "knews: Bogus entry in newsrc, skipping: \n  %s",
		    buffer + 6);
	else if (c == ':') {
	    char	*reply;
	    long	first, last;

	    buffer[i] = '\r';
	    buffer[i+1] = '\n';
	    buffer[i+2] = '\0';
	    server_write(&global.server, buffer);
	    reply = server_read(&global.server);
	    buffer[i] = '\0';

	    if (!reply) {
		(void)fclose(fin);
		(void)fclose(fout);
		return -1;
	    }

	    if (atoi(reply) == OK_GROUP &&
		sscanf(reply, "%*d%*d%ld%ld", &first, &last) == 2) {
		GROUP_NODE	*temp = add_group(buffer + 6);

		if (temp->found_in_newsrc)
		    fprintf(stderr,
			    "knews: Group %s appeared twice "
			    "in newsrc file!\n",
			    buffer + 6);
		else {
		    temp->first_art = first;
		    temp->last_art = last;
		    temp->subscribed = True;
		    temp->found_in_newsrc = True;
		    temp->disp = n++;
		    temp->read_arts = parse_read_arts_list(fin, fout);
		    c = '\n'; /* hack */
		}
	    } else {
		fprintf(stderr,
			"knews: Bogus group in newsrc: %s\n"
			"       Failed message from server is: %s\n",
			buffer + 6, reply);
	    }
	} else { /* c == '!' */
	    GROUP_NODE	*temp = add_group(buffer + 6);

	    if (temp->found_in_newsrc)
		fprintf(stderr,
			"knews: Group %s appeared twice "
			"in newsrc file!\n",
			buffer + 6);
	    else {
		temp->found_in_newsrc = True;
		temp->read_arts = parse_read_arts_list(fin, fout);
		c = '\n'; /* hack */
	    }
	}

	while (c != EOF && c != '\n')
	    c = get_char(fin, fout);
	if (c != EOF)
	    c = get_char(fin, fout);
    }

    (void)fclose(fin);
    (void)fclose(fout);
    sort_groups();

    return OK_GROUPS;
}

int update_newsrc()
{
    char		*newsrc_file;
    FILE		*fout;
    long		i;
    ART_LIST_NODE	*loop;
    int			fill = NntpObjectFillNewsrcFile(global.nntp_object);

    newsrc_file = NntpObjectNewsrcFile(global.nntp_object);
    if (!newsrc_file || !(fout = fopen(newsrc_file, "w")))
	return False;

    for (i = 0 ; i < global.no_groups ; i++) {
	if (fill || global.groups[i]->read_arts ||
	    global.groups[i]->subscribed) {
	    if (fprintf(fout, "%s%c", global.groups[i]->name,
			global.groups[i]->subscribed ? ':' : '!') == EOF)
		return False;
	    loop = global.groups[i]->read_arts;
	    if (loop && fprintf(fout, " ") == EOF)
		return False;
	    while (loop) {
		if (fprintf(fout, "%ld", loop->first) == EOF)
		    return False;
		if (loop->first != loop->last &&
		    fprintf(fout, "-%ld", loop->last) == EOF)
		    return False;
		loop = loop->next;
		if (loop && fprintf(fout, ",") == EOF)
		    return False;
	    }
	    if (fprintf(fout, "\n") == EOF)
		return False;
	}
    }
    if (fclose(fout) == EOF)
	return False;

    return True;
}

void calc_no_unread(GROUP_NODE *group)
{
    ART_LIST_NODE *loop;

    if (group->first_art == 0 && group->last_art == 0) {
	group->no_unread = 0;
	return;
    }

    group->no_unread = group->last_art - group->first_art + 1;
    for (loop = group->read_arts ; loop ; loop = loop->next) {
	if (loop->last > group->last_art) {
	    if (loop->first <= group->last_art)
		group->no_unread -= (group->last_art - loop->first + 1);
	    return;
	} else if (loop->first < group->first_art) {
	    if (loop->last >= group->first_art)
		group->no_unread -= (loop->last - group->first_art +1);
	} else {
	    group->no_unread -= (loop->last - loop->first + 1);
	}
    }
}

char *rescan()
{
    long	n;
    char	*p;
    char	message[128];

    for (n = 0 ; n < global.no_groups ; n++)
	if (!global.groups[n]->subscribed)
	    break;
    if (n == 0)
	return CODE_TO_STR(OK_GROUPS);

    sprintf(message, "Rescan in progress...   ");
    p = message + strlen(message);

    if (NntpObjectReadActiveFile(global.nntp_object)) {
	char	*buffer;
	long	i, j, k, m;

	buffer = server_comm(&global.server, "LIST\r\n", True);
	if (!buffer || atoi(buffer) != OK_GROUPS)
	    return buffer;

	buffer = server_read(&global.server);
	j = m = global.no_groups / 16;
	i = 0;
	while (buffer && (buffer[0] != '.' || buffer[1] != '\0')) {
	    char	*c = strchr(buffer, ' ');

	    if (c) {
		*c++ = '\0';
		for (k = 0 ; k < n ; k++) {
		    if (strcmp(buffer, global.groups[k]->name) == 0) {
			sscanf(c, "%ld%ld",
			       &global.groups[k]->last_art,
			       &global.groups[k]->first_art);
			break;
		    }
		}
	    }

	    if (j-- <= 0) {
		sprintf(p, "%3ld%%", i / global.no_groups);
		set_message(message, False);
		j = m;
	    }
	
	    buffer = server_read(&global.server);
	    i += 100;
	}

	if (!buffer)
	    return NULL;
    } else {
	char	command[512];
	char	*buffer;
	long	i, j;

	for (i = 0, j = 100 ; i < n ; i++, j += 100) {
	    if (strlen(global.groups[i]->name) < 480) {
		sprintf(command, "GROUP %s\r\n", global.groups[i]->name);
		server_write(&global.server, command);
		buffer = server_read(&global.server);

		if (!buffer)
		    return NULL;
		if (atoi(buffer) == OK_GROUP) {
		    sscanf(buffer, "%*d%*d%ld%ld",
			   &global.groups[i]->first_art,
			   &global.groups[i]->last_art);
		}
	    }

	    sprintf(p, "%3ld%%", j / n);
	    set_message(message, False);
	}
    }

    return CODE_TO_STR(OK_GROUPS);
}

int get_descriptions()
{
    char	*buffer;
    long	n;

    buffer = server_comm(&global.server, "LIST NEWSGROUPS\r\n", False);
    if (!buffer)
	return -1;
    sscanf(buffer, "%ld", &n);
    if (n != OK_GROUPS)
	return n;

    for (buffer = server_read(&global.server) ;
	 buffer && (buffer[0] != '.' || buffer[1] != '\0') ;
	 buffer = server_read(&global.server)) {
	GROUP_NODE	*group;
	char		*c;

	c = buffer;
	while (*c != '\0' && *c != '\t' && *c != ' ') c++;
	if (*c == '\0')
	    continue;

	*c++ = '\0';
	while (*c == ' ' || *c == '\t') c++;
	group = find_group(buffer);
	if (group) {
	    if (group->description) {
		long	len = strlen(group->description);

		/* if duplicate descriptions, take the longest */
		if (len >= strlen(c)) continue;
		XtFree(group->description);
	    }

	    group->description = XtNewString(c);
	}
    }

    if (!buffer)
	return -1;

    return OK_GROUPS;
}

GROUP_NODE *create_group(char *name)
{
    GROUP_NODE	*group;

    global.groups =
	(GROUP_NODE **)XtRealloc((char *)global.groups,
				 (global.no_groups + 1) *
				 sizeof(GROUP_NODE *));
    group = global.groups[global.no_groups++] =
	(GROUP_NODE *)XtMalloc(sizeof(GROUP_NODE));
    group->name = XtNewString(name);
    group->description = NULL;
    group->ahead = NULL;
    group->no_unread = 0;
    group->first_art = 0;
    group->last_art = 0;
    group->read_arts = NULL;
    group->disp = -1;
    group->subscribed = False;
    group->moderated = False;
    group->found_in_newsrc = False;

    return group;
}
