/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "../configure.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "parse.h"
#include "pedit.h"
#include "post.h"
#include "server.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtText.h"
#include "../Widgets/Dialogue.h"
#include "../Widgets/Menu.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/SeparatorG.h"
#include "../Widgets/StringG.h"
#include "../Widgets/ToggleG.h"
#include "../Widgets/PopdownSh.h"

static Widget		include_toggle;

static void fork_editor(PostContext*);

static int create_temp_file(FILE **fp, char **file_name)
{
    *fp = NULL;
    *file_name = tmpnam(NULL);

    if (*file_name) {
	*fp = fopen(*file_name, "w");
	if (*fp) {
	    *file_name = XtNewString(*file_name);
	    return 0;
	}
	*file_name = NULL;
    }

    return -1;
}

static void free_post_context(PostContext *context)
{
    if (context->file_name) {
	unlink(context->file_name);
	XtFree(context->file_name);
    }
    if (context->fp)
	fclose(context->fp);
    XtFree(context->newsgroups);
    XtFree(context->reply_to);
    XtFree((char *)context);
}

static void post_notice_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    NoticeReply	reply = (NoticeReply)call_data;
    PostContext	*context = (PostContext *)client_data;
    char	buffer[512];

    if (global.busy) return;

    switch (reply) {
    case NoticeReplyTimeout:
	return;
    case NoticeReplyLeft:
	if (!(context->flags & OK_TO_POST)) {
	    XBell(XtDisplay(w), 0);
	    return;
	} else if (global.mode == NewsModeDisconnected) {
	    set_message("Not connected!", True);
	    return;
	} else {
	    int		do_post, do_mail;
	    PostStatus	post_status = PostStatusOk;
	    MailStatus	mail_status = MailStatusOk;
	    char	message[STDERR_BUFFLEN], *c;
	    char	*agent = NntpObjectPostingAgent(global.nntp_object);

	    do_post = (context->flags & POST) && !(context->flags & POST_DONE);
	    do_mail = (context->flags & MAIL) && !(context->flags & MAIL_DONE);

	    if (!do_post && !do_mail) {
		XBell(XtDisplay(w), 0);
		return;
	    }

	    if (do_post) {
		message[0] = '\0';
		set_message("Posting article...", False);
		if (agent)
		    post_status = post_to_agent(context, agent, message);
		else
		    post_status = post_article(context, message);
	    }

	    if (post_status == PostStatusOk && do_mail) {
		message[0] = '\0';
		set_message("Mailing article...", False);
		mail_status = mail_article(context, message);
	    }

	    if (post_status == PostStatusOk && mail_status == MailStatusOk) {
		if (do_post) {
		    if (do_mail)
			set_message("Post and mail were successful.", False);
		    else
			set_message("Post was successful.", False);
		} else
		    set_message("Mail was successful.", False);
		free_post_context(context);
	    } else if (post_status == PostStatusOk) {
		/* post successful, but not mail */
		NoticeSetLeftLabel(w, "Mail");

		switch (mail_status) {
		case MailStatusOk: /* notreached */
		    break;
		case MailStatusFailed:
		    if (do_post)
			sprintf(buffer,
				"Post was successful, but email failed.\n"
				MAIL_COMMAND " didn't accept it.");
		    else
			sprintf(buffer,
				"Email failed.\n"
				MAIL_COMMAND " didn't accept it.");
		    stderr_popup(message, 0);
		    break;
		case MailStatusStart:
		    if (do_post)
			sprintf(buffer,
				"Post was successful, but email failed.\n"
				"Failed to start " MAIL_COMMAND);
		    else
			sprintf(buffer,
				"Email failed.\n"
				"Failed to start " MAIL_COMMAND);
		    stderr_popup(message, 0);
		    break;
		case MailStatusOpen:
		    if (do_post)
			sprintf(buffer,
				"Post was successful, but email failed.\n"
				"Failed to open temp file.");
		    else
			sprintf(buffer,
				"Email failed.\n"
				"Failed to open temp tile.");
		    break;
		case MailStatusFork:
		    if (do_post)
			sprintf(buffer,
				"Post was successful, but email failed.\n"
				"Fork failed.");
		    else
			sprintf(buffer, "Fork failed.");
		    break;
		}
		NoticeSetMessage(w, buffer);
		c = strchr(buffer, '\n');
		if (c) *c = '\0';
		set_message(buffer, True);
		return;
	    } else { /* post unsuccessful, mail not tried */
		switch (post_status) {
		case PostStatusOk: /* notreached */
		    break;
		case PostStatusBroken:
		    sprintf(buffer, "Posting (probably) not completed");
		    break;
		case PostStatusOpen:
		    sprintf(buffer, "Failed to open temp file!");
		    break;
		case PostStatusFailed:
		    if (agent) {
			sprintf(buffer,
				"Posting failed, %s didn't accept it.",
				agent);
			stderr_popup(message, 0);
		    } else {
			if (strlen(message) > 128) message[128] = '\0';
			sprintf(buffer,
				"Posting failed, message "
				"from server is:\n\n%s",
				message);
		    }
		    break;
		case PostStatusStart:
		    if (agent) {
			sprintf(buffer,
				"Posting failed: couldn't start %s.",
				agent);
			stderr_popup(message, 0);
		    } else /* not reached */
			sprintf(buffer, "Unknown error!");
		    break;
		}
		NoticeSetMessage(w, buffer);
		c = strchr(buffer, '\n');
		if (c) *c = '\0';
		set_message(buffer, True);
		return;
	    }
	}
	break;
    case NoticeReplyMiddle:
	fork_editor(context);
	break;
    case NoticeReplyRight:
    case NoticeReplyClose:
	free_post_context(context);
	set_standard_message();
	break;
    }

    XtPopdown(w);
    XtDestroyWidget(w);
}

static void post_context_callback(void *client_data,
				  int status, char *stderr_buf)
{
    PostContext	*context = (PostContext *)client_data;
    char	message[STDERR_BUFFLEN];
    char	*left_label = "Post";
    int		ok = False;
    Widget	notice;

    if (WIFEXITED(status)) {
	status = WEXITSTATUS(status);
	if (status == 0)
	    ok = check_article_to_post(context, message);
	else if (status == 127)
	    strcpy(message, "Failed to start editor!");
	else
	    strcpy(message, "Editor exited abnormally!");
    } else if (WIFSIGNALED(status)) {
	status = WTERMSIG(status);
	sprintf(message, "Editor caught %s!",
		signal_string(status));
    } else
	strcpy(message, "Unknown problem with editor!");

    if (ok) {
	if ((context->flags & POST) && !(context->flags & POST_DONE)) {
	    if ((context->flags & MAIL) &&
		!(context->flags & MAIL_DONE))
		left_label = "Post&Mail";
	} else if ((context->flags & MAIL) &&
		   !(context->flags & MAIL_DONE))
	    left_label = "Mail";
    } else {
	stderr_popup(stderr_buf, 0);
    }

    notice = NoticeCreateAndPopup("postnotice", main_widgets.shell, message,
				  left_label ? left_label : "Post",
				  "Edit", "Abort", 0,
				  post_notice_callback, (XtPointer)context,
				  XtGrabNone);
    if (!ok) {
	Arg	arg;
	Widget	knapp;

	XtSetArg(arg, XtNleftKnapp, &knapp);
	XtGetValues(notice, &arg, 1);
	XtSetSensitive(knapp, False);
	XBell(XtDisplay(main_widgets.shell), 0);
    }
}

static void fork_editor(PostContext *context)
{
    pid_t	pid;

    pid = fork_nicely(context, post_context_callback, True);

    if (pid < 0) { /* fork failed */
	set_message("Failed to start editor!", True);
	free_post_context(context);
    } else if (pid == 0) { /* child */
	char	editcmd[512];
	char	*p1, *p2;

	p1 = strstr(global.edit_command, "%s");
	p2 = strstr(global.edit_command, "%i");

	sprintf(editcmd, "exec ");
	if (p1) {
	    if (p2) {
		if (p1 < p2)
		    sprintf(editcmd + 5, global.edit_command,
			    context->file_name, context->line);
		else
		    sprintf(editcmd + 5, global.edit_command,
			    context->line, context->file_name);
	    } else
		sprintf(editcmd + 5, global.edit_command, context->file_name);
	} else {
	    if (p2)
		sprintf(editcmd + 5, global.edit_command, context->line);
	    else
		sprintf(editcmd + 5, global.edit_command);
	}
	
	execl("/bin/sh", "sh", "-c", editcmd, (char *)NULL);
	perror("knews: execl");
	_exit(127);
    } else { /* parent */
	if (global.mode == NewsModeDisconnected)
	    set_message("Editor started.  Warning: not connected.", False);
	else if (global.posting_allowed ||
		 !(context->flags & POST))
	    set_message("Editor started.", False);
	else
	    set_message("Editor started.  Warning: this server "
			"doesn't allow posting.", False);
    }
}

static int insert_extra_headers(FILE *fp)
{
    char	*reply_to = GroupObjectReplyTo(global.group_object);
    char	*organization = GroupObjectOrganization(global.group_object);
    char	*extra_headers = GroupObjectExtraHeaders(global.group_object);
    int		result = 0;

    fprintf(fp, "%s\n", "X-Newsreader: knews " KNEWS_VERSION);
    result++;

    if (reply_to) {
	fprintf(fp, "Reply-To: %s\n", reply_to);
	result++;
    }

    if (organization) {
	fprintf(fp, "Organization: %s\n", organization);
	result++;
    }

    if (extra_headers && extra_headers[0] != '\0') {
	char	*c;

	fprintf(fp, "%s", extra_headers);
	for (c = extra_headers ; ; c++) {
	    if (*c == '\n')
		result++;
	    else if (*c == '\0') {
		if (c[-1] != '\n') {
		    fputc('\n', fp);
		    result++;
		}
		break;
	    }
	}
    }

    return result;
}

static void append_signature(PostContext *context)
{
    char	*sig_file_name = GroupObjectSignatureFile(global.group_object);
    FILE	*sig;

    if (sig_file_name &&
	(sig = fopen(sig_file_name, "r"))) {
	int	c;

	fprintf(context->fp, "\n-- \n");
	while ((c = getc(sig)) != EOF)
	    (void)putc(c, context->fp);
	(void)fclose(sig);
    }
}

#define MAX_REFS	8
static int print_references_header(FILE *fp, ARTICLE *art)
{
    ARTICLE	*arts[MAX_REFS + 2];
    int		i = MAX_REFS + 1, col, rows;

    do {
	arts[i--] = art;
	art = A_PARENT(art);
    } while (i > 0 && art);

    if (art) {
	while (A_PARENT(art))
	    art = A_PARENT(art);
	arts[i] = art;
    } else {
	i++;
    }

    rows = 1;
    col = 11 + fprintf(fp, " <%s>", arts[i++]->msgid);
    while (i < MAX_REFS + 2) {
	if (col + arts[i]->hash_len > 75) {
	    rows++;
	    fprintf(fp, "\n ");
	    col = 1;
	}
	col += fprintf(fp, " <%s>", arts[i++]->msgid);
    }
    fprintf(fp, "\n");

    return rows;
}

static void print_attribution(FILE *fp, char *attr, ARTICLE *art)
{
    if (fp && attr && attr[0] != '\0' && art) {
	while (*attr != '\0') {
	    if (*attr == '%') {
		attr++;
		switch (*attr) {
		case 'f':
		case 'F':
		    fprintf(fp, "%s", art->from);
		    break;
		case 'M':
		case 'm':
		    fprintf(fp, "<%s>", art->msgid);
		    break;
		case '%':
		    fputc('%', fp);
		    break;
		default:
		    break;
		}
	    } else {
		fputc(*attr, fp);
	    }
	    attr++;
	}
	fputc('\n', fp);
    }
}

static int include_quotation(long art_no, FILE *fp,
			     char *quote_string, Boolean head)
{
    char	command[32];
    char	*buffer;
    int		status;

    if (!quote_string) quote_string = "";

    sprintf(command, "%s %ld\r\n", head ? "ARTICLE" : "BODY", art_no);
    buffer = server_comm(&global.server, command, True);

    if (!buffer)
	return -1;
    status = atoi(buffer);
    if (status != OK_ARTICLE && status != OK_BODY)
	return 0; /* fail silently */

    buffer = server_read(&global.server);
    while (buffer && (buffer[0] != '.' || buffer[1] != '\0')) {
	(void)fprintf(fp, "%s%s\n", quote_string, buffer);
	buffer = server_read(&global.server);
    }

    if (!buffer)
	return -1;

    return 0;
}

static void setup_file_and_fork_editor(PostContext *context)
{
    FILE	*fp = context->fp;
    int		line = 1;
    char	*full_name = GroupObjectFullName(global.group_object);

    line += insert_extra_headers(context->fp);

    if (context->art) { /* a followup */
	fprintf(fp, "References:");
	line += print_references_header(fp, context->art);

	fprintf(fp, "From: %s@%s (%s)\n",
		global.mail_name, global.domain_name,
		full_name ? full_name : "");
	line++;

	fprintf(fp, "Subject: Re: %s\n", context->art->subject->subject);
	line++;

	if (context->flags & POST) {
	    fprintf(fp, "Newsgroups: %s\n",
		    context->newsgroups ? context->newsgroups : "");
	    line++;
	} else if (context->newsgroups) {
	    fprintf(fp, "X-Original-Newsgroups: %s\n", context->newsgroups);
	    line++;
	}

	if (context->flags & MAIL) {
	    if (context->art) {
		fprintf(fp, "In-Reply-To: <%s>\n", context->art->msgid);
		line++;
	    }

	    fprintf(fp, "To: %s\n",
		    context->reply_to ?
		    context->reply_to : context->art->from);
	    line++;		    
	}

	line++;
    } else {
	fprintf(fp, "From: %s@%s (%s)\n",
		global.mail_name, global.domain_name,
		full_name ? full_name : "");
	line++;

	fprintf(fp, "Subject: \n");

	fprintf(fp, "Newsgroups: ");
	if (global.mode == NewsModeGroup || global.mode == NewsModeThread)
	    fprintf(fp, "%s\n", global.thr.group->name);
	else
	    fputc('\n', fp);
    }

    fputc('\n', fp);
    context->line = line;

    if (ToggleGadgetGet(include_toggle) && context->art) {
	char	*attribution = GroupObjectAttribution(global.group_object);
	char	*quote_string = GroupObjectQuoteString(global.group_object);

	print_attribution(context->fp, attribution, context->art);
	if (include_quotation(context->art->no, context->fp,
			      quote_string, False) < 0) {
	    free_post_context(context);
	    set_busy(True, True);
	    (void)reconnect_server(True);
	    unset_busy();
	    return;
	}
    }

    append_signature(context);

    fclose(fp);
    context->fp = NULL;

    fork_editor(context);
}

static void forward_dialogue_callback(Widget w,
				      XtPointer client_data,
				      XtPointer call_data)
{
    PostContext		*context = (PostContext *)client_data;
    DialogueReport	*report = (DialogueReport *)call_data;
    char		*full_name = GroupObjectFullName(global.group_object);
    int			line = 0;

    if (!report)
	return;

    switch (report->reply) {
    case DialogueReplyClose:
    case DialogueReplyRight:
	XtPopdown(w);
	XtDestroyWidget(w);
	free_post_context(context);
	return;
    case DialogueReplyTab:
	return;
    case DialogueReplyMiddle:
	break;
    case DialogueReplyLeft:
    case DialogueReplyEnter:
	if (!report->buffer)
	    return;
	break;
    }

    XtPopdown(w);
    XtDestroyWidget(w);

    line += insert_extra_headers(context->fp);

    if (full_name)
	fprintf(context->fp, "From: %s@%s (%s)\n",
		global.mail_name, global.domain_name, full_name);
    else
	fprintf(context->fp, "From: %s@%s\n",
		global.mail_name, global.domain_name);
    line++;

    fprintf(context->fp, "Subject: %s%s (fwd)\n",
	    PARENT(context->art) ? "Re: " : "",
	    context->art->subject->subject);
    line++;

    fprintf(context->fp,
	    "To: %s\n\n----- Forwarded Message -----\n",
	    report->buffer);
    line++;

    if (include_quotation(context->art->no,
			  context->fp, NULL, True) < 0) {
	(void)reconnect_server(True);
	free_post_context(context);
	return;
    }
    context->art = NULL;

    (void)fclose(context->fp);
    context->fp = NULL;

    if (report->reply == DialogueReplyMiddle)
	fork_editor(context);
    else {
	char		message[STDERR_BUFFLEN];
	MailStatus	status;

	set_message("Mailing...", False);
	message[0] = '\0';
	status = mail_article(context, message);

	switch (status) {
	case MailStatusOk:
	    set_message("Mailing done.", False);
	    break;
	case MailStatusFailed:
	    set_message("Mail failed: " MAIL_COMMAND
			" didn't accept it!", True);
	    stderr_popup(message, 0);
	case MailStatusStart:
	    set_message("Failed to start " MAIL_COMMAND, True);
	    stderr_popup(message, 0);
	    break;
	case MailStatusOpen:
	    set_message("Failed to open temp file!", True);
	    break;
	case MailStatusFork:
	    set_message("Fork failed!", True);
	    break;

	}

	free_post_context(context);
    }
}

static void followup_to_poster_callback(Widget w,
					XtPointer client_data,
					XtPointer call_data)
{
    PostContext	*context = (PostContext *)client_data;
    NoticeReply	reply = (NoticeReply)call_data;

    switch (reply) {
    case NoticeReplyMiddle:	/* Email */
	context->flags = MAIL | ORIG_MAIL;
	/* fall through */
    case NoticeReplyLeft:	/* Post anyway */
	setup_file_and_fork_editor(context);
	break;
    default:
	free_post_context(context);
	break;
    }

    XtPopdown(w);
    XtDestroyWidget(w);
}

static void post_menu_callback(Widget gw,
			       XtPointer client_data,
			       XtPointer call_data)
{
    PostContext	*context;
    int		flags = (long)client_data;
    char	command[512];
    char	*buffer;
    ARTICLE	*art = NULL;
    char	*reply_to = NULL;
    char	*newsgroups = NULL;
    char	*followup_to = NULL;

    if (!global.domain_name) {
	set_message("Domain name not available, "
		    "posting will not be allowed!",
		    True);
	return;
    } else if (!global.user_id) {
	set_message("User id not available, "
		    "posting will not be allowed!",
		    True);
	return;
    }

    if (global.mode == NewsModeGroup || global.mode == NewsModeThread) {
	if (!global.curr_art || !global.curr_art->from) {
	    set_message("No selected article!", True);
	    return;
	}
	art = global.curr_art;
    } else {
	set_message("Not in a newsgroup!", True);
	return;
    }

    set_busy(True, True);
    sprintf(command, "HEAD %ld\r\n", art->no);
    buffer = server_comm(&global.server, command, True);
    if (!buffer) {
	reconnect_server(True);
	unset_busy();
	return;
    } else if (atoi(buffer) != OK_HEAD) {
	unset_busy();
	set_message("Failed to retrieve article!", True);
	return;
    }

    buffer = server_read(&global.server);
    while (buffer && (buffer[0] != '.' || buffer[1] != '\0')) {
	if (!newsgroups &&
	    strncasecmp(buffer, "Newsgroups: ", 12) == 0)
	    newsgroups = XtNewString(buffer + 12);
	else if (!followup_to &&
		 strncasecmp(buffer, "Followup-To: ", 13) == 0)
	    followup_to = XtNewString(buffer + 13);
	else if (!reply_to &&
		 strncasecmp(buffer, "Reply-To: ", 10) == 0)
	    reply_to = XtNewString(buffer + 10);

	buffer = server_read(&global.server);
    }
    if (!buffer) {
	XtFree(newsgroups);
	XtFree(followup_to);
	XtFree(reply_to);
	reconnect_server(True);
	unset_busy();
	return;
    }
    unset_busy();

    context = (PostContext *)XtMalloc(sizeof(PostContext));
    context->reply_to = reply_to;
    context->art = art;
    context->flags = flags;
    context->line = 0;
    if (create_temp_file(&context->fp, &context->file_name) < 0) {
	set_message("Failed to create temp file!", True);
	XtFree((char *)context);
	XtFree(newsgroups);
	XtFree(followup_to);
	XtFree(reply_to);
	return;
    }

    if (followup_to) {
	if ((flags & POST) && strcmp(followup_to, "poster") == 0) {
	    XtFree(followup_to);
	    NoticeCreateAndPopup("notice", main_widgets.shell,
				 "The original author has requested\n"
				 "that followups be directed to email.\n\n"
				 "Post, email or abort?",
				 "Post", "Email", "Abort", 0,
				 followup_to_poster_callback,
				 (XtPointer)context,
				 XtGrabExclusive);
	    context->newsgroups = newsgroups;
	    XtFree(followup_to);
	    return;
	}
	context->newsgroups = followup_to;
	XtFree(newsgroups);
    } else {
	context->newsgroups = newsgroups;
    }
    if (!(context->flags & POST))
	context->flags |= ORIG_MAIL;

    setup_file_and_fork_editor(context);
}

static void post_new_callback(Widget w,
			      XtPointer client_data,
			      XtPointer call_data)
{
    PostContext *context;
    char        *file_name;
    FILE        *fp;

    if (!global.domain_name) {
        set_message("Domain name not available, "
		    "posting will not be allowed!",
		    True);
        return;
    } else if (!global.user_id) {
        set_message("User id not available, "
		    "posting will not be allowed!",
		    True);
        return;
    } else if (create_temp_file(&fp, &file_name) < 0) {
        set_message("Failed to create temporary file!", True);
        return;
    }

    context = (PostContext *)XtMalloc(sizeof(PostContext));
    context->file_name = file_name;
    context->fp = fp;
    context->line = 0;
    context->newsgroups = NULL;
    context->reply_to = NULL;
    context->art = NULL;
    context->flags = POST;

    setup_file_and_fork_editor(context);
}

static void forward_menu_callback(Widget gw,
				  XtPointer client_data,
				  XtPointer call_data)
{
    PostContext	*context;
    Widget	w;
    Arg		args[6];
    Position	x, y;
    FILE	*fp;
    char	*file_name;

    if (!global.domain_name) {
	set_message("Domain name not available, "
		    "posting will not be allowed!",
		    True);
	return;
    } else if (!global.user_id) {
	set_message("User id not available, "
		    "posting will not be allowed!",
		    True);
	return;
    } else if (global.mode != NewsModeGroup &&
	       global.mode != NewsModeThread) {
	set_message("Not in a newsgroup!", True);
	return;
    } else if (!global.curr_art || !global.curr_art->from) {
	set_message("No selected article!", True);
	return;
    } else if (create_temp_file(&fp, &file_name) < 0) {
        set_message("Failed to create temporary file!", True);
        return;
    }

    context = (PostContext *)XtMalloc(sizeof(PostContext));
    context->file_name = file_name;
    context->fp = fp;
    context->line = 0;
    context->newsgroups = NULL;
    context->reply_to = NULL;
    context->art = global.curr_art;
    context->flags = MAIL|ORIG_MAIL;

    XtTranslateCoords(gw, -10, -10, &x, &y);
    XtSetArg(args[0], XtNleftLabel, "Mail");
    XtSetArg(args[1], XtNmiddleLabel, "Edit");
    XtSetArg(args[2], XtNrightLabel, "Cancel");
    XtSetArg(args[3], XtNmessage, "Forward article to");
    XtSetArg(args[4], XtNx, x);
    XtSetArg(args[5], XtNy, y);
    w = XtCreatePopupShell("forward", dialogueWidgetClass,
			   main_widgets.shell, args, 6);
    XtAddCallback(w, XtNcallback,
		  forward_dialogue_callback,(XtPointer)context);
    XtPopup(w, XtGrabExclusive);
}

static void cancel_post_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    PostContext	*context;
    FILE	*fp;
    char	*file_name;
    char	*full_name = GroupObjectFullName(global.group_object);
    int		line = 1;

    if (!global.domain_name) {
	set_message("Domain name not available, "
		    "posting will not be allowed!",
		    True);
	return;
    } else if (!global.user_id) {
	set_message("User id not available, "
		    "posting will not be allowed!",
		    True);
	return;
    } else if (global.mode != NewsModeGroup &&
	       global.mode != NewsModeThread) {
	set_message("Not in a newsgroup!", True);
	return;
    } else if (!global.curr_art) {
	set_message("No selected article!", True);
	return;
    } else if (create_temp_file(&fp, &file_name) < 0) {
	set_message("Failed to create temporary file!", True);
	return;
    }

    context = (PostContext *)XtMalloc(sizeof(PostContext));
    context->art = global.curr_art;
    context->file_name = file_name;
    context->fp = NULL;
    context->newsgroups = NULL;
    context->reply_to = NULL;
    context->flags = POST;

    line += insert_extra_headers(fp);

    fprintf(fp, "From: %s@%s (%s)\n",
	    global.mail_name, global.domain_name,
	    full_name ? full_name : "");
    line++;

    fprintf(fp, "Subject: cmsg cancel <%s>\n", context->art->msgid);
    line++;

    fprintf(fp, "Control: cancel <%s>\n", context->art->msgid);
    line++;

    fprintf(fp, "Newsgroups: %s\n", global.thr.group->name);
    line++;

    fputc('\n', fp);
    line++;

    fprintf(fp, "Article canceled from within knews.\n");

    (void)fclose(fp);

    context->line = line;
    fork_editor(context);
}

/*************************************************************************/

void create_post_menu(Widget main_shell)
{
    Widget	menu, temp;
    Arg		args[2];

    main_widgets.post_menu =
	XtCreatePopupShell("postshell", popdownShellWidgetClass,
			   main_shell, NULL, 0);
    menu = XtCreateManagedWidget("postmenu", menuWidgetClass,
				 main_widgets.post_menu, NULL, 0);

    XtSetArg(args[0], XtNlabel, "Post a followup");
    temp = MenuCreateGadget("followup", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback, post_menu_callback, (XtPointer)POST);

    XtSetArg(args[0], XtNlabel, "Reply by mail");
    temp = MenuCreateGadget("mailreply", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback, post_menu_callback, (XtPointer)MAIL);

    XtSetArg(args[0], XtNlabel, "Followup and reply");
    temp = MenuCreateGadget("followupreply", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback,
		  post_menu_callback, (XtPointer)(POST | MAIL));

    XtSetArg(args[0], XtNlabel, "Post a new article");
    temp = MenuCreateGadget("postnew", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback, post_new_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Cancel article");
    temp = MenuCreateGadget("cancel", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback, cancel_post_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Forward by mail");
    temp = MenuCreateGadget("forward", stringGadgetClass, menu, args, 1);
    XtAddCallback(temp, XtNcallback, forward_menu_callback, NULL);

    (void)MenuCreateGadget("separator", separatorGadgetClass, menu, NULL, 0);

    XtSetArg(args[0], XtNlabel, "Include quoted text");
    include_toggle =
	MenuCreateGadget("quotetoggle", toggleGadgetClass, menu, args, 1);
}

static void do_incl(String *params, Cardinal *no_params)
{
    if (*no_params > 0) {
	if (strcasecmp(params[0], "true") == 0 ||
	    strcasecmp(params[0], "yes"))
	    ToggleGadgetSet(include_toggle, True);
	else if (strcasecmp(params[0], "false") == 0 ||
		 strcasecmp(params[0], "no") == 0)
	    ToggleGadgetSet(include_toggle, False);
    }
}

void action_followup(Widget w, XEvent *event,
		     String *params, Cardinal *no_params)
{
    Boolean	old = ToggleGadgetGet(include_toggle);

    do_incl(params, no_params);
    post_menu_callback(w, (XtPointer)POST, NULL);
    ToggleGadgetSet(include_toggle, old);
}

void action_reply(Widget w, XEvent *event,
		  String *params, Cardinal *no_params)
{
    Boolean	old = ToggleGadgetGet(include_toggle);

    do_incl(params, no_params);
    post_menu_callback(w, (XtPointer)MAIL, NULL);
    ToggleGadgetSet(include_toggle, old);
}

void action_followup_and_reply(Widget w, XEvent *event,
			       String *params, Cardinal *no_params)
{
    Boolean	old = ToggleGadgetGet(include_toggle);

    do_incl(params, no_params);
    post_menu_callback(w, (XtPointer)(POST|MAIL), NULL);
    ToggleGadgetSet(include_toggle, old);
}

void action_post_new(Widget w, XEvent *event,
		     String *params, Cardinal *no_params)
{
    post_new_callback(w, NULL, NULL);
}

void action_forward_by_mail(Widget w, XEvent *event,
			    String *params, Cardinal *no_params)
{
    forward_menu_callback(w, NULL, NULL);
}
