#include "global.h"
#include "../configure.h"
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "hash.h"
#include "pedit.h"
#include "post.h"
#include "server.h"
#include "widgets.h"
#include "xutil.h"

#define IS_SPACE(c) ((c) == ' ' || (c) == '\t')

static char *mystrnstr(char *stack, char *needle, int n)
{
    char	*p;
    int		len;

    if (!needle || (len = strlen(needle)) == 0)
	return NULL;

    for (p = strchr(stack, needle[0]) ;
	 p && p < stack + n ;
	 p = strchr(p + 1, needle[0])) {
	if (strncmp(p, needle, len) == 0)
	    return p;
    }

    return NULL;
}

static int an_admin()
{
    static int	inited = False;
    static int	result = False;

    if (!inited) {
	uid_t		uid;

	inited = True;
	if ((uid = getuid()) == 0)
	    result = True;
	else {
	    struct passwd	*pwd;

	    pwd = getpwnam("news");
	    if (pwd && uid == pwd->pw_uid)
		result = True;
	}
    }

    return result;
}

static char *check_cancel(char *start, char *stop)
{
    for (;;) {
	ARTICLE	*art;
	char	*c;

	while (IS_SPACE(*start))
	    start++;

	if (start >= stop || *start == '\n')
	    return NULL;

	if (*start != '<')
	    return "Syntax error in control message.";

	start++;
	for (c = start ; c < stop && *c != '>' && *c != '<'; c++);
	if (c >= stop || *c != '>')
	    return "Syntax error in control message.";

	if (global.mode != NewsModeGroup &&
	    global.mode != NewsModeThread)
	    return "Error! Failed to authenticate control message.";

	art = *art_hash_find(global.thr.art_ht, start, c - start);
	if (!art)
	    return "Error! Failed to authenticate control message.";

	/* possible security hole */
	if (!mystrnstr(art->from, global.mail_name, strlen(art->from)) ||
	    !mystrnstr(art->from, global.domain_name, strlen(art->from)))
	    return "Error! You cannot cancel somebody else's article.";

	start = c + 1;
    }

    return NULL; /* ok */
}

int check_article_to_post(PostContext *context, char *message)
{
    struct stat		st_buf;
    int			fd;
    char		*art, *c;
    int			n, q;
    Boolean		has_subject_header = False;
    Boolean		has_newsgroups_header = False;
    Boolean		has_from_header = False;
    Boolean		has_mail_headers = False;
    Boolean		needs_sender_header = False;

    context->flags &= ~OK_TO_POST;
    message[0] = '\0';

    if (stat(context->file_name, &st_buf) < 0) {
	perror("knews: stat");
	strcpy(message, "Error! Failed to open temp file.");
	return False;
    }

    if (!S_ISREG(st_buf.st_mode)) {
	strcpy(message, "Error! Failed to open temp file.");
	return False;
    }

    fd = open(context->file_name, O_RDONLY);
    if (fd < 0) {
	perror("knews: open");
	strcpy(message, "Error! Failed to open temp file.");
	return False;
    }

    art = malloc(st_buf.st_size + 8);
    if (!art) {
	(void)close(fd);
	strcpy(message,
	       "Error! Failed to allocate\n"
	       "sufficient memory for article/mail.");
	return False;
    }

    n = st_buf.st_size;
    c = art;
    while (n > 0) {
	int	i;

	i = read(fd, c, n);
	if (i < 0) {
	    perror("knews: read");
	    (void)close(fd);
	    free(art);
	    strcpy(message, "Error! Failed to read temp file.");
	    return False;
	} else if (i == 0) {
	    *c = '\0';
	    break;
	}
	n -= i;
	c += i;
    }
    art[st_buf.st_size] = '\0';
    (void)close(fd);

    /* check headers */
    n = 0; /* line number */
    c = art;
    for (;;) {
	char	*p1, *p2, *p3;
	int	len;
	int	lines;

	if (*c == '\n' && c > art) { /* we've reached the body */
	    c++;
	    break;
	}

	for (p1 = c ; *p1 != ':' ; p1++) {
	    if (p1 > c + 80 ||
		!isascii((unsigned char)*p1) ||
		!isprint((unsigned char)*p1) ||
		*p1 == '\t' || *p1 == ' ' ||
		*p1 == '\0') {
		p2 = strchr(p1, '\n');
		if (p2 && p2 < p1 + 8)
		    *p2 = '\0';
		else
		    p1[8] = '\0';
		if (strlen(c) > 48)
		    c[48] = '\0';
		sprintf(message,
			"Syntax error in '%s:' header\n"
			"Remember to put an EMPTY line between\n"
			"the headers and the body (the article/mail text)!",
			c);
		context->line = n;
		free(art);
		return False;
	    }
	}

	p1[0] = '\0';
	if (p1[1] != ' ' && p1[1] != '\t') {
	    p2 = strchr(p1, '\n');
	    if (p2 && p2 < p1 + 8)
		*p2 = '\0';
	    else
		p1[8] = '\0';
	    if (strlen(c) > 48)
		c[48] = '\0';
	    sprintf(message,
		    "Syntax error in '%s:' header.\n"
		    "Must have a space after the colon!",
		    c);
	    context->line = n;
	    free(art);
	    return False;
	}
	len = p1 - c;

	lines = 1;
	p1 += 2;
	while ( (p2 = strchr(p1, '\n')) ) {
	    p1 = p2 + 1;
	    if (*p1 == ' ' || *p1 == '\t')
		lines++;
	    else
		break;
	}

	context->line = n + 1;

#define RETURN_ERROR(a) \
	do { \
	    strcpy(message, a); \
	    free(art); \
	    return False; \
	} while (0)
#define	IS_HEADER(header) \
        (len + 1 == sizeof(header) && strcasecmp(c, header) == 0)

	if (IS_HEADER("Subject")) {
	    if (has_subject_header)
		RETURN_ERROR("Error! Duplicate 'Subject:' header.");
	    p2 = c + 8;
	    while (*p2 == ' ' || *p2 == '\t')
		p2++;
	    if (*p2 == '\n' || *p2 == '\0')
		RETURN_ERROR("Error! Empty 'Subject:' header.");
	    has_subject_header = True;
	} else if (IS_HEADER("From")) {
	    if (has_from_header)
		RETURN_ERROR("Error! Duplicate 'From:' header.");
	    has_from_header = True;
	    if (mystrnstr(c + 5, global.domain_name, p1 - c - 5) &&
		mystrnstr(c + 5, global.user_id, p1 - c - 5))
		needs_sender_header = False;
	    else
		needs_sender_header = True;
	} else if (IS_HEADER("Newsgroups")) {
	    char	*p3;

	    if (lines > 1)
		RETURN_ERROR("Error! The 'Newsgroups:' header cannot\n"
			     "be continued on several lines.");
	    p2 = c + 11;
	    while (*p2 == ' ' || *p2 == '\t')
		p2++;
	    if (*p2 == '\n')
		RETURN_ERROR("Error! Empty 'Newsgroups:' header");
	    if ( ( (p3 = strchr(p2, ' ')) && (p3 < p1) ) ||
		 ( (p3 = strchr(p2, '\t')) && (p3 < p1)) )
		RETURN_ERROR("Error! There can be no spaces between the\n"
			     "commas in the 'Newsgroups:' header");
	    has_newsgroups_header = True;
	} else if (IS_HEADER("Followup-To")) {
	    if (lines > 1)
		RETURN_ERROR("Error! The 'Followup-To:' header cannot\n"
			     "be continued on several lines");
	    p2 = c + 12;
	    if (*p2 == '\n')
		RETURN_ERROR("Error! Empty 'Followup-To:' header");
	    while (*p2 == ' ' || *p2 == '\t')
		p2++;
	    if ( ( (p3 = strchr(p2, ' ')) && (p3 < p1) ) ||
		 ( (p3 = strchr(p2, '\t')) && (p3 < p1)) )
		RETURN_ERROR("Error! There can be no spaces between the\n"
			     "commas in the 'Followup-To:' header");
	} else if (IS_HEADER("Control")) {
	    if (!an_admin()) {
		p2 = c + 9;
		while (*p2 == ' ' || *p2 == '\t')
		    p2++;
		if (strncmp(p2, "cancel ", 7) != 0)
		    RETURN_ERROR("Error! No permission "
				 "to send control message.");
		else {
		    char	*temp = check_cancel(p2 + 7, p1);

		    if (temp)
			RETURN_ERROR(temp);
		}
	    }
	} else if (IS_HEADER("Also-Control")) {
	    if (!an_admin()) {
		p2 = c + 14;
		while (*p2 == ' ' || *p2 == '\t')
		    p2++;
		if (strncmp(p2, "cancel ", 7) != 0)
		    RETURN_ERROR("Error! No permission "
				 "to send control message.");
		else {
		    char	*temp = check_cancel(p2 + 7, p1);

		    if (temp) RETURN_ERROR(temp);
		}
	    }
	} else if (IS_HEADER("Supersedes")) {
	    if (!an_admin()) {
		char	*temp = check_cancel(c + 12, p1);
		
		if (temp)
		    RETURN_ERROR(temp);
	    }
	} else if (IS_HEADER("Sender")) {
	    RETURN_ERROR("Error! You cannot specify a 'Sender:' header.");
	} else if (IS_HEADER("To")) {
	    has_mail_headers = True;
	} else if (IS_HEADER("Cc")) {
	    has_mail_headers = True;
	}
#undef	IS_HEADER

	n += lines;
	c = p1;
    }

    if (!has_subject_header)
	RETURN_ERROR("Error! Article/mail has no 'Subject:' header.");
    if (!has_from_header)
	RETURN_ERROR("Error! Article/mail has no 'From:' header.");
    if (!has_newsgroups_header && !has_mail_headers)
	RETURN_ERROR("Error! Article/mail has no "
		     "'Newsgroups:', 'To:' or 'Cc:' header.");

    context->line += 2;
    /* check the body */
    n = 0; /* max line length */
    q = 0; /* no quoted - no unquoted + no '<' */
    while (c) {
	char	*p1;
	int	i;

	p1 = strchr(c, '\n');
	if (p1) {
	    i = p1 - c;
	    c = p1 + 1;
	    switch (*c) {
	    case '>':
		q++;
		break;
	    case '<':
		q--;
		/* fall through */
	    default:
		q--;
	    }
	} else {
	    i = strlen(c);
	    c = NULL;
	}
	if (i > n)
	    n = i;	    
    }

    if (n > 1024)
	RETURN_ERROR("Error! The article/mail contains lines\n"
		     "longer than 1024 characters.\n"
		     "Posting will not be allowed.");
    else if (n > 75)
	strcpy(message,
	       "Note: The article/mail contains lines\n"
	       "longer than 75 characters.\n"
	       "You should consider editing it.");
    else if (q > 0)
	strcpy(message,
	       "Note: This article/mail has more\n"
	       "quoted text than new.");
    else
	strcpy(message, "The article/mail seems to be ok...");

    free(art);

    if (has_mail_headers)
	context->flags |= MAIL;
    else
	context->flags &= ~MAIL;

    if (has_newsgroups_header)
	context->flags |= POST;
    else
	context->flags &= ~POST;

    if (needs_sender_header)
	context->flags |= NEEDS_SENDER;
    else
	context->flags &= ~NEEDS_SENDER;

    context->flags |= OK_TO_POST;

    return True;
}

PostStatus post_article(PostContext *context, char *message)
{
    char	buffer[4096];
    char	*reply, *c;
    int		status;
    FILE	*fp;

    if (message)
	message[0] = '\0';

    fp = fopen(context->file_name, "r");
    if (!fp)
	return PostStatusOpen;

    set_busy(True, False); /* don't quit nicely */

    reply = server_comm(&global.server, "POST\r\n", True);
    if (!reply) {
	(void)reconnect_server(True);
	unset_busy();
	(void)fclose(fp);
	return PostStatusBroken;
    }

    status = atoi(reply);
    if (status != CONT_POST) {
	if (message) {
	    if (strlen(reply) > 256)
		reply[256] = '\0';
	    strcpy(message, reply);
	}
	(void)fclose(fp);
	unset_busy();
	return PostStatusFailed;
    }

    buffer[0] = '\0';
    if (context->flags & NEEDS_SENDER) {
	char	*full_name = GroupObjectFullName(global.group_object);

	if (full_name)
	    sprintf(buffer, "Sender: %s@%s (%s)\r\n",
		    global.user_id, global.domain_name, full_name);
	else
	    sprintf(buffer, "Sender: %s@%s\r\n",
		    global.user_id, global.domain_name);
    }
    c = buffer + strlen(buffer);

    for (;;) {
	int	len = sizeof(buffer) - (c - buffer);

	if (len < 128) {
	    server_write(&global.server, buffer);
	    c = buffer;
	    c[0] = '\0';
	    len = sizeof(buffer);
	}

	if (fgets(c, len - 8, fp)) {
	    if (c[0] == '.' && (c[1] == '\n' || c[1] == '\0')) {
		c[1] = '.';
		c[2] = '\r';
		c[3] = '\n';
		c[4] = '\0';
		c += 4;
	    } else {
		char	*p = strchr(c, '\n');

		if (p) {
		    p[0] = '\r';
		    p[1] = '\n';
		    p[2] = '\0';
		    c = p + 2;
		} else {
		    c += strlen(c);
		}
	    }
	} else {
	    sprintf(c, "\r\n.\r\n");
	    server_write(&global.server, buffer);
	    break;
	}
    }

    (void)fclose(fp);
    reply = server_read(&global.server);
    if (!reply) {
	(void)reconnect_server(True);
	unset_busy();
	return PostStatusBroken;
    }
    unset_busy();
    status = atoi(reply);
    if (status == OK_POSTED) {
	context->flags |= POST_DONE;
	return PostStatusOk;
    } else {
	if (message) {
	    if (strlen(reply) > 256)
		reply[256] = '\0';
	    strcpy(message, reply);
	}
	return PostStatusFailed;
    }
}

MailStatus mail_article(PostContext *context, char *stderr_buf)
{
    int		fd, status;
    pid_t	pid, wpid;

    fd = open(context->file_name, O_RDONLY);
    if (fd < 0) {
	perror("knews: open");
	return MailStatusOpen;
    }

    pid = fork_nicely(NULL, NULL, True);

    if (pid < 0) {
	perror("knews: fork");
	(void)close(fd);
	return MailStatusFork;
    } else if (pid == 0) { /* child */
	if (fd != STDIN_FILENO) {
	    fd = dup2(fd, STDIN_FILENO);
	    if (fd < 0) {
		perror("knews: dup2");
		_exit(127);
	    }
	}

/*	(void)execl("/bin/sh", "sh", "-c", "xterm", (char *)0);*/
#ifndef MAIL_COMMAND
#error You need to define MAIL_COMMAND in configure.h
#endif
	(void)execl("/bin/sh", "sh", "-c", MAIL_COMMAND, (char *)0);
	perror("knews: execl");
	_exit(127);
    }

    (void)close(fd);
    set_busy(False, False);
    wpid = wait_for_pid(pid, &status, stderr_buf);
    unset_busy();

    if (wpid == pid && WIFEXITED(status))
	status = WEXITSTATUS(status);
    else
	status = 1;

    if (status == 0)
        return MailStatusOk;
    else if (status == 127)
	return MailStatusStart;
    else
	return MailStatusFailed;
}

PostStatus post_to_agent(PostContext *context, char *agent, char *stderr_buf)
{
    pid_t	pid, wpid;
    int		status;

    pid = fork_nicely(NULL, NULL, True);

    if (pid < 0)
	return PostStatusStart;

    if (pid == 0) {
	int	fd;

	fd = open(context->file_name, O_RDONLY);
	if (fd < 0) {
	    perror("knews: open");
	    _exit(126);
	}

	if (fd != STDIN_FILENO &&
	    dup2(fd, STDIN_FILENO) != STDIN_FILENO) {
	    perror("knews: dup2");
	    _exit(126);
	}

	execl("/bin/sh", "sh", "-c", agent, (char *)0);
	perror("knews: execl");
	_exit(127);
    }

    set_busy(False, False);
    wpid = wait_for_pid(pid, &status, stderr_buf);
    unset_busy();

    if (wpid == pid && WIFEXITED(status))
	status = WEXITSTATUS(status);
    else
	status = 1;

    if (status == 0)
        return PostStatusOk;
    else if (status == 126)
	return PostStatusOpen;
    else if (status == 127)
	return PostStatusStart;
    else
	return PostStatusFailed;
}
