/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include "codes.h"
#include "connect.h"
#include "kill.h"
#include "read.h"
#include "server.h"
#include "util.h"
#include "viewer.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtText.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/Message.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"

typedef enum {
    MimeEncUnknown,
    MimeEncNone, /* 7bit, 8bit and binary */
    MimeEncBase64,
    MimeEncQP
} MimeEnc;

#define IS_SPACE(c) \
	((c) == ' ' || (c) == '\t')
#define IS_DOT(buffer) \
	((buffer)[0] == '.' && (buffer)[1] == '\0')
#define IS_BOUNDARY(buffer, boundary, len) \
	(boundary && (buffer)[0] == '-' && (buffer)[1] =='-' && \
	 strncmp((buffer) + 2, (boundary), (len)) == 0)
#define IS_LAST_BOUNDARY(buffer, boundary, len) \
	(IS_BOUNDARY(buffer, boundary, len) && \
	 (buffer)[(len) + 2] == '-' && (buffer)[(len) + 3] == '-')

#define MIME_BUF_LEN	127

static MimeEnc parse_content_enc(char *header)
{
    char	*c = header + 26;

    for (;;) { /* so that we can continue */
	while (IS_SPACE(*c))
	    c++;

	switch (*c++) {
	case '(':
	    while (*c != '\0' && *c != ')')
		c++;
	    if (*c == ')')
		c++;
	    continue;
	case '7':
	    if (strncasecmp(c, "bit", 3) == 0)
		return MimeEncNone;
	    break;
	case '8':
	    if (strncasecmp(c, "bit", 3) == 0)
		return MimeEncNone;
	    break;
	case 'b':
	case 'B':
	    if (strncasecmp(c, "inary", 5) == 0)
		return MimeEncNone;
	    if (strncasecmp(c, "ase64", 5) == 0)
		return MimeEncBase64;
	    break;
	case 'q':
	case 'Q':
	    if (strncasecmp(c, "uoted-printable", 15) == 0)
		return MimeEncQP;
	    break;
	default:
	    return MimeEncUnknown;
	}

	return MimeEncUnknown;
    }
    /* not reached */
}

static int get_token(char **c, char **p)
{
    int	len;

    if (**c == '"') {
	*p = ++*c;
	while (**p != '\0' && **p != '"')
	    (*p)++;
	if (**p == '\0')
	    len = -1;
	else
	    len = (*p)++ - *c;
    } else {
	*p = *c;
	while (**p != '\0' && **p != ' ' &&
	       **p != '\t' && **p != '/' &&
	       **p != ';' && **p != '(' &&
	       **p != '=' && **p != '"')
	    (*p)++;
	len = *p - *c;
    }

    if (len > MIME_BUF_LEN)
	return -1;

    return len;
}

static Boolean parse_content_type(char **headers, char *type_buf,
				  char *subtype_buf, Arg *args, int n_args)
{
    enum {
	LookingForType,
	LookingForSlash,
	LookingForSubtype,
	LookingForParam,
	LookingForEqual,
	LookingForValue,
	LookingForSemicolon
    }		state = LookingForType;
    char	*c, *p;
    int		len, n;

    c = headers[0] + 13;
    n = 0;
    do {
	for (;;) {
	    while (IS_SPACE(*c))
		c++;

	    if (*c == '\0')
		break;
	    if (*c == '(') { /* skip comments */
		while (*c != '\0' && *c != ')')
		    c++;
		continue;
	    }

	    switch (state) {
	    case LookingForType:
		len = get_token(&c, &p);
		if (len < 0)
		    return False;
		(void)memcpy(type_buf, c, len);
		type_buf[len] = '\0';
		c = p;
		state = LookingForSlash;
		break;
	    case LookingForSlash:
		if (*c++ != '/')
		    return False;
		state = LookingForSubtype;
		break;
	    case LookingForSubtype:
		len = get_token(&c, &p);
		if (len < 0)
		    return False;
		(void)memcpy(subtype_buf, c, len);
		subtype_buf[len] = '\0';
		c = p;
		state = LookingForSemicolon;
		break;
	    case LookingForSemicolon:
		if (*c++ != ';')
		    return True; /* lenience */
		state = LookingForParam;
		break;
	    case LookingForParam:
		if (n >= n_args)
		    return True;
		len = get_token(&c, &p);
		if (len < 0)
		    return True; /* lenience */
		args[n].name = XtMalloc(len + 1);
		(void)memcpy((void *)args[n].name, c, len);
		args[n].name[len] = '\0';
		c = p;
		state = LookingForEqual;
		break;
	    case LookingForEqual:
		if (*c++ != '=')
		    return True; /* lenience */
		state = LookingForValue;
		break;
	    case LookingForValue:
		len = get_token(&c, &p);
		if (len < 0)
		    return True; /* lenience */
		args[n].value = (XtArgVal)XtMalloc(len + 1);
		(void)memcpy((void *)args[n].value, c, len);
		((char *)args[n++].value)[len] = '\0';
		c = p;
		state = LookingForSemicolon;
		break;
	    }
	}

	c = *++headers;
    } while (c && IS_SPACE(*c));

    return (state != LookingForType &&
	    state != LookingForSlash &&
	    state != LookingForSubtype);
}

static int insert_headers(char **headers, int n, int full_header)
{
    char	**loop = GroupObjectHeaderFormat(global.group_object);
    int		added_some = False;
    int		i;

    if (loop && !full_header) {
	while (*loop) {
	    int	len = strlen(*loop);

	    for (i = 0 ; i < n ; i++) {
		if (strncasecmp(*loop, headers[i], len) == 0) {
		    do {
			ArtTextAddLine(main_widgets.text, headers[i],
				       LineTypeHeader);
			i++;
		    } while (i < n && IS_SPACE(headers[i][0]));
		    added_some = True;
		    break;
		}
	    }
	    loop++;
	}
    } else {
	if (n > 0)
	    added_some = True;

	for (i = 0 ; i < n ; i++)
	    ArtTextAddLine(main_widgets.text, headers[i], LineTypeHeader);
    }

    if (added_some)
	ArtTextAddLine(main_widgets.text, "", LineTypeOrdinary);

    return added_some;
}

/* returns true if soft line break */
static int do_quoted_printable(char *buffer)
{
    char	*src = buffer;
    char	*dest = buffer;
    int		c1, c2;

    for (;;) {
	switch (*src) {
	case '=':
	    c1 = src[1];
	    if (c1 >= '0' && c1 <= '9')
		c1 -= '0';
	    else if (c1 >= 'A' && c1 <= 'F')
		c1 -= 'A' - 10;
	    else if (c1 >= 'a' && c1 <= 'f')
		c1 -= 'a' - 10;
	    else if (c1 == '\0') {
		*dest = '\0';
		return True;
	    } else
		break; /* not a hex digit */

	    c2 = src[2];
	    if (c2 >= '0' && c2 <= '9')
		c2 -= '0';
	    else if (c2 >= 'A' && c2 <= 'F')
		c2 -= 'A' - 10;
	    else if (c2 >= 'a' && c2 <= 'f')
		c2 -= 'a' - 10;
	    else
		break; /* not a hex digit */

	    *dest++ = c1 * 16 + c2;
	    src += 3;
	    continue;
	default:
	    break;
	}

	if ((*dest++ = *src++) == '\0')
	    break;
    }

    return False;
}

static long do_base64(char *buffer)
{
    static int	base64_alpha[256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
	-1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    };
    char	*src;
    long	n, len, m;

    src = strchr(buffer, '=');
    if (src)
	len = src - buffer;
    else
	len = strlen(buffer);

    src = buffer;
    n = 0;
    m = len / 4;
    while (m-- > 0) {
	int	c1 = base64_alpha[(unsigned char)*src++];
	int	c2 = base64_alpha[(unsigned char)*src++];
	int	c3 = base64_alpha[(unsigned char)*src++];
	int	c4 = base64_alpha[(unsigned char)*src++];

	if (c1 < 0 || c2 < 0 || c3 < 0 || c4 < 0)
	    return -1;

	buffer[n++] = (c1 << 2) | ((c2 & 0x30) >> 4);
	buffer[n++] = ((c2 & 0x0f) << 4) | ((c3 & 0x3c) >> 2);
	buffer[n++] = ((c3 & 0x03) << 6) | c4;
    }

    switch (len % 4) {
	int c1, c2, c3;
    case 3:
	c1 = base64_alpha[(unsigned char)*src++];
	c2 = base64_alpha[(unsigned char)*src++];
	c3 = base64_alpha[(unsigned char)*src++];
	if (c1 < 0 || c2 < 0 || c3 < 0)
	    return -1;
	buffer[n++] = (c1 << 2) | ((c2 & 0x30) >> 4);
	buffer[n++] = ((c2 & 0x0f) << 4) | ((c3 & 0x3c) >> 2);
	break;
    case 2:
	c1 = base64_alpha[(unsigned char)*src++];
	c2 = base64_alpha[(unsigned char)*src++];
	if (c1 < 0 || c2 < 0)
	    return -1;
	buffer[n++] = (c1 << 2) | ((c2 & 0x30) >> 4);
	break;
    case 1:
	return -1;
    }

    buffer[n] = '\0';

    return n;
}

static char *add_clickable(char *buffer, char **headers, char *viewer,
			   MimeEnc enc, char *bound, int len)
{
    char	*data = XtMalloc(16);
    long	alloced = 16;
    long	data_len = 0;
    VIEWABLE	*viewable;
    Boolean	error = False;

    while (buffer && !IS_DOT(buffer) &&
	   !IS_BOUNDARY(buffer, bound, len)) {
	long	n = 0;
	Boolean	soft;

	switch (enc) {
	case MimeEncQP:
	    soft = do_quoted_printable(buffer);
	    n = strlen(buffer);
	    if (!soft)
		buffer [n++] = '\n';
	    break;
	case MimeEncNone:
	case MimeEncUnknown:
	    n = strlen(buffer);
	    buffer[n++] = '\n';
	    break;
	case MimeEncBase64:
	    n = do_base64(buffer);
	    break;
	}

	if (n < 0)
	    error = True;
	else if (n > 0) {
	    if (n + data_len >= alloced) {
		alloced = 2 * (data_len + n);
		data = XtRealloc(data, alloced);
	    }

	    (void)memmove(data + data_len, buffer, n);
	    data_len += n;
	}

	buffer = server_read(&global.server);
    }

    data = XtRealloc(data, data_len);

    viewable = (VIEWABLE *)XtMalloc(sizeof(VIEWABLE));
    viewable->data = data;
    viewable->len = data_len;
    viewable->viewer = viewer;
    viewable->ref_count = 1;

    ArtTextAddClickable(main_widgets.text, *headers,
			click_callback, (XtPointer)viewable);
    headers++;
    while (*headers && (**headers == ' ' || **headers == '\t'))
	ArtTextAddLine(main_widgets.text, *headers++, LineTypeOrdinary);
    if (error)
	ArtTextAddLine(main_widgets.text, "    [decode error]",
		       LineTypeQuoted);
    else if (enc == MimeEncUnknown)
	ArtTextAddLine(main_widgets.text,
		       "    [unrecogized content-transfer-encoding]",
		       LineTypeQuoted);
    ArtTextAddLine(main_widgets.text, "", LineTypeOrdinary);

    return buffer;
}

static void add_body_line(char *buffer, regex_t *re)
{
    LineType	type;

    if (re && regexec(re, buffer, 0, NULL, 0) == 0)
	type = LineTypeQuoted;
    else
	type = LineTypeOrdinary;

    ArtTextAddLine(main_widgets.text, buffer, type);
}

static char *insert_body(char *buffer, MimeEnc enc, char *bound, int len)
{
    regex_t	*re = GroupObjectQuoteRegexp(global.group_object);
    Boolean	flag;

    switch (enc) {
    case MimeEncUnknown:
    case MimeEncNone:
	while (buffer && !IS_DOT(buffer) &&
	       !IS_BOUNDARY(buffer, bound, len)) {
	    add_body_line(buffer, re);
	    buffer = server_read(&global.server);
	}
	break;
    case MimeEncQP:
	flag = False;
	while (buffer && !IS_DOT(buffer) &&
	       !IS_BOUNDARY(buffer, bound, len)) {
	    Boolean	soft;

	    soft = do_quoted_printable(buffer);
	    if (flag)
		ArtTextAppendToLast(main_widgets.text, buffer);
	    else
		add_body_line(buffer, re);

	    flag = soft;
	    buffer = server_read(&global.server);
	}
	break;
    case MimeEncBase64:
	flag = False;
	while (buffer && !IS_DOT(buffer) &&
	       !IS_BOUNDARY(buffer, bound, len)) {
	    if (do_base64(buffer) > 0) {
		char	*p, *c = buffer;

		for (p = c ; ; p++) {
		    if (*p == '\0') {
			if (flag)
			    ArtTextAppendToLast(main_widgets.text, c);
			else
			    add_body_line(c, re);
			flag = True;
			break;
		    } else if (*p == '\n') {
			*p++ = '\0';
			if (flag)
			    ArtTextAppendToLast(main_widgets.text, c);
			else
			    add_body_line(c, re);
			if (*p == '\r')
			    p++;
			if (*p == '\0')
			    break;
			c = p;
			flag = False;
		    }
		}
	    }
	    buffer = server_read(&global.server);
	}
	break;
    default:
	while (buffer && !IS_DOT(buffer) &&
	       !IS_BOUNDARY(buffer, bound, len))
	    buffer = server_read(&global.server);
	break;
    }

    return buffer;
}

static char *do_mime(char *buffer, Boolean default_to_rfc822,
		     char *ext_bound, int ext_len)
{
#define N_ARGS 8
    MimeEnc	enc = MimeEncNone;
    char	**headers;
    char	type_buf[MIME_BUF_LEN + 1], subtype_buf[MIME_BUF_LEN + 1];
    Arg		args[N_ARGS + 1];
    long	no_alloc, n, i, ct_index;

    n = 0;
    no_alloc = 16;
    headers = (char **)XtMalloc(no_alloc * sizeof(char *));
    for (i = 0 ; i < no_alloc ; i++) headers[i] = NULL;
    while (buffer && buffer[0] != '\0' &&
	   !IS_DOT(buffer) && !IS_BOUNDARY(buffer, ext_bound, ext_len)) {
	if (n + 3 > no_alloc) {
	    i = no_alloc;
	    no_alloc *= 2;
	    headers = (char **)XtRealloc((char *)headers,
					 no_alloc * sizeof(char *));
	    while (i < no_alloc)
		headers[i++] = NULL;
	}

	headers[n] = XtNewString(buffer);
	n++;
	buffer = server_read(&global.server);
    }

    ct_index = -1;
    for (i = 0 ; i < n ; i++)
	if (strncasecmp(headers[i], "Content-Type:", 13) == 0)
	    ct_index = i;
	else if (strncasecmp(headers[i],
			     "Content-Transfer-Encoding:", 26) == 0)
	    enc = parse_content_enc(headers[i]);

    if (!ext_bound)
	ArtTextClearLines(main_widgets.text); /* top level */

    for (i = 0 ; i <= N_ARGS ; i++) {
	args[i].name = NULL;
	args[i].value = (XtArgVal)0;
    }

    if (ct_index >= 0) {
	if (!parse_content_type(headers + ct_index, type_buf,
				subtype_buf, args, N_ARGS))
	    ct_index = -1;
    } else if (default_to_rfc822) {
	strcpy(type_buf, "message");
	strcpy(subtype_buf, "rfc822");
	ct_index = 0; /* hack */
    }

    if (ct_index < 0) {
	if (buffer && buffer[0] == '\0') {
	    insert_headers(headers, n, False);
	    buffer = insert_body(server_read(&global.server),
				 enc, ext_bound, ext_len);
	}
    } else {
	char		*viewer;
	unsigned char	*c;

	for (c = (unsigned char *)type_buf    ; *c != '\0' ; c++)
	    if (isupper(*c)) *c = tolower(*c);
	for (c = (unsigned char *)subtype_buf ; *c != '\0' ; c++)
	    if (isupper(*c)) *c = tolower(*c);

	viewer = get_viewer(type_buf, subtype_buf, args);
	if (viewer) {
	    insert_headers(headers, n, False);
	    buffer = add_clickable(server_read(&global.server),
				   headers + ct_index,
				   viewer, enc, ext_bound, ext_len);
	} else if (strcmp(type_buf, "text") == 0) {
	    insert_headers(headers, n, False);
	    buffer = insert_body(server_read(&global.server),
				 enc, ext_bound, ext_len);
	} else if (strcmp(type_buf, "multipart") == 0) {
	    char	*bound = NULL;

	    /*
	     * treat all subtypes as multipart/mixed,
	     * alternative is just too much trouble
	     */

	    for (i = 0 ; args[i].value ; i++) {
		if (strcasecmp(args[i].name, "boundary") == 0) {
		    bound = (char *)args[i].value;
		    break;
		}
	    }

	    if (bound) {
		int	do_line;

		do_line = insert_headers(headers, n, False);

		n = strlen(bound);
		if (strcmp(subtype_buf, "digest") == 0)
		    default_to_rfc822 = True;
		else
		    default_to_rfc822 = False;

		while (buffer && !IS_DOT(buffer) &&
		       !IS_BOUNDARY(buffer, bound, n) &&
		       !IS_BOUNDARY(buffer, ext_bound, ext_len))
		    buffer = server_read(&global.server);

		if (buffer && !IS_DOT(buffer) &&
		    !IS_BOUNDARY(buffer, ext_bound, ext_len)) {
		    if (!IS_LAST_BOUNDARY(buffer, bound, n)) {
			do {
			    if (do_line) {
				ArtTextAddSeparator(main_widgets.text, 2);
				ArtTextAddSpace(main_widgets.text, 4);
			    }
			    do_line = True;
			    buffer = do_mime(server_read(&global.server),
					     default_to_rfc822, bound, n);
			} while (buffer && !IS_DOT(buffer) &&
				 !IS_LAST_BOUNDARY(buffer, bound, n));
		    }

		    while (buffer && !IS_DOT(buffer) &&
			   !IS_BOUNDARY(buffer, ext_bound, ext_len))
			buffer = server_read(&global.server);
		}
	    } else { /* no boundary, do raw text */
		insert_headers(headers, n, ext_bound != NULL);
		buffer = insert_body(server_read(&global.server),
				     enc, ext_bound, ext_len);
	    }
	} else if (strcmp(type_buf, "message") == 0) {
	    if (strcmp(subtype_buf, "rfc822") == 0) {
		insert_headers(headers, n, False);
		buffer = do_mime(server_read(&global.server),
				 False, ext_bound, ext_len);
	    } else { /* do raw text */
		insert_headers(headers, n, ext_bound != NULL);
		buffer = insert_body(server_read(&global.server),
				     enc, ext_bound, ext_len);
	    }
	} else { /* use internal viewer */
	    insert_headers(headers, n, False);
	    buffer = add_clickable(server_read(&global.server),
				   headers + ct_index,
				   NULL, enc, ext_bound, ext_len);
	}
    }

    for (i = 0 ; i < N_ARGS ; i++) {
	XtFree(args[i].name);
	XtFree((char *)args[i].value);
    }

    for (i = 0 ; i < no_alloc ; i++)
	XtFree(headers[i]);
    XtFree((char *)headers);

    return buffer;    
}

static char *get_article(ARTICLE *art, int full_header, long *sel_data)
{
    char	command[256];
    char	*buffer;
    long	first, shown, size;

    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);

    if (art->from) {
	sprintf(command, "ARTICLE %ld\r\n", art->no);
	buffer = server_comm(&global.server, command, True);
    } else if (art->hash_len < 240) {
	sprintf(command, "ARTICLE <%s>\r\n", art->msgid);
	buffer = server_comm(&global.server, command, True);
    } else {
	char	*temp = XtMalloc(art->hash_len + 16);

	sprintf(temp, "ARTICLE <%s>\r\n", art->msgid);
	buffer = server_comm(&global.server, temp, True);
	XtFree(temp);
    }

    if (!buffer || atoi(buffer) != OK_ARTICLE) {
	ArtTextClearLines(main_widgets.text);
	return buffer;
    }

    if (art->lines > 0) {
	Arg		arg;

	XtSetArg(arg, XtNnAlloc, art->lines + 16);
	XtSetValues(main_widgets.text, &arg, 1);
    }

    buffer = server_read(&global.server);
    if (full_header) {
	long	i;

	ArtTextClearLines(main_widgets.text);

	i = -1;
	while (buffer && buffer[0] != '\0' &&
	       (buffer[0] != '.' || buffer[1] != '\0')) {
	    if (sel_data && sel_data[0] == i)
		ArtTextAddSelected(main_widgets.text, buffer, LineTypeHeader,
				   sel_data[1], sel_data[2]);
	    else
		ArtTextAddLine(main_widgets.text, buffer, LineTypeHeader);
	    i--;
	    buffer = server_read(&global.server);
	}

	if (buffer && buffer[0] == '\0') {
	    regex_t	*re = GroupObjectQuoteRegexp(global.group_object);

	    ArtTextAddLine(main_widgets.text, "", LineTypeOrdinary);

	    buffer = server_read(&global.server);
	    i = 0;
	    while (buffer && (buffer[0] != '.' || buffer[1] != '\0')) {
		LineType	type;

		if (re && regexec(re, buffer, 0, NULL, 0) == 0)
		    type = LineTypeQuoted;
		else
		    type = LineTypeOrdinary;

		if (sel_data && i == sel_data[0])
		    ArtTextAddSelected(main_widgets.text, buffer, type,
				       sel_data[1], sel_data[2]);
		else
		    ArtTextAddLine(main_widgets.text, buffer, type);
		buffer = server_read(&global.server);
		i++;
	    }
	}
    } else {
	buffer = do_mime(buffer, False, NULL, 0);
    }

    ArtTextGetFirstShownSize(main_widgets.text, &first, &shown, &size);
    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, size, shown, first);
    
    if (!buffer)
	return NULL;
    
    if (art->from && !art->read) {
	art->read = True;
	art->subject->no_unread--;
	global.thr.group->no_unread--;

	if (global.mode == NewsModeThread) {
	    ArtTreeNodeSetInner(main_widgets.arttree,
				(ART_TREE_NODE *)art, False);
	    ArtTreeNodeSetPixmap(main_widgets.arttree,
				 (ART_TREE_NODE *)art, None);
	}
	if (art->hot > 0)
	    update_subj_hot_value(art->subject);
	update_subj_entry(art->subject,
			  GroupObjectFixedAuthor(global.group_object));

	if (GroupObjectProcessXrefs(global.group_object))
	    process_xref(art);
    }

    return CODE_TO_STR(OK_ARTICLE);
}

int read_article(ARTICLE *art, int full_header, long *sel_data)
{
    if (art) {
	char	*reply;

	set_busy(True, True);
	reply = get_article(art,
			    full_header ||
			    GroupObjectFullHeader(global.group_object),
			    sel_data);
	if (!reply) {
	    (void)reconnect_server(True);
	    unset_busy();
	    return False;
	}
	unset_busy();

	if (atoi(reply) == OK_ARTICLE)
	    set_standard_message();
	else {
	    char	message[256];

	    if (strlen(reply) > 200) reply[200] ='\0';
	    sprintf(message,
		    "Error! Failed message from server is: %s", reply);
	    set_message(message, True);
	}
    } else {
	ArtTextClearLines(main_widgets.text);
	ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);
	set_standard_message();
    }

    return True;
}
