/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "save.h"
#include "server.h"
#include "tag.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include <X11/Shell.h>
#include "../Widgets/ArtText.h"
#include "../Widgets/FileSel.h"
#include "../Widgets/Knapp.h"
#include "../Widgets/Layout.h"
#include "../Widgets/Message.h"
#include "../Widgets/TextField.h"
#include "../Widgets/Toggle.h"
#include "../Widgets/Util.h"

typedef enum {
    SaveScopeArtwin,
    SaveScopeArticle,
    SaveScopeSubject,
    SaveScopeThread,
    SaveScopeSubthread,
    SaveScopeTagged
} SaveScope;

#define SAVE_BOGUS_FROM	(1<<0)
#define SAVE_BOGUS_SUBJ	(1<<1)
#define SAVE_HEAD	(1<<2)
#define SAVE_BODY	(1<<3)
#define SAVE_EMPTY	(1<<4)

static struct {
    Widget	file_message;
    Widget	file_field;
    Widget	shell_message;
    Widget	shell_field;
    Widget	file_chooser;
    Widget	choose_knapp;
    Widget	ok_knapp;
    Widget	close_knapp;
    /***/
    Widget	bogus_from_toggle;
    Widget	bogus_subj_toggle;
    Widget	head_toggle;
    Widget	body_toggle;
    Widget	empty_toggle;
    /***/
    Widget	artwin_toggle;
    Widget	article_toggle;
    Widget	subject_toggle;
    Widget	thread_toggle;
    Widget	subthread_toggle;
    Widget	tagged_toggle;
} save_widgets;

static SaveScope	save_scope = SaveScopeArticle;

/*************************************************************************/

#define IS_DOT(buffer)       (buffer[0] == '.' && buffer[1] == '\0')
#define IS_SEP(buffer, body) (body && buffer[0] == '\0')

/* only handles modes "a", "w", "r" */
FILE *fopen_expand(char *file_name, char *mode)
{
    int		fd, flags;
    FILE	*result;

    switch (*mode) {
    case 'a':
	flags = O_WRONLY|O_APPEND|O_CREAT;
	break;
    case 'r':
	flags = O_RDONLY;
	break;
    case 'w':
	flags = O_WRONLY|O_TRUNC|O_CREAT;
	break;
    default:
	errno = EINVAL;
	return NULL;
    }

    fd = open_expand(file_name, flags, 0644);
    if (fd < 0)
	return NULL;

    result = fdopen(fd, mode);
    if (!result)
	(void)close(fd);

    return result;
}

int open_expand(char *file_name, int flags, mode_t mode)
{
    char	buffer[1024];
    char	*c;
    int		n = 0;
    char	ch;

    errno = EINVAL;

    if (*file_name == '~') {
	char	*env_var = getenv("HOME");

	file_name++;
	if (!env_var)
	    return -1;
	n = strlen(env_var);
	if (n >= sizeof(buffer))
	    return -1;
	strcpy(buffer, env_var);
    }

    for (ch = *file_name++ ; ch != '\0' ; ch = *file_name++) {
	if (n >= sizeof(buffer) - 8)
	    return -1; /* too long */

	if (ch == '%') {
	    int	cap, slash;

	    ch = *file_name++;
	    switch (ch) {
	    case '%':
		buffer[n++] = '%';
		continue; /* don't fall through */
	    case 'n':
		cap = False;
		slash = False;
		break;
	    case 'N':
		cap = True;
		slash = False;
		break;
	    case 'g':
		cap = False;
		slash = True;
		break;
	    case 'G':
		cap = True;
		slash = True;
		break;
	    default:
		return -1;
	    }

	    if ((global.mode != NewsModeGroup &&
		 global.mode != NewsModeThread) ||
		!global.thr.group)
		return -1;
	    c = global.thr.group->name;
	    if (!c)
		return -1;
	    ch = *c++;
	    if (ch == '\0')
		return -1;

	    if (cap && islower(ch))
		ch = toupper(ch);

	    buffer[n++] = ch;
	    while (*c != '\0') {
		ch = *c++;

		if (n > sizeof(buffer) - 8)
		    return -1;

		if (ch == '.' && slash)
		    ch ='/';
		buffer[n++] = ch;
	    }
	} else {
	    buffer[n++] = ch;
	}
    }
    buffer[n] = '\0';

    if (flags & O_CREAT)
	n = open(buffer, flags, mode);
    else
	n = open(buffer, flags);

    if (n >= 0)
	return n;

#ifndef O_ACCMODE
#define O_ACCMODE 3
#endif
    if (errno == ENOENT && (flags & O_ACCMODE) == O_WRONLY) {
	/* try to build a directory hierarchy */
	if (buffer[0] == '/')
	    c = strchr(buffer + 1, '/');
	else
	    c = strchr(buffer, '/');

	while (c) {
	    struct stat	st_buf;

	    *c = '\0';
	    if (stat(buffer, &st_buf) < 0) {
		if (errno != ENOENT ||
		    mkdir(buffer, 0755) < 0) {
		    perror("knews: stat/mkdir");
		    return -1;
		}
	    }
	    *c++ = '/';

	    c = strchr(c, '/');
	}
    }

    if (flags & O_CREAT)
	n = open(buffer, flags, mode);
    else
	n = open(buffer, flags);

    if (n < 0)
	perror("knews: open");

    return n;
}

static void get_email(ARTICLE *art, char *buf, int max_len)
{
    char	*c = art->from;
    char	*p1, *p2;
    int		len;

    if ((p1 = strchr(c, '<')) && (p2 = strchr(p1, '>'))) {
	len = p2 - p1 - 1;
	if (len > max_len)
	    len = max_len;
	memcpy(buf, p1 + 1, len);
	buf[len] = '\0';
	return;
    }

    p1 = strchr(c, '(');
    if (p1) {
	do {
	    p1--;
	} while (p1 > c && (*p1 == ' ' || *p1 == '\t'));

	if (p1 > c) {
	    len = p1 - c + 1;
	    if (len > max_len)
		len = max_len;
	    memcpy(buf, c, len);
	    buf[len] = '\0';
	    return;
	}
    }

    len = strlen(c);
    if (len > max_len)
	len = max_len;
    memcpy(buf, c, len);
    buf[len] = '\0';
}

static ARTICLE **get_art_list(SaveScope scope)
{
    ARTICLE	*art = NULL, *top = NULL;
    ARTICLE	**result = NULL;
    SUBJECT	*subj = NULL;
    int		n, n_alloc;

    switch (scope) {
    case SaveScopeArtwin:
	result = (ARTICLE **)XtMalloc(sizeof(ARTICLE *));
	*result = NULL;
	return result;
    case SaveScopeArticle:
	if (global.curr_art) {
	    result = (ARTICLE **)XtMalloc(2 * sizeof(ARTICLE *));
	    result[0] = global.curr_art;
	    result[1] = NULL;
	}
	return result;
    case SaveScopeTagged:
	{
	    ARTICLE **tagged = get_tagged_articles();

	    n_alloc = no_tagged_articles();
	    if (n_alloc <= 0)
		return NULL;
	    result = (ARTICLE **)XtMalloc((n_alloc + 1) * sizeof(ARTICLE *));
	    for (n = 0 ; n < n_alloc ; n++)
		result[n] = tagged[n];
	    result[n++] = NULL;
	    return result;
	}
    case SaveScopeSubject:
	subj = global.curr_subj;
	if (!subj)
	    return NULL;
	art = subj->thread;
	break;
    case SaveScopeThread:
	if (!global.curr_subj)
	    return NULL;
	art = global.curr_subj->thread;
	break;
    case SaveScopeSubthread:
	if (!global.curr_art)
	    return NULL;
	top = art = global.curr_art;
	break;
    }

    while (art && (!art->from || (subj && art->subject != subj)))
	if (top)
	    art = next_in_subthread_preorder(art, top);
	else
	    art = next_in_thread_preorder(art);

    if (!art)
	return NULL;

    n = 0;
    n_alloc = 16;
    result = (ARTICLE **)XtMalloc(n_alloc * sizeof(ARTICLE *));

    while (art) {
	if (n + 8 > n_alloc) {
	    n_alloc += 32;
	    result =
		(ARTICLE **)XtRealloc((char *)result,
				      n_alloc * sizeof(ARTICLE *));
	}

	result[n++] = art;

	do {
	    if (top)
		art = next_in_subthread_preorder(art, top);
	    else
		art = next_in_thread_preorder(art);
	} while (art && (!art->from || (subj && art->subject != subj)));
    }
    result[n] = NULL;

    return result;
}

static int save_to_file(FILE *file, char *message,
			ARTICLE **art_list, int what, long result[2])
{
    char	*buffer = CODE_TO_STR(OK_ARTICLE);
    char	*p = message + strlen(message);
    long	n_pending, n_done, n_put, n_failed, n;
    int		file_err = False;

    n_done = 0;
    n_failed = 0;
    n_put = n_pending =
	server_write_chunk_arts(&global.server, art_list,
				what & SAVE_HEAD, what & SAVE_BODY);

    while (n_pending-- > 0) {
	if (n_pending < global.chunk_size) {
	    n = server_write_chunk_arts(&global.server, art_list + n_put,
					what & SAVE_HEAD, what & SAVE_BODY);
	    n_pending += n;
	    n_put += n;
	}

	buffer = server_read(&global.server);
	if (!buffer)
	    return -1;

	n = atoi(buffer);
	if (n == OK_ARTICLE || n == OK_HEAD || n == OK_BODY) {
	    if (what & SAVE_BOGUS_FROM) {
		if (!file_err) {
		    char	email[128];

		    get_email(art_list[n_done + n_failed],
			      email, sizeof(email) - 1);
		    if (fprintf(file, "From %s", email) == EOF) {
			perror("knews: fprintf");
			file_err = True;
		    }
		}

		if (!file_err) {
		    time_t	t = art_list[n_done + n_failed]->date;
		    char	*c = ctime(&t);

		    if (fprintf(file, " %s", c) == EOF) {
			perror("knews: fprintf");
			file_err = True;
		    }
		}
	    }

	    if (what & SAVE_BOGUS_SUBJ) {
		if (!file_err) {
		    if (fprintf(file, "Subject: %s\n",
				art_list[n_done + n_failed]->
				subject->subject) == EOF) {
			perror("knews: fprintf");
			file_err = True;
		    }
		}
	    }

	    buffer = server_read(&global.server);

	    if (what & SAVE_HEAD) {
		while (buffer && !IS_DOT(buffer) &&
		       !IS_SEP(buffer, what & SAVE_BODY)) {
		    if (!file_err) {
			if (fprintf(file, "%s\n", buffer) == EOF) {
			    perror("knews: fprintf");
			    file_err = True;
			}
		    }
		    buffer = server_read(&global.server);
		}

		if (!buffer)
		    return -1;
	    }

	    if (what & SAVE_BODY) {
		while (buffer && !IS_DOT(buffer)) {
		    if (!file_err) {
			if (fprintf(file, "%s\n", buffer) == EOF) {
			    perror("knews: fprintf");
			    file_err = True;
			}
		    }
		    buffer = server_read(&global.server);
		}

		if (!buffer)
		    return -1;
	    }

	    if (what & SAVE_EMPTY) {
		if (!file_err) {
		    if (fputc('\n', file) == EOF) {
			perror("knews: fputc");
			file_err = True;
		    }
		}
	    }

	    if (file_err)
		break;

	    n_done++;
	} else {
	    n_failed++;
	}

	sprintf(p, "%ld done.", n_done);
	if (n_failed > 0)
	    sprintf(p + strlen(p),
		    "   Failed to get %ld articles.", n_failed);
	set_message(message, False);
    }

    while (n_pending-- > 0) {
	int	status;

	buffer = server_read(&global.server);
	if (!buffer)
	    return -1;

	status = atoi(buffer);
	if (status == OK_ARTICLE || status == OK_HEAD || status == OK_BODY) {
	    buffer = server_read_chunk(&global.server);
	    if (!buffer)
		return -1;
	}
    }

    if (file_err)
	result[0] = -1;
    else {
	result[0] = n_done;
	result[1] = n_failed;
    }

    return 0;
}

static int get_what()
{
    Boolean	bogus_from = ToggleGet(save_widgets.bogus_from_toggle);
    Boolean	bogus_subj = ToggleGet(save_widgets.bogus_subj_toggle);
    Boolean	head       = ToggleGet(save_widgets.head_toggle);
    Boolean	body       = ToggleGet(save_widgets.body_toggle);
    Boolean	empty      = ToggleGet(save_widgets.empty_toggle);
    int		what = 0;

    if (bogus_from)
	what |= SAVE_BOGUS_FROM;
    if (bogus_subj)
	what |= SAVE_BOGUS_SUBJ;
    if (head)
	what |= SAVE_HEAD;
    if (body)
	what |= SAVE_BODY;
    if (empty)
	what |= SAVE_EMPTY;

    return what;
}

static void do_save(char *file_name, SaveScope scope, int what)
{
    FILE	*file;
    ARTICLE	**art_list = NULL;
    long	result[2];
    int		status, fclose_status;
    char	message[128];

    if (global.busy) return;

    if (global.mode != NewsModeGroup &&
	global.mode != NewsModeThread) {
	set_message("Not in a newsgroup!", True);
	return;
    } else if (!file_name || file_name[0] == '\0') {
	set_message("No file specified!", True);
	return;
    } else if (!(what & (SAVE_HEAD | SAVE_BODY)) ||
	       !(art_list = get_art_list(scope))) {
	set_message("Nothing to save!", True);
	return;
    }

    file = fopen_expand(file_name, "a");
    if (!file) {
	set_message("Failed to open file!", True);
	XtFree((char *)art_list);
	return;
    }

    sprintf(message, "Saving...   ");

    if (scope == SaveScopeArtwin) {
	status = ArtTextDumpToFile(main_widgets.text, file);
	result[0] = 1;
	result[1] = 0;
    } else {
	set_busy(True, True);
	status = save_to_file(file, message, art_list, what, result);
	unset_busy();
    }

    if ((fclose_status = fclose(file)) < 0)
	perror("knews: fclose");
    XtFree((char *)art_list);

    if (status < 0) {
	set_busy(True, True);
	(void)reconnect_server(True);
	unset_busy();
	return;
    } else if (result[0] < 0 || fclose_status < 0) {
	set_message("File error!", True);
	return;
    } else if (result[1] > 0) {
	sprintf(message, "%ld articles saved, failed to get %ld articles.",
		result[0], result[1]);
	set_message(message, True);
    } else {
	sprintf(message, "%ld articles saved.", result[0]);
	set_message(message, False);
    }

    if (scope == SaveScopeTagged)
	clear_tagged_articles();
}

void pipe_context_callback(void *data, int status, char *stderr_buf)
{
    char	message[128];
    char	*cmd = data;
    int		ok = False;

    if (strlen(cmd) > 100) cmd[100] = '\0';

    if (WIFEXITED(status)) {
	status = WEXITSTATUS(status);
	if (status == 0) {
	    sprintf(message, "'%s' exited ok.", cmd);
	    ok = True;
	} else if (status == 127)
	    sprintf(message, "Failed to start '%s'!", cmd);
	else
	    sprintf(message, "'%s' exited abnormally!", cmd);
    } else if (WIFSIGNALED(status)) {
	status = WTERMSIG(status);
	sprintf(message, "'%s' caught %s!",
		cmd, signal_string(status));
    } else
	sprintf(message, "Unknown problem with '%s'!", cmd);

    if (!ok)
	stderr_popup(stderr_buf, -1);

    set_message(message, !ok);

    XtFree(cmd);
}

static void do_pipe(char *command, SaveScope scope, int what)
{
    FILE	*file = NULL;
    char	*file_name;
    ARTICLE	**art_list = NULL;
    long	result[2];
    int		status, fflush_status;
    char	message[128];
    pid_t	pid;
    char	*temp;

    if (global.busy) return;

    if (global.mode != NewsModeGroup &&
	global.mode != NewsModeThread) {
	set_message("Not in a newsgroup!", True);
	return;
    } else if (!command || command[0] == '\0') {
	set_message("No command specified!", True);
	return;
    } else if (!(what & (SAVE_HEAD | SAVE_BODY)) ||
	       !(art_list = get_art_list(scope))) {
	set_message("Nothing to pipe!", True);
	return;
    }

    file_name = tmpnam(NULL);
    if (file_name) {
	file = fopen(file_name, "w+");
	(void)unlink(file_name);
    }

    if (!file) {
	set_message("Failed to open temporary file!", True);
	XtFree((char *)art_list);
	return;
    }

    sprintf(message, "Saving to temp file...   ");

    if (scope == SaveScopeArtwin) {
	status = ArtTextDumpToFile(main_widgets.text, file);
	result[0] = 1;
	result[1] = 0;
    } else {
	set_busy(True, True);
	status = save_to_file(file, message, art_list, what, result);
	unset_busy();
    }

    if ((fflush_status = fflush(file)) < 0)
	perror("knews: fflush");
    XtFree((char *)art_list);

    if (status < 0) {
	(void)fclose(file);
	set_busy(True, True);
	(void)reconnect_server(True);
	unset_busy();
	return;
    } else if (result[0] < 0 || fflush_status < 0) {
	(void)fclose(file);
	set_message("Error with temp file!", True);
	return;
    } else if (result[1] > 0) {
	sprintf(message,
		"%ld articles saved to temp file, failed to get %ld "
		"articles.  Pipe started.",
		result[0], result[1]);
	set_message(message, True);
    } else {
	sprintf(message,
		"%ld articles saved to temp file.  Pipe started.",
		result[0]);
	set_message(message, False);
    }

    if (scope == SaveScopeTagged)
	clear_tagged_articles();

    temp = XtNewString(command);
    pid = fork_nicely(temp, pipe_context_callback,
		      global.stderr_timeout >= 0);

    if (pid < 0) {
	perror("knews: fork");
	set_message("Fork failed!", True);
	XtFree(temp);
    } else if (pid == 0) { /* child */
	int	fd;

	fd = fileno(file);
	if (fd != STDIN_FILENO) {
	    fd = dup2(fd, STDIN_FILENO);
	    if (fd < 0) {
		perror("knews: dup2");
		_exit(127);
	    }
	}

	if (lseek(fd, SEEK_SET, 0) < 0) {
	    perror("knews: lseek");
	    _exit(127);
	}

	(void)execl("/bin/sh", "sh", "-c", command, (char *)NULL);
	perror("knews: execl");
	_exit(127);
    } else { /* parent */
	set_message("Pipe started.", False);
    }

    (void)fclose(file);
}

static void save_callback(Widget w,
			  XtPointer client_data,
			  XtPointer call_data)
{
    Arg		arg;
    char	*file_name = NULL;

    XtSetArg(arg, XtNbuffer, &file_name);
    XtGetValues(save_widgets.file_field, &arg, 1);

    do_save(file_name, save_scope, get_what());
}

static void pipe_callback(Widget w,
			  XtPointer client_data,
			  XtPointer call_data)
{
    Arg		arg;
    char	*command = NULL;

    XtSetArg(arg, XtNbuffer, &command);
    XtGetValues(save_widgets.shell_field, &arg, 1);

    do_pipe(command, save_scope, get_what());
}

static void unset_scope_toggle(SaveScope scope)
{
    Widget	w = NULL;

    switch (scope) {
    case SaveScopeArtwin:
	w = save_widgets.artwin_toggle;
	break;
    case SaveScopeArticle:
	w = save_widgets.article_toggle;
	break;
    case SaveScopeSubject:
	w = save_widgets.subject_toggle;
	break;
    case SaveScopeThread:
	w = save_widgets.thread_toggle;
	break;
    case SaveScopeSubthread:
	w = save_widgets.subthread_toggle;
	break;
    case SaveScopeTagged:
	w = save_widgets.tagged_toggle;
	break;
    }

    ToggleUnset(w);
}

static void set_toggles_sensitive(Boolean sens)
{
    XtSetSensitive(save_widgets.bogus_from_toggle, sens);
    XtSetSensitive(save_widgets.bogus_subj_toggle, sens);
    XtSetSensitive(save_widgets.head_toggle, sens);
    XtSetSensitive(save_widgets.body_toggle, sens);
    XtSetSensitive(save_widgets.empty_toggle, sens);
}

static void file_chooser_callback(Widget gw,
				  XtPointer client_data,
				  XtPointer call_data)
{
    char	*buffer = (char *)call_data;

    if (buffer) {
	Arg	arg;

	XtSetArg(arg, XtNbuffer, buffer);
	XtSetValues(save_widgets.file_field, &arg, 1);
	XtSetKeyboardFocus(main_widgets.save_shell,
			   save_widgets.file_field);
	XtSetArg(arg, XtNlabel, "Save");
	XtSetValues(save_widgets.ok_knapp, &arg, 1);
    }
}

static void choose_knapp_callback(Widget gw,
				  XtPointer client_data,
				  XtPointer call_data)
{
    if (!save_widgets.file_chooser) {
	save_widgets.file_chooser =
	    XtCreatePopupShell("filechooser", fileSelWidgetClass,
			       main_widgets.shell, NULL, 0);
	XtAddCallback(save_widgets.file_chooser, XtNchooseCallback,
		      file_chooser_callback, NULL);
    }

    XtPopup(save_widgets.file_chooser, XtGrabExclusive);
}

static void close_knapp_callback(Widget gw,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (!global.busy)
	XtPopdown(main_widgets.save_shell);
}

static void toggle_callback(Widget gw,
			    XtPointer client_data,
			    XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    *set = !*set;
}

static void scope_toggle_callback(Widget w,
				  XtPointer client_data,
				  XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    if (!*set) {
	unset_scope_toggle(save_scope);
	save_scope = (SaveScope)client_data;
	*set = True;
	if (w == save_widgets.artwin_toggle)
	    set_toggles_sensitive(False);
	else
	    set_toggles_sensitive(True);
    }
}

static void knapp_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    Arg		arg;
    char	*label = NULL;

    XtSetArg(arg, XtNlabel, &label);
    XtGetValues(w, &arg, 1);

    if (label) {
	if (label[0] == 'S')
	    save_callback(w, client_data, call_data);
	else
	    pipe_callback(w, client_data, call_data);
    }
}

static void focus_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    if (w == save_widgets.shell_field)
	KnappSetLabel(save_widgets.ok_knapp, "Pipe", True);
    else if (w == save_widgets.file_field)
	KnappSetLabel(save_widgets.ok_knapp, "Save", True);
}

static void tab_callback(Widget w,
			 XtPointer client_data,
			 XtPointer call_data)
{
    if (w == save_widgets.shell_field) {
	XtSetKeyboardFocus(main_widgets.save_shell,
			   save_widgets.file_field);
	KnappSetLabel(save_widgets.ok_knapp, "Save", True);
    } else if (w == save_widgets.file_field) {
	XtSetKeyboardFocus(main_widgets.save_shell,
			   save_widgets.shell_field);
	KnappSetLabel(save_widgets.ok_knapp, "Pipe", True);
    }
}

/*************************************************************************/

static void create_save_widgets()
{
    Arg		args[8];
    Widget	layout;

    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetArg(args[1], XtNinput, True);
    XtSetArg(args[2], XtNtitle, "knews: save/pipe");
    main_widgets.save_shell =
	XtCreatePopupShell("saveshell", topLevelShellWidgetClass,
			   main_widgets.shell, args, 3);

    save_widgets.file_chooser = NULL;

    layout =
	XtCreateManagedWidget("savelayout", layoutWidgetClass,
			      main_widgets.save_shell, NULL, 0);

    XtSetArg(args[0], XtNcenter, True);
    XtSetArg(args[1], XtNbuffer, "Save to file or pipe to shell");
    (void)XtCreateManagedWidget("message", messageWidgetClass,
				layout, args, 2);

    XtSetArg(args[0], XtNcenter, False);
    XtSetArg(args[1], XtNbuffer, "Shell-command:");
    save_widgets.shell_message =
	XtCreateManagedWidget("shellmessage", messageWidgetClass,
			      layout, args, 2);

    XtSetArg(args[0], XtNfocusRoot, main_widgets.save_shell);
    save_widgets.shell_field =
	XtCreateManagedWidget("shellfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(save_widgets.shell_field, XtNcallback,
		  pipe_callback, False);
    XtAddCallback(save_widgets.shell_field, XtNfocusCallback,
		  focus_callback, NULL);
    XtAddCallback(save_widgets.shell_field, XtNtabCallback,
		  tab_callback, NULL);

    XtSetArg(args[0], XtNcenter, False);
    XtSetArg(args[1], XtNbuffer, "File:");
    save_widgets.file_message =
	XtCreateManagedWidget("filemessage", messageWidgetClass,
			      layout, args, 2);

    XtSetArg(args[0], XtNfocusRoot, main_widgets.save_shell);
    save_widgets.file_field =
	XtCreateManagedWidget("filefield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(save_widgets.file_field, XtNcallback,
		  save_callback, NULL);
    XtAddCallback(save_widgets.file_field, XtNfocusCallback,
		  focus_callback, NULL);
    XtAddCallback(save_widgets.file_field, XtNtabCallback,
		  tab_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Save");
    XtSetArg(args[1], XtNresizable, False);
    save_widgets.ok_knapp =
	XtCreateManagedWidget("ok", knappWidgetClass, layout, args, 2);
    XtAddCallback(save_widgets.ok_knapp, XtNcallback,
		  knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Close");
    save_widgets.close_knapp =
	XtCreateManagedWidget("close", knappWidgetClass, layout, args, 1);
    XtAddCallback(save_widgets.close_knapp, XtNcallback,
		  close_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Choose");
    save_widgets.choose_knapp =
	XtCreateManagedWidget("choose", knappWidgetClass, layout, args, 1);
    XtAddCallback(save_widgets.choose_knapp, XtNcallback,
		  choose_knapp_callback, NULL);

    /***/

    XtSetArg(args[0], XtNlabel, "Bogus From header");
    XtSetArg(args[1], XtNset, True);
    save_widgets.bogus_from_toggle =
	XtCreateManagedWidget("bogusfrom", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.bogus_from_toggle, XtNcallback,
		  toggle_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Bogus Subject header");
    XtSetArg(args[1], XtNset, False);
    save_widgets.bogus_subj_toggle =
	XtCreateManagedWidget("bogussubj", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.bogus_subj_toggle, XtNcallback,
		  toggle_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Header");
    XtSetArg(args[1], XtNset, True);
    save_widgets.head_toggle =
	XtCreateManagedWidget("header", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.head_toggle, XtNcallback,
		  toggle_callback, NULL);


    XtSetArg(args[0], XtNlabel, "Body");
    XtSetArg(args[1], XtNset, True);
    save_widgets.body_toggle =
	XtCreateManagedWidget("body", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.body_toggle, XtNcallback,
		  toggle_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Empty line");
    XtSetArg(args[1], XtNset, True);
    save_widgets.empty_toggle =
	XtCreateManagedWidget("empty", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.empty_toggle, XtNcallback,
		  toggle_callback, NULL);

    /***/

    XtSetArg(args[0], XtNlabel, "Article win");
    XtSetArg(args[1], XtNset, False);
    save_widgets.artwin_toggle =
	XtCreateManagedWidget("artwin", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.artwin_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeArtwin);

    XtSetArg(args[0], XtNlabel, "Article");
    XtSetArg(args[1], XtNset, True);
    save_widgets.article_toggle =
	XtCreateManagedWidget("article", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.article_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeArticle);

    XtSetArg(args[0], XtNlabel, "Subject");
    XtSetArg(args[1], XtNset, False);
    save_widgets.subject_toggle =
	XtCreateManagedWidget("subject", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.subject_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeSubject);

    XtSetArg(args[0], XtNlabel, "Thread");
    XtSetArg(args[1], XtNset, False);
    save_widgets.thread_toggle =
	XtCreateManagedWidget("thread", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.thread_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeThread);

    XtSetArg(args[0], XtNlabel, "Subthread");
    XtSetArg(args[1], XtNset, False);
    save_widgets.subthread_toggle =
	XtCreateManagedWidget("subthread", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.subthread_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeSubthread);

    XtSetArg(args[0], XtNlabel, "Tagged articles");
    XtSetArg(args[1], XtNset, False);
    save_widgets.tagged_toggle =
	XtCreateManagedWidget("tagged", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(save_widgets.tagged_toggle, XtNcallback,
		  scope_toggle_callback, (XtPointer)SaveScopeTagged);

    XtSetKeyboardFocus(main_widgets.save_shell, save_widgets.file_field);
    XtRealizeWidget(main_widgets.save_shell);
    XtInstallAllAccelerators(main_widgets.save_shell,
			     main_widgets.save_shell);

    add_WM_DELETE_WINDOW_callback(main_widgets.save_shell,
				  close_knapp_callback, NULL);
}

static Boolean save_created = False;

void popup_save()
{
    if (!save_created) {
	create_save_widgets();
	save_created = True;
    }

    XtPopup(main_widgets.save_shell, XtGrabNone);	
}

void popdown_save()
{
    if (save_created)
	XtPopdown(main_widgets.save_shell);
}

static void do_action(Boolean save, String *params, Cardinal n)
{
    int		what = 0;
    char	*c;
    SaveScope	scope = SaveScopeArticle;

    if (n == 0) {
	if (save)
	    set_message("No filename specified in save/pipe action!",
			True);
	else
	    set_message("No shell command specified in save/pipe action!",
			True);
	return;
    } else if (n == 1) {
	set_message("To few arguments to save/pipe action!", True);
	return;
    }

    c = params[1];
    while (*c != '\0') {
	switch (*c++) {
	case 'f':
	case 'F':
	    what |= SAVE_BOGUS_FROM;
	    break;
	case 'S':
	case 's':
	    what |= SAVE_BOGUS_SUBJ;
	    break;
	case 'H':
	case 'h':
	    what |= SAVE_HEAD;
	    break;
	case 'B':
	case 'b':
	    what |= SAVE_BODY;
	    break;
	case 'E':
	case 'e':
	    what |= SAVE_EMPTY;
	    break;
	}
    }

    if (!(what & (SAVE_BODY|SAVE_HEAD)) ||
	!params[0] || params[0][0] == '\0') {
	if (save)
	    set_message("Nothing to save!", True);
	else
	    set_message("Nothing to pipe!", True);
	return;
    }

    if (n >= 3) {
	switch (params[2][0]) {
	case 'W':
	case 'w':
	    scope = SaveScopeArtwin;
	    break;
	case 'A':
	case 'a':
	    scope = SaveScopeArticle;
	    break;
	case 'S':
	case 's':
	    if ((params[2][1] == 'U' || params[2][1] == 'u') &&
		(params[2][2] == 'B' || params[2][2] == 'b')) {
		switch (params[2][3]) {
		case 'J':
		case 'j':
		    scope = SaveScopeSubject;
		    break;
		case 'T':
		case 't':
		    scope = SaveScopeSubthread;
		    break;
		}
	    }
	    break;
	case 'T':
	case 't':
	    switch (params[2][1]) {
	    case 'A':
	    case 'a':
		scope = SaveScopeTagged;
		break;
	    case 'H':
	    case 'h':
		scope = SaveScopeThread;
		break;
	    }
	    break;
	}
    }

    if (save)
	do_save(params[0], scope, what);
    else
	do_pipe(params[0], scope, what);
}

void action_save(Widget w, XEvent *string,
		 String *params, Cardinal *no_params)
{
    do_action(True, params, *no_params);
}

void action_pipe(Widget w, XEvent *string,
		 String *params, Cardinal *no_params)
{
    do_action(False, params, *no_params);
}
