/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include "codes.h"
#include "connect.h"
#include "newsrc.h"
#include "read.h"
#include "search.h"
#include "server.h"
#include "thread.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include <X11/Shell.h>
#include "../Widgets/Dialogue.h"
#include "../Widgets/FileSel.h"
#include "../Widgets/Knapp.h"
#include "../Widgets/Layout.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/Sash.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "../Widgets/TextField.h"
#include "../Widgets/Toggle.h"
#include "../Widgets/Util.h"

typedef enum {
    SearchModeRegexp,
    SearchModeXpat,
    SearchModeNewsgroups
} SearchMode;

static SearchMode search_mode = SearchModeRegexp;

static long	*xpats = NULL;
static long	n_xpats = 0;
static long	curr_xpat = -1;

static struct {
    Widget	regexp_field;
    Widget	wildcard_field;
    Widget	header_field;
    /***/
    Widget	from_toggle;
    Widget	subject_toggle;
    Widget	head_toggle;
    Widget	body_toggle;
    /***/
    Widget	search_knapp;
    Widget	submit_knapp;
    Widget	clear_knapp;
    Widget	next_knapp;
    Widget	first_knapp;
} search_widgets;

/*************************************************************************/

static void clear_xpat_data()
{
    XtFree((char *)xpats);
    xpats = NULL;
    n_xpats = 0;
    curr_xpat = -1;
}

static void set_xpat_message(int error)
{
    if (curr_xpat >= 0 && n_xpats > 0) {
	char	message[80];

	if (error)
	    sprintf(message, "Couldn't find match %ld", curr_xpat + 1);
	else
	    sprintf(message, "Match %ld of %ld", curr_xpat + 1, n_xpats);
	set_message(message, error);
    }
}

static ARTICLE *no_to_art(long no, int even_if_read)
{
    ARTICLE	*art;

    for (art = global.thr.articles ; art ; art = art->next)
	if (art->no == no) {
	    if (art->read && !even_if_read)
		return NULL;
	    else
		return art;
	}

    return NULL;
}

static void set_search_mode(SearchMode mode)
{
    search_mode = mode;

    if (mode != SearchModeXpat)
	clear_xpat_data();

    XtSetSensitive(search_widgets.header_field,
		   mode != SearchModeNewsgroups);
    XtSetSensitive(search_widgets.wildcard_field,
		   mode != SearchModeNewsgroups);

    XtSetSensitive(search_widgets.from_toggle,
		   mode == SearchModeRegexp);
    XtSetSensitive(search_widgets.subject_toggle,
		   mode == SearchModeRegexp);
    XtSetSensitive(search_widgets.head_toggle,
		   mode == SearchModeRegexp);
    XtSetSensitive(search_widgets.body_toggle,
		   mode == SearchModeRegexp);

    KnappSetLabel(search_widgets.search_knapp, NULL,
		  mode != SearchModeXpat);
    KnappSetLabel(search_widgets.submit_knapp, NULL,
		  mode == SearchModeXpat);
    KnappSetLabel(search_widgets.clear_knapp, NULL,
		   mode == SearchModeXpat && xpats != NULL);
    KnappSetLabel(search_widgets.next_knapp, NULL,
		  mode == SearchModeXpat &&
		  curr_xpat >= 0 && curr_xpat < n_xpats - 1);
    KnappSetLabel(search_widgets.first_knapp, NULL,
		  mode == SearchModeXpat &&
		  n_xpats > 0);
}

static int valid_header(char *header)
{
    for (;;) {
	unsigned char	c = *(unsigned char *)header++;

	if (c == '\0')
	    return True;

	if (!isalpha(c) && !isdigit(c) && c != '-')
	    return False;
    }

    /* not reached */
}

/*************************************************************************/

static void do_group_search(regex_t *re)
{
    long	i, n;

    i = ScrListGetFirstSelected(main_widgets.group_list);
    if (i < 0)
	i = 0;
    else
	ScrListSetSelected(main_widgets.group_list, i++, False);

    while (i < global.no_groups) {
	if (regexec(re, global.groups[i]->name, 0, NULL, 0) == 0)
	    break;
	i++;
    }

    if (i >= global.no_groups)
	set_message("No match!", True);
    else {
	global.thr.group = global.groups[i];
	ScrListSetSelected(main_widgets.group_list, i, True);
	ScrListSetFirst(main_widgets.group_list, i);
	ScrListGetFirstShownSize(main_widgets.group_list,
				 NULL, &n, NULL);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar,
			       global.no_groups, n, i);
    }
}

static void do_art_search(regex_t *re)
{
    ARTICLE	*art = NULL;
    SUBJECT	*subj = NULL;
    long	sel_data[3];
    long	*sel_p = NULL;
    Boolean	full_header = False;

    if (ToggleGet(search_widgets.subject_toggle)) {
	if (global.curr_subj) {
	    subj = global.curr_subj->next;
	    if (!subj)
		subj = global.thr.subjects;
	} else {
	    subj = global.thr.subjects;
	}

	while (subj) {
	    if (subj->disp >= 0)
		if (regexec(re, subj->subject, 0, NULL, 0) == 0)
		    break;

	    subj = subj->next;
	}
    } else {
	art = global.curr_art;
	if (art) {
	    subj = art->subject;
	    art = next_in_thread_preorder(art);
	    if (!art) {
		do {
		    subj = subj->next;
		} while (subj && subj->prev &&
			 subj->prev->thread == subj->thread);

		if (subj)
		    art = subj->thread;
		else {
		    subj = global.thr.subjects;
		    if (subj)
			art = subj->thread;
		    else
			art = NULL;
		}
	    }
	} else {
	    subj = global.curr_subj;
	    if (subj)
		subj = subj->next;

	    if (subj)
		art = subj->thread;
	    else {
		subj = global.thr.subjects;
		if (subj)
		    art = subj->thread;
		else
		    art = NULL;
	    }
	}

	if (ToggleGet(search_widgets.from_toggle)) {
	    while (subj) {
		while (art) {
		    if (art->from && !art->read)
			if (regexec(re, art->from, 0, NULL, 0) == 0)
			    break;

		    art = next_in_thread_preorder(art);
		}
		if (art)
		    break;

		do {
		    subj = subj->next;
		} while (subj && subj->prev &&
			 subj->prev->thread == subj->thread);
		if (subj)
		    art = subj->thread;
	    }
	} else {
	    Boolean	head = ToggleGet(search_widgets.head_toggle);
	    Boolean 	body = ToggleGet(search_widgets.body_toggle);
	    regmatch_t	pmatch[1];
	    long	line;
	    char	command[1024];
	    char	*reply = "STFU"; /* Shut gcc up */

	    if (!head && !body) {
		if (global.bell)
		    XBell(XtDisplay(main_widgets.shell), 0);
		return;
	    }

	    set_message("Searching...", False);
	    set_busy(True, True);

	    while (subj) {
		while (art) {
		    if (art->from && !art->read) {
			if (head) {
			    if (body)
				sprintf(command, "ARTICLE %ld\r\n", art->no);
			    else
				sprintf(command, "HEAD %ld\r\n", art->no);
			    reply = server_comm(&global.server,
					       command, True);
			    
			    if (!reply) {
				(void)reconnect_server(True);
				unset_busy();
				return;
			    }

			    if (atoi(reply) == OK_HEAD ||
				atoi(reply) == OK_ARTICLE) {
				int	found = False;

				line = -1;
				reply = server_read(&global.server);
				while (reply &&
				       (!body || reply[0] != '\0') &&
				       (reply[0] != '.' ||
					reply[1] != '\0')) {
				    if (!found &&
					regexec(re, reply, 1,
						pmatch, 0) == 0) {
					sel_p = sel_data;
					sel_data[0] = line;
					sel_data[1] = pmatch[0].rm_so;
					sel_data[2] = pmatch[0].rm_eo - 1;
					full_header = True;
					found = True;
				    }
				    line--;
				    reply = server_read(&global.server);
				}

				if (!reply) {
				    (void)reconnect_server(True);
				    unset_busy();
				    return;
				}

				if (found) {
				    if (body) {
					while (reply &&
					       (reply[0] != '.' ||
						reply[1] != '\0'))
					    reply =
						server_read(&global.server);

					if (!reply) {
					    (void)reconnect_server(True);
					    unset_busy();
					    return;
					}
				    }
				    break;
				}
			    }
			}

			if (body) {
			    if (!head) {
				sprintf(command, "BODY %ld\r\n", art->no);
				reply = server_comm(&global.server,
						   command, True);

				if (!reply) {
				    (void)reconnect_server(True);
				    unset_busy();
				    return;
				}
			    }

			    if (head || atoi(reply) == OK_BODY) {
				int	found = False;
				
				line = 0;
				reply = server_read(&global.server);
				while (reply &&
				       (reply[0] != '.' ||
					reply[1] != '\0')) {
				    if (!found &&
					regexec(re, reply, 1,
						pmatch, 0) == 0) {
					sel_p = sel_data;
					sel_data[0] = line;
					sel_data[1] = pmatch[0].rm_so;
					sel_data[2] = pmatch[0].rm_eo - 1;
					full_header = True;
					found = True;
				    }
				    line++;
				    reply = server_read(&global.server);
				}
				
				if (!reply) {
				    (void)reconnect_server(True);
				    unset_busy();
				    return;
				}

				if (found)
				    break;
			    }
			}
		    }
		    art = next_in_thread_preorder(art);
		}
		if (art)
		    break;

		do {
		    subj = subj->next;
		} while (subj && subj->prev &&
			 subj->prev->thread == subj->thread);
		if (subj)
		    art = subj->thread;
	    }

	    unset_busy();
	}
    }

    if (art) {
	set_curr_art(art, True, True);
	if (global.mode == NewsModeThread)
	    setNewsModeThread();
	(void)read_article(art, full_header, sel_p);
    } else if (subj) { /* Subject search */
	global.curr_art = NULL;
	if (global.mode == NewsModeThread) {
	    global.curr_subj = subj;
	    setNewsModeGroup(False);
	} else {
	    set_curr_subj(subj);
	}
    } else {
	set_message("No matches!", True);
    }
}

static void do_xpat_search(char *header, char *wildcard)
{
    char	command[512];
    char	*buffer, *c;
    long	n_alloc = 0;

    clear_xpat_data();
    KnappSetLabel(search_widgets.first_knapp, NULL, False);
    KnappSetLabel(search_widgets.clear_knapp, NULL, False);
    KnappSetLabel(search_widgets.next_knapp, NULL, False);

    if (strlen(header) + strlen(wildcard) >
	sizeof(command) - 64) {
	set_message("Header or wildcard too long!", True);
	return;
    }

    if (!global.thr.group ||
	global.thr.group->first_art > global.thr.group->last_art) {
	set_message("No articles!", True);
	return;
    }

    sprintf(command, "XPAT %s %ld-%ld %s\r\n",
	    header,
	    global.thr.group->first_art,
	    global.thr.group->last_art,
	    wildcard);

    set_busy(True, True);
    set_message("Server contacted, waiting for response...", False);
    buffer = server_comm(&global.server, command, True);
    if (!buffer) {
	(void)reconnect_server(True);
	unset_busy();
	return;
    }

    if (atoi(buffer) != OK_HEAD) {
	unset_busy();
	if (strlen(buffer) > 100)
	    buffer[100] = '\0';
	sprintf(command, "Error!  Failed message from server is: %s",
		buffer);
	set_message(command, True);
	return;
    }

    sprintf(command, "Number of matches:  ");
    c = command + strlen(command);
    n_xpats = 0;

    buffer = server_read(&global.server);
    while (buffer && (buffer[0] != '.' || buffer [1] != '\0')) {
	long	no;

	if (sscanf(buffer, "%ld", &no) == 1 &&
	    no_to_art(no, False)) {
	    if (n_xpats > n_alloc - 2) {
		n_alloc = 2 * (n_alloc + 1);
		xpats =
		    (long *)XtRealloc((char *)xpats,
				      n_alloc * sizeof(long));
	    }

	    xpats[n_xpats++] = no;
	    sprintf(c, "%ld", n_xpats);
	    set_message(command, False);
	}

	buffer = server_read(&global.server);
    }

    if (!buffer) {
	clear_xpat_data();
	(void)reconnect_server(True);
	unset_busy();
	return;
    }

    unset_busy();

    if (n_xpats == 0)
	set_message("No matches!", True);
    else {
	ARTICLE	*art = no_to_art(xpats[0], True);

	curr_xpat = 0;

	set_curr_art(art, True, True);
	if (global.mode == NewsModeThread)
	    setNewsModeThread();
	(void)read_article(art, False, NULL);
	set_xpat_message(art == NULL);
	KnappSetLabel(search_widgets.submit_knapp, NULL, False);
	KnappSetLabel(search_widgets.first_knapp, NULL, True);
	KnappSetLabel(search_widgets.next_knapp, NULL, n_xpats > 1);
	KnappSetLabel(search_widgets.clear_knapp, NULL, True);
    }
}

static void submit_knapp_callback(Widget w,
				  XtPointer client_data,
				  XtPointer call_data)
{

    if (global.busy) return;

    if (search_mode == SearchModeXpat) {
	char	*header = NULL;
	char	*wildcard = NULL;
	Arg	arg;

	if (global.mode != NewsModeGroup &&
	    global.mode != NewsModeThread)
	    return;
	
	XtSetArg(arg, XtNbuffer, &header);
	XtGetValues(search_widgets.header_field, &arg, 1);
	XtSetArg(arg, XtNbuffer, &wildcard);
	XtGetValues(search_widgets.wildcard_field, &arg, 1);

	if (header && wildcard) {
	    if (valid_header(header))
		do_xpat_search(header, wildcard);
	    else
		set_message("Invalid header name!", True);
	} else {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	}
    }
}

static void search_knapp_callback(Widget w,
				  XtPointer client_data,
				  XtPointer call_data)
{
    Arg		arg;
    char	*buffer = NULL;
    regex_t	re;

    if (global.busy ||
	global.mode == SearchModeXpat)
	return;

    XtSetArg(arg, XtNbuffer, &buffer);
    XtGetValues(search_widgets.regexp_field, &arg, 1);
    if (!buffer) buffer ="";

    if (regcomp(&re, buffer, REG_ICASE | REG_EXTENDED) != 0) {
	NoticeCreateAndPopup("searchnotice", main_widgets.shell,
			     "Parse error in regexp!",
			     "Ok", NULL, NULL, 2000,
			     NULL, NULL, XtGrabNone);
	regfree(&re);
	return;
    }

    if (search_mode == SearchModeNewsgroups)
	do_group_search(&re);
    else
	do_art_search(&re);

    regfree(&re);
}

static void close_knapp_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (!global.busy)
	popdown_search();
}

static void clear_knapp_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (search_mode == SearchModeXpat) {
	clear_xpat_data();
	KnappSetLabel(search_widgets.submit_knapp, NULL, True);
	KnappSetLabel(search_widgets.clear_knapp, NULL, False);
	KnappSetLabel(search_widgets.next_knapp, NULL, False);
	KnappSetLabel(search_widgets.first_knapp, NULL, False);
    }
}

static void first_knapp_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    if (search_mode == SearchModeXpat &&
	n_xpats > 0) {
	ARTICLE	*art = no_to_art(xpats[0], True);

	curr_xpat = 0;

	set_curr_art(art, True, True);
	if (global.mode == NewsModeThread)
	    setNewsModeThread();
	(void)read_article(art, False, NULL);
	set_xpat_message(art == NULL);
	KnappSetLabel(search_widgets.next_knapp, NULL, n_xpats > 1);
    }
}

static void next_knapp_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    if (search_mode == SearchModeXpat &&
	curr_xpat >= 0 && curr_xpat + 1 < n_xpats) {
	ARTICLE	*art = no_to_art(xpats[++curr_xpat], True);

	set_curr_art(art, True, True);
	if (global.mode == NewsModeThread)
	    setNewsModeThread();
	(void)read_article(art, False, NULL);
	set_xpat_message(art == NULL);
	KnappSetLabel(search_widgets.next_knapp, NULL,
		      curr_xpat + 1 < n_xpats);
    }
}

static void from_toggle_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    if (set) {
	*set = !*set;
	if (*set) {
	    ToggleUnset(search_widgets.subject_toggle);
	    ToggleUnset(search_widgets.head_toggle);
	    ToggleUnset(search_widgets.body_toggle);
	}
    }
}

static void subject_toggle_callback(Widget w,
				    XtPointer client_data,
				    XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    if (set) {
	*set = !*set;
	if (*set) {
	    ToggleUnset(search_widgets.from_toggle);
	    ToggleUnset(search_widgets.head_toggle);
	    ToggleUnset(search_widgets.body_toggle);
	}
    }
}

static void head_toggle_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    if (set) {
	*set = !*set;
	if (*set) {
	    ToggleUnset(search_widgets.from_toggle);
	    ToggleUnset(search_widgets.subject_toggle);
	}
    }
}

static void body_toggle_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    Boolean	*set = (Boolean *)call_data;

    if (set) {
	*set = !*set;
	if (*set) {
	    ToggleUnset(search_widgets.from_toggle);
	    ToggleUnset(search_widgets.subject_toggle);
	}
    }
}

static void xpat_field_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    if (w == search_widgets.wildcard_field)
	XtSetKeyboardFocus(main_widgets.search_shell,
			   search_widgets.header_field);
    else
	XtSetKeyboardFocus(main_widgets.search_shell,
			   search_widgets.wildcard_field);
}

static void focus_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    if (global.mode == NewsModeAllgroups)
	set_search_mode(SearchModeNewsgroups);
    else if (w == search_widgets.regexp_field)
	set_search_mode(SearchModeRegexp);
    else
	set_search_mode(SearchModeXpat);
}

/*************************************************************************/
			   
static void create_search_widgets()
{
    Arg		args[8];
    Widget	layout;
    Widget	temp;

    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetArg(args[1], XtNinput, True);
    XtSetArg(args[2], XtNtitle, "knews: search");
    main_widgets.search_shell =
	XtCreatePopupShell("searchshell", topLevelShellWidgetClass,
			   main_widgets.shell, args, 3);

    layout =
	XtCreateManagedWidget("searchlayout", layoutWidgetClass,
			      main_widgets.search_shell, NULL, 0);

    XtSetArg(args[0], XtNbuffer, "Regular expression searching");
    (void)XtCreateManagedWidget("regexptitle", messageWidgetClass,
				layout, args, 1);

    XtSetArg(args[0], XtNbuffer, "XPAT wildcard searching");
    (void)XtCreateManagedWidget("xpattitle", messageWidgetClass,
				layout, args, 1);

    (void)XtCreateManagedWidget("sash", sashWidgetClass,
				layout, NULL, 0);

    XtSetArg(args[0], XtNcenter, False);
    XtSetArg(args[1], XtNbuffer, "Regexp:");
    (void)XtCreateManagedWidget("regexpmessage", messageWidgetClass,
				layout, args, 2);

    XtSetArg(args[1], XtNbuffer, "Wildcard:");
    (void)XtCreateManagedWidget("wildcardmessage", messageWidgetClass,
				layout, args, 2);

    XtSetArg(args[1], XtNbuffer, "Header:");
    (void)XtCreateManagedWidget("headermessage", messageWidgetClass,
				layout, args, 2);

    XtSetArg(args[0], XtNfocusRoot, main_widgets.search_shell);
    search_widgets.regexp_field =
	XtCreateManagedWidget("regexpfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.regexp_field, XtNcallback,
		  search_knapp_callback, NULL);
    XtAddCallback(search_widgets.regexp_field, XtNfocusCallback,
		  focus_callback, NULL);

    search_widgets.wildcard_field =
	XtCreateManagedWidget("wildcardfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.wildcard_field, XtNcallback,
		  xpat_field_callback, NULL);
    XtAddCallback(search_widgets.wildcard_field, XtNtabCallback,
		  xpat_field_callback, NULL);
    XtAddCallback(search_widgets.wildcard_field, XtNfocusCallback,
		  focus_callback, NULL);

    search_widgets.header_field =
	XtCreateManagedWidget("headerfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.header_field, XtNcallback,
		  xpat_field_callback, NULL);
    XtAddCallback(search_widgets.header_field, XtNtabCallback,
		  xpat_field_callback, NULL);
    XtAddCallback(search_widgets.header_field, XtNfocusCallback,
		  focus_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Search");
    search_widgets.search_knapp =
	XtCreateManagedWidget("search", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.search_knapp, XtNcallback,
		  search_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Submit");
    search_widgets.submit_knapp =
	XtCreateManagedWidget("submit", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.submit_knapp, XtNcallback,
		  submit_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Clear");
    search_widgets.clear_knapp =
	XtCreateManagedWidget("clear", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.clear_knapp, XtNcallback,
		  clear_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Next");
    search_widgets.next_knapp =
	XtCreateManagedWidget("next", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.next_knapp, XtNcallback,
		  next_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "First");
    search_widgets.first_knapp =
	XtCreateManagedWidget("first", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(search_widgets.first_knapp, XtNcallback,
		  first_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Close");
    temp = XtCreateManagedWidget("close", knappWidgetClass,
				 layout, args, 1);
    XtAddCallback(temp, XtNcallback, close_knapp_callback, NULL);

    XtSetArg(args[0], XtNset, True);
    XtSetArg(args[1], XtNlabel, "'From:' header");
    search_widgets.from_toggle =
	XtCreateManagedWidget("fromtoggle", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(search_widgets.from_toggle, XtNcallback,
		  from_toggle_callback, NULL);

    XtSetArg(args[0], XtNset, False);
    XtSetArg(args[1], XtNlabel, "'Subject:' header");
    search_widgets.subject_toggle =
	XtCreateManagedWidget("subjecttoggle", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(search_widgets.subject_toggle, XtNcallback,
		  subject_toggle_callback, NULL);

    XtSetArg(args[0], XtNset, False);
    XtSetArg(args[1], XtNlabel, "Article head");
    search_widgets.head_toggle =
	XtCreateManagedWidget("headtoggle", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(search_widgets.head_toggle, XtNcallback,
		  head_toggle_callback, NULL);

    XtSetArg(args[0], XtNset, False);
    XtSetArg(args[1], XtNlabel, "Article body");
    search_widgets.body_toggle =
	XtCreateManagedWidget("bodytoggle", toggleWidgetClass,
			      layout, args, 2);
    XtAddCallback(search_widgets.body_toggle, XtNcallback,
		  body_toggle_callback, NULL);

    XtSetKeyboardFocus(main_widgets.search_shell,
		       search_widgets.regexp_field);
    XtRealizeWidget(main_widgets.search_shell);
    XtInstallAllAccelerators(main_widgets.search_shell,
			     main_widgets.search_shell);

    add_WM_DELETE_WINDOW_callback(main_widgets.search_shell,
				  close_knapp_callback, NULL);
}

static Boolean search_created = False;

void popup_search()
{
    if (!search_created) {
	create_search_widgets();
	search_created = True;
	set_search_mode(SearchModeRegexp);
    }

    if (global.mode == NewsModeAllgroups)
	set_search_mode(SearchModeNewsgroups);
    else if (search_mode == SearchModeNewsgroups)
	set_search_mode(SearchModeRegexp);

    XtPopup(main_widgets.search_shell, XtGrabNone);	
}

void popdown_search()
{
    if (search_created) {
	if (search_mode == SearchModeXpat) {
	    clear_xpat_data();
	    set_search_mode(SearchModeXpat);
	}
	XtPopdown(main_widgets.search_shell);
    }
}

/*************************************************************************/

static Widget find_group_widget = NULL;

static long getnmatching(char *c1, char *c2)
{
    long	n = 0;

    while (*c1 != '\0' && *c2 != '\0' && *c1 == *c2) {
	c1++;
	c2++;
	n++;
    }

    return n;
}

static void find_group_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    DialogueReport	*report = (DialogueReport *)call_data;
    Arg			arg;

    if (global.busy || !report ||
	(global.mode != NewsModeConnected &&
	 global.mode != NewsModeAllgroups))
	return;

    switch (report->reply) {
    case DialogueReplyEnter: /* find group */
    case DialogueReplyLeft:
	if (!report->buffer || report->buffer[0] == '\0')
	    XtPopdown(w);
	else if (!report->buffer || strlen(report->buffer) > 500) {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	} else {
	    char	command[512];
	    char	*reply;

	    set_busy(True, True);
	    set_message("Asking server about group...", False);
	    sprintf(command, "GROUP %s\r\n", report->buffer);
	    reply = server_comm(&global.server, command, True);
	    if (!reply) {
		(void)reconnect_server(True);
		unset_busy();
		break;
	    }
	    unset_busy();

	    if (atoi(reply) == OK_GROUP) {
		XtPopdown(w);

		global.thr.group = find_group(report->buffer);
		if (!global.thr.group) {
		    global.thr.group = create_group(report->buffer);
		    sort_groups();
		}

		if (global.mode == NewsModeAllgroups)
		    setNewsModeAllgroups(); /* if we're aborted later */
		read_group(reply);
	    } else {
		if (strlen(reply) > 80)
		    reply[80] = '\0';
		sprintf(command, "Error!  Message from server is: %s", reply);
		set_message(command, True);
	    }
	}
	break;
    case DialogueReplyMiddle: /* clear */
	XtSetArg(arg, XtNbuffer, "");
	XtSetValues(w, &arg, 1);
	break;
    case DialogueReplyTab: /* complete */
	if (report->buffer && report->buffer[0] != '\0') {
	    long	len = strlen(report->buffer);
	    long	first, last, n;
	    char	*c;

	    for (first = 0 ; first < global.no_groups ; first++)
		if (!global.groups[first]->subscribed)
		    break;

	    while (first < global.no_groups) {
		if (strncmp(global.groups[first]->name,
			    report->buffer, len) == 0)
		    break;
		first++;
	    }

	    for (last = first + 1 ; last < global.no_groups ; last++)
		if (strncmp(global.groups[last]->name,
			    report->buffer, len) != 0)
		    break;

	    if (first == global.no_groups) {
		c = NULL;
		n = 0;
	    } else {
		c = global.groups[first]->name;
		n = getnmatching(c, global.groups[last-1]->name);
	    }

	    for (first = 0 ; first < global.no_groups ; first++)
		if (!global.groups[first]->subscribed)
		    break;
		else if (strncmp(global.groups[first]->name,
				 report->buffer, len) == 0) {
		    if (c) {
			long	temp;

			temp = getnmatching(c, global.groups[first]->name);
			if (temp < n)
			    n = temp;
		    } else {
			c = global.groups[first]->name;
			n = strlen(c);
		    }
		}

	    if (!c || n > 500) {
		if (global.bell)
		    XBell(XtDisplay(w), 0);
	    } else {
		char	buffer[512];

		memcpy(buffer, c, n);
		buffer[n] = '\0';
		XtSetArg(arg, XtNbuffer, buffer);
		XtSetValues(w, &arg, 1);
	    }
	}
	break;
    case DialogueReplyClose: /* cancel */
    case DialogueReplyRight:
	XtPopdown(w);
	break;
    }
}

void popup_find_group(int x, int y)
{
    if (global.busy ||
	(global.mode != NewsModeConnected &&
	 global.mode != NewsModeAllgroups))
	return;

    if (!find_group_widget) {
	Arg	args[6];

	XtSetArg(args[0], XtNleftLabel, "Goto group");
	XtSetArg(args[1], XtNmiddleLabel, "Clear");
	XtSetArg(args[2], XtNrightLabel, "Cancel");
	XtSetArg(args[3], XtNmessage, "Find group by name");
	XtSetArg(args[4], XtNx, x);
	XtSetArg(args[5], XtNy, y);

	find_group_widget =
	    XtCreatePopupShell("findgroup", dialogueWidgetClass,
			       main_widgets.shell, args, 6);
	XtAddCallback(find_group_widget, XtNcallback,
		      find_group_callback, NULL);
	XtRealizeWidget(find_group_widget);
    }

    XtPopup(find_group_widget, XtGrabNone);
}

void popdown_find_group()
{
    if (find_group_widget)
	XtPopdown(find_group_widget);
}
