/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "server.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Message.h"

#ifndef INADDR_NONE
#define INADDR_NONE (0xffffffff)    /* a hack, hope it works */
#endif

#ifndef USE_POLL
#  if defined(SYSV) && !defined(__hpux)
#    define USE_POLL 1
#  else
#    define USE_POLL 0
#  endif
#endif

#if USE_POLL
#include <sys/poll.h>
#endif

#define BUFFERLEN (8192 - 3)

int disp_conn_num;

int do_wait(int *fd, int rd, XtInputMask mask)
{
#if USE_POLL
    struct pollfd	fds[2];
#else
    fd_set		read_fds;
    fd_set		write_fds;
    int			maxfdp1;

    maxfdp1 = (*fd > disp_conn_num ? *fd : disp_conn_num) + 1;
#endif

    for (;;) {
	while (XtAppPending(global.app_cont) & mask) {
	    XtAppProcessEvent(global.app_cont, mask);
	    if (*fd < 0)
		return -1;
	}

#if USE_POLL
	fds[0].fd = disp_conn_num;
	fds[0].events = POLLIN;
	fds[0].revents = 0;
	fds[1].fd = *fd;
	if (rd)
	    fds[1].events = POLLIN;
	else
	    fds[1].events = POLLOUT;
	fds[1].revents = 0;

	while (poll(fds, 2, -1) < 0)
	    if (errno != EINTR) {
		perror("knews: poll");
		return -1;
	    }

	if (fds[1].revents)
	    return 0;
	if (fds[0].revents) {
	    XtAppProcessEvent(global.app_cont, mask);
	    if (*fd < 0)
		return -1;
	}
#else
	FD_ZERO(&read_fds);
	FD_ZERO(&write_fds);
	FD_SET(disp_conn_num, &read_fds);
	if (rd)
	    FD_SET(*fd, &read_fds);
	else
	    FD_SET(*fd, &write_fds);

	while (select(maxfdp1,
#ifdef __hpux  /* don't ask */
		      (int *)&read_fds, rd ? NULL : (int *)&write_fds,
#else
		      &read_fds, rd ? NULL : &write_fds,
#endif
		      NULL, NULL) < 0)
	    if (errno != EINTR) {
		perror("knews: select");
		return -1;
	    }

	if (FD_ISSET(*fd, &read_fds) ||
	    FD_ISSET(*fd, &write_fds))
	    return 0;
	if (FD_ISSET(disp_conn_num, &read_fds)) {
	    XtAppProcessEvent(global.app_cont, mask);
	    if (*fd < 0)
		return -1;
	}
#endif
    }
    /* not reached */
}

/*  tell != 0   means tell about errors
 *  tell >  1   means say what you're doing
 */
int server_open(SERVER *server, char *host, int tell)
{
    static struct sockaddr_in	sin;
    long	flags;
    long	temp;
    int		port;
    int		alive = 1;

    if (server->fd >= 0)
	(void)server_close(server);

    server->fd = -1;
    server->buffer = NULL;
    server->curr_pos = 0;
    server->len = 0;
    server->cookie = NULL;

    if (tell > 1)
	set_message("Looking up hostname...", False);

    if (host) {
	char		*c;
	unsigned long	l;

	port = 0;
	c = strchr(host, ':');
	if (c) {
	    *c = '\0';
	    if (c[1] >= '0' && c[1] <= '9')
		port = atoi(c + 1);
	}

	if (port != 0)
	    port = htons(port);
	else {
	    struct servent	*service;

	    service = getservbyname("nntp", "tcp");
	    if (!service) {
		if (tell)
		    set_message("Error: tcp/nntp: unknown service!", True);
		if (c) *c = ':';
		return -1;
	    }
	    port = service->s_port;
	}

	(void)memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = port;

	if ((l = inet_addr(host)) != INADDR_NONE) {
	    (void)memcpy(&sin.sin_addr, &l, sizeof(l));
	    if (c) *c = ':';
	} else {
	    struct hostent	*hp;

	    hp = gethostbyname(host);
	    if (c) *c = ':';
	    if (!hp) {
		if (tell)
		    set_message("Error: No such host!", True);
		return -1;
	    }
	    sin.sin_family = hp->h_addrtype;
	    (void)memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);
	}
    } /* else same as last time */

    while ((server->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 &&
	   errno == EINTR);
    if (server->fd < 0) {
	perror("knews: socket");
	if (tell)
	    set_message("Error: Failed to create socket!", True);
	return -1;
    }

    if ( (flags = fcntl(server->fd, F_GETFL)) < 0 ||
	 fcntl(server->fd, F_SETFL, flags|O_NONBLOCK) < 0) {
	perror("knews: fcntl(F_GETFL)");
	if (tell)
	    set_message("Error: Failed to set nonblocking I/O!", True);
	return -1;
    }

    while ((temp = setsockopt(server->fd, SOL_SOCKET, SO_KEEPALIVE,
			      (char *)&alive, sizeof(alive))) < 0 &&
	   errno == EINTR);
    if (temp < 0) {
	perror("knews: setsockopt");
	if (tell)
	    set_message("Failed to set socket keepalive!", True);
	return -1;
    }

    if (tell > 1)
	set_message("Server contacted, waiting for response...", False);

    while ((temp = connect(server->fd, (struct sockaddr *)&sin,
			   sizeof(sin))) < 0 &&
	   errno == EINTR);

    server->len = BUFFERLEN;
    server->buffer = XtMalloc(BUFFERLEN + 3);
    server->buffer[0] = '\0';

    if (temp < 0 &&
	(errno == EINPROGRESS ||
	 errno == EAGAIN ||
	 errno == EWOULDBLOCK)) {
	if (host) {
	    temp = 0;
	    (void)do_wait(&server->fd, True, XtIMXEvent);
	    if (server->fd < 0)
		return -1;
	} else {
	    return 0; /* don't block */
	}
    }

    server->buffer[0] = '\0';
    if (temp >= 0)
	while ((temp = read(server->fd, server->buffer, server->len)) < 0 &&
	       errno == EINTR);

    if (temp >= 0) {
	server->buffer[temp] = '\0';
	return 0;
    }

    switch (errno) {
    case EINPROGRESS:
    case EAGAIN:
#if EWOULDBLOCK != EAGAIN
    case EWOULDBLOCK:
#endif
	return 0;
    case ECONNREFUSED:
	if (tell)
	    set_message("Error: Connection refused!", True);
	else
	    perror("knews: read");
	break;
    case ENETUNREACH:
	if (tell)
	    set_message("Error: Network unreachable!", True);
	else
	    perror("knews: read");
	break;
    case ETIMEDOUT:
	if (tell)
	    set_message("Error: Connection timed out!", True);
	else
	    perror("knews: read");
	break;
    case EADDRNOTAVAIL:
	if (tell)
	    set_message("Error: Address not available!", True);
	else
	    perror("knews: read");
	break;
    default:
	perror("knews: read");
	if (tell)
	    set_message("Error: Connection failed!", True);
	break;
    }

    return -1;
}

static void server_context_callback(void *data, int status, char *stderr_buf)
{
    /* for debugging purposes */
}

int server_fork(SERVER *server, char *command, int tell)
{
#if 0
    if (tell)
	set_message("Forking not supported!", True);
    return -1;
#else
    static char	*cmd = NULL;
    pid_t	pid;
    int		fd[2];
    long	flags;

    if (command)
	cmd = command;

    if (server->fd >= 0)
	server_close(server);

    server->fd = -1;
    server->buffer = NULL;
    server->curr_pos = 0;
    server->len = 0;
    server->cookie = NULL;

#if 1
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd) < 0) {
	perror("knews: socketpair");
	if (tell)
	    set_message("Error!  socketpair() failed!", True);
	return -1;
    }
#else
    if (pipe(fd) < 0) {
	perror("knews: pipe");
	if (tell)
	    set_message("Error!  pipe() failed!", True);
	return -1;
    }
#endif

    if ( (flags = fcntl(fd[0], F_GETFL)) < 0 ||
	 fcntl(fd[0], F_SETFL, flags|O_NONBLOCK) < 0) {
	perror("knews: fcntl");
	(void)close(fd[0]);
	(void)close(fd[1]);
	if (tell)
	    set_message("Error: Failed to set nonblocking I/O!", True);
	return -1;
    }

    pid = fork_nicely(NULL, server_context_callback, False);
    if (pid < 0) {
	(void)close(fd[0]);
	(void)close(fd[1]);
	if (tell)
	    set_message("Error!  Fork failed!", True);
	return -1;
    }

    if (pid == 0) {
	char	*c = CODE_TO_STR(ERR_FAULT) " This ain't happening, man!\r\n";

	(void)close(fd[0]);
	if (fd[1] != STDIN_FILENO) {
	    if (dup2(fd[1], STDIN_FILENO) != STDIN_FILENO) {
		perror("knews: dup2");
		write(fd[1], c, strlen(c));
		_exit(0);
	    }
	    (void)close(fd[1]);
	}
	if (dup2(STDIN_FILENO, STDOUT_FILENO) != STDOUT_FILENO) {
	    perror("knews: dup2");
	    write(STDIN_FILENO, c, strlen(c));
	    _exit(0);
	}

	execl("/bin/sh", "sh", "-c", command, (char *)0);
	perror("knews: execl");
	write(STDOUT_FILENO, c, strlen(c));
	_exit(0);
    }

    (void)close(fd[1]);
    server->fd = fd[0];
    server->buffer = XtMalloc(BUFFERLEN + 3);
    server->len = BUFFERLEN;
    server->buffer[0] = '\0';

    return 0;
#endif
}

int server_close(SERVER *server)
{
    if (server->buffer) {
	XtFree(server->buffer);
	server->buffer = NULL;
	server->len = 0;
	server->curr_pos = 0;
    }

    if (server->cookie) {
	(void)fclose(server->cookie);
	server->cookie = NULL;
    }

    if (server->fd >= 0) {
	if (close(server->fd) < 0)
	    perror("knews: close");
	server->fd = -1;
	return 1;
    }

    return 0;
}

int server_write_raw(SERVER *server, char *message)
{
    long	n, i;

    if (server->fd < 0)
	return -1;

    n = strlen(message);

    while (n > 0) {
	i = write(server->fd, message, n);
	if (i < 0) {
	    if (errno == EINTR)
		continue;
	    else
		return -1;
	}
	if (i == 0)
	    return -1;

	n -= i;
	message += i;
    }

    return 0;
}

int server_write(SERVER *server, char *message)
{
    long	n = strlen(message);

    if (server->fd < 0)
	return -1;

    while (n > 0) {
	long	i = write(server->fd, message, n);

	if (i < 0) {
	    switch (errno) {
	    case EINTR:
		continue;
	    case EAGAIN:
#if EAGAIN != EWOULDBLOCK
	    case EWOULDBLOCK:
#endif
		if (do_wait(&server->fd, False, XtIMXEvent) < 0)
		    return -1;
		continue;
	    default:
		perror("knews: write");
		return -1;
	    }
	}

	if (i == 0)
	    return -1;

	n -= i;
	message += i;
    }

    return 0;
}

char *server_read(SERVER *server)
{
    char *c, *temp;

    if (server->fd < 0)
	return NULL;

    temp = strchr(server->buffer + server->curr_pos, '\n');
    if (!temp) {
	long	end;

	if (server->curr_pos == 0)
	    end = strlen(server->buffer);
	else {
	    end = strlen(server->buffer + server->curr_pos);
	    memmove(server->buffer,
		    server->buffer + server->curr_pos,
		    end + 1);
	    server->curr_pos = 0;
	}

	for(;;) {
	    long	n;

	    if (end >= server->len) {
		server->len *= 2;
		server->buffer = XtRealloc(server->buffer, server->len + 3);
	    }

	    n = read(server->fd, server->buffer + end, server->len - end);

	    if (n < 0) {
		switch (errno) {
		case EINTR:
		    continue;
		case EAGAIN:
#if EAGAIN != EWOULDBLOCK
		case EWOULDBLOCK:
#endif
		    if (do_wait(&server->fd, True, XtIMXEvent) < 0)
			return NULL;
		    continue;
		default:
		    perror("knews: read");
		    return NULL;
		}
	    }

	    if (n == 0)
		return NULL;

	    server->buffer[end + n] = '\0';
	    if ( (temp = strchr(server->buffer + end, '\n')) )
		break;
	    end += n;
	}
    }

    if (*(temp-1) == '\r')
	*(temp-1) = '\0';
    *temp = '\0';
    c = server->buffer + server->curr_pos;
    server->curr_pos = temp - server->buffer + 1;

    return c;
}

static char *find_crlf_dot_crlf(char *buffer)
{
    char	*c;

    c = strstr(buffer, "\r\n.\r\n");
    if (c)
	return c + 2;

#if 1
    c = strstr(buffer, "\n.\n");
    if (c)
	return c + 1;
#endif

    return NULL;
}

char *server_read_chunk(SERVER *server)
{
    char	*c, *dot;

    if (server->fd < 0)
	return NULL;

    dot = find_crlf_dot_crlf(server->buffer + server->curr_pos);
    if (!dot) {
	long	end;

	if (server->curr_pos == 0)
	    end = strlen(server->buffer);
	else {
	    end = strlen(server->buffer + server->curr_pos);
	    memmove(server->buffer, server->buffer + server->curr_pos, end + 1);
	    server->curr_pos = 0;
	}

	for(;;) {
	    long	n;

	    if (end >= server->len) {
		server->len *= 2;
		server->buffer = XtRealloc(server->buffer, server->len + 3);
	    }

	    n = read(server->fd, server->buffer + end, server->len - end);

	    if (n < 0) {
		switch (errno) {
		case EINTR:
		    continue;
		case EAGAIN:
#if EAGAIN != EWOULDBLOCK
		case EWOULDBLOCK:
#endif
		    if (do_wait(&server->fd, True, XtIMXEvent) < 0)
			return NULL;
		    continue;
		default:
		    perror("knews: read");
		    return NULL;
		}
	    }

	    if (n == 0)
		return NULL;

	    end += n;
	    server->buffer[end] = '\0';
	    if ( (dot = find_crlf_dot_crlf(server->buffer)) )
		break;
	}
    }

    *dot++ = '\0';
    if (*dot++ == '\r')
	dot++;
    c = server->buffer + server->curr_pos;
    server->curr_pos = dot - server->buffer;

    return c;
}

char *server_get_line(SERVER *server)
{
    char	*c = server->buffer + server->curr_pos;
    char	*temp;

    temp = strchr(c, '\n');
    if (temp) {
	if (*(temp-1) == '\r')
	    *(temp-1) = '\0';
	*temp++ = '\0';
	server->curr_pos = temp - server->buffer;
	return c;
    } else {
	long	len = strlen(c);

	memmove(server->buffer, c, len + 1);
	server->curr_pos = 0;
	return NULL;
    }    
}

char *server_get_chunk(SERVER *server)
{
    char	*c = server->buffer + server->curr_pos;
    char	*dot;

    dot = find_crlf_dot_crlf(c);
    if (dot) {
	*dot++ = '\0';
	if (*dot++ == '\r')
	    dot++;
	server->curr_pos = dot - server->buffer;
	return c;
    } else {
	long	len = strlen(c);

	memmove(server->buffer, c, len + 1);
	server->curr_pos = 0;
	return NULL;
    }
}

char *server_comm(SERVER *server, char *command, int reconnect)
{
    char	*reply;

    if (server_write(server, command) < 0) {
	if (!reconnect)
	    return NULL;
	else if (global.aborted ||
		 reconnect_server(False) < 0 ||
		 server_write(server, command) < 0)
	    return NULL;
    }
    reply = server_read(server);

    if (reply && atoi(reply) == ERR_NEED_AUTH) {
	char	buffer[1024];
	char	*auth_user = NntpObjectAuthInfoUser(global.nntp_object);
	char	*auth_pass = NntpObjectAuthInfoPass(global.nntp_object);
	char	*auth_gen  = NntpObjectAuthInfoGeneric(global.nntp_object);

	if (auth_user && auth_pass) {
	    sprintf(buffer, "AUTHINFO USER %s\r\n", auth_user);
	    (void)server_write(server, buffer);
	    reply = server_read(server);
	    if (!reply || atoi(reply) != CONT_AUTH)
		return reply;

	    sprintf(buffer, "AUTHINFO PASS %s\r\n", auth_pass);
	    (void)server_write(server, buffer);
	    reply = server_read(server);
	    if (!reply || atoi(reply) != OK_AUTH)
		return reply;

	    (void)server_write(server, command);
	    reply = server_read(server);
	} else if (auth_gen) {
	    pid_t	pid;
	    int		status;

	    if (!server->cookie) {
		char	*file_name;

		file_name = tmpnam(NULL);
		if (!file_name)
		    return reply;
		server->cookie = fopen(file_name, "w+");
		(void)unlink(file_name);
		if (!server->cookie)
		    return reply;
	    }	    

	    sprintf(buffer, "AUTHINFO GENERIC %s\r\n", auth_gen);
	    (void)server_write(server, buffer);
	    reply = server_read(server);
	    if (!reply || reply[0] == '5')
		return reply;

	    change_interruptible(False);
	    pid = fork_nicely(NULL, NULL, False);
	    if (pid < 0) { /* simulate broken connection */
		perror("knews: fork");
		server_write_raw(server, "\r\nQUIT\r\n");
		server_close(server);
		return NULL;
	    }

	    if (pid == 0) {
		static char	env_var[80];
		int	sd;

		sd = dup(server->fd); /* why??? */
		if (sd < 0)
		    sd = server->fd;
		sprintf(env_var, "NNTP_AUTH_FDS=%d.%d.%d",
			server->fd, sd, fileno(server->cookie));
		putenv(env_var); /* why is this not prototyped?? */
		sprintf(buffer, "xterm -e %s", auth_gen);
		execl("/bin/sh", "sh", "-c", buffer, (char *)0);
		_exit(127);
	    }

	    pid = wait_for_pid(pid, &status, NULL);
	    change_interruptible(True);

	    if (pid < 0) {
		server_write_raw(server, "\r\nQUIT\r\n");
		server_close(server);
		return NULL;
	    }

	    reply = server_read(server);
	    if (!reply || atoi(reply) != OK_AUTH)
		return reply;

	    (void)server_write(server, command);
	    reply = server_read(server);
	}
    }

    return reply;
}

int server_write_chunk_arts(SERVER *server, ARTICLE **art_list,
			    int head, int body)
{
    char	*buffer;
    char	*cmd;
    char	*c;
    int		n;

    if (!*art_list)
	return 0;

    if (head && body)
	cmd = "ARTICLE";
    else if (head)
	cmd = "HEAD";
    else if (body)
	cmd = "BODY";
    else
	return 0;

    n = 0;
    c = buffer = XtMalloc(32 * global.chunk_size);
    while (art_list[n] && n < global.chunk_size) {
	sprintf(c, "%s %ld\r\n", cmd, art_list[n]->no);
	c += strlen(c);
	/*  Should be c += sprintf(...)
	 */
	n++;
    }
    *c = '\0';
    (void)server_write(server, buffer);
    XtFree(buffer);

    return n;
}
