/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */

/* The general threading algorithm, and in particular the functions
 * link_child(), unlink_child() and thread_article(), are modified
 * from the source to trn.
 */

#include "global.h"
#include <time.h>
#include "ahead.h"
#include "codes.h"
#include "connect.h"
#include "hash.h"
#include "kill.h"
#include "misc.h"
#include "parse.h"
#include "server.h"
#include "thread.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Message.h"

#define IS_SPACE(c) ((c) == ' ' || (c) == '\t')

static void overview_error(long art_no, char *string)
{
    fprintf(stderr, "Bad overview record");
    if (art_no) fprintf(stderr, ", article no %ld", art_no);
    fprintf(stderr, "; %s\n", string);
}

static void unlink_child(ARTICLE *child)
{
    ARTICLE	*last = A_PARENT(child);

    if (last) {
	if (child == A_CHILD1(last))
	    CHILD1(last) = SIBLING(child);
	else {
	    last = A_CHILD1(last);
	    while (child != A_SIBLING(last))
		last = A_SIBLING(last);
	    SIBLING(last) = SIBLING(child);
	}
    } else {
	SUBJECT	*subj = child->subject;

	if ((last = subj->thread) == child) {
	    do {
		subj->thread = A_SIBLING(child);
		subj = subj->prev;
	    } while (subj && subj->thread == last);

	    subj = child->subject->next;

	    while (subj && subj->thread == last) {
		subj->thread = A_SIBLING(child);
		subj = subj->next;
	    }
	} else {
	    while (child != A_SIBLING(last))
		last = A_SIBLING(last);
	    SIBLING(last) = SIBLING(child);
	}
    }
}

static void link_child(ARTICLE *child)
{
    ARTICLE	*art = A_PARENT(child);

    if (art) {
	art = A_CHILD1(art);
	if (!art || child->date < art->date) {
	    SIBLING(child) = (ART_TREE_NODE *)art;
	    CHILD1(A_PARENT(child)) = (ART_TREE_NODE *)child;
	} else {
	    while (A_SIBLING(art) && A_SIBLING(art)->date <= child->date)
		art = A_SIBLING(art);
	    SIBLING(child) = SIBLING(art);
	    SIBLING(art) = (ART_TREE_NODE *)child;
	}
    } else {
	SUBJECT	*subj = child->subject;

	art = subj->thread;
	if (!art || child->date < art->date) {
	    do {
		subj->thread = child;
		subj = subj->prev;
	    } while (subj && subj->thread == art);

	    subj = child->subject->next;

	    while (subj && subj->thread == art) {
		subj->thread = child;
		subj = subj->next;
	    }
	    SIBLING(child) = (ART_TREE_NODE *)art;
	} else {
	    while (A_SIBLING(art) && A_SIBLING(art)->date <= child->date)
		art = A_SIBLING(art);
	    SIBLING(child) = SIBLING(art);
	    SIBLING(art) = (ART_TREE_NODE *)child;
	}
    }
}

static void merge_threads(THREAD_CONTEXT *context, SUBJECT *s1, SUBJECT *s2)
{
    SUBJECT *s;
    ARTICLE *t1 = s1->thread;
    ARTICLE *t2 = s2->thread;

    while (s1->next && s1->next->thread == t1)
	s1 = s1->next;
    while (s2->prev && s2->prev->thread == t2)
	s2 = s2->prev;

    s = s2;
    s->thread = t1;
    while (s->next && s->next->thread == t2) {
	s = s->next;
	s->thread = t1;
    }

    if (s2->prev)
	s2->prev->next = s->next;
    else
	context->subjects = s->next;
    if (s->next)
	s->next->prev = s2->prev;

    if ( (s->next = s1->next) )
	s->next->prev = s;
    s1->next = s2;
    s2->prev = s1;

    for (t1 = t2 ; t1 ; t1 = t2) {
	t2 = A_SIBLING(t2);
	link_child(t1);
    }
}

static SUBJECT *get_subject(THREAD_CONTEXT *context, char *str)
{
    SUBJECT	    **sp;

    if (!*(sp = subj_hash_find(context->subj_ht, str, strlen(str)))) {
	(*sp) = (SUBJECT *)XtMalloc(sizeof(SUBJECT));
	(*sp)->hash_next = NULL;
	(*sp)->next = NULL;
	(*sp)->prev = NULL;
	(*sp)->thread = NULL;
	(*sp)->no_unread = 0;
	(*sp)->disp = 0;
	(*sp)->hot = 0;
	(*sp)->has_tagged = False;
	(*sp)->subject = XtNewString(str);
	(*sp)->hash_len = strlen(str);
	if (context->subjects) {
	    if (!context->last_subj || context->last_subj->next) {
		/* subjects have been reordered by merge_threads */
		context->last_subj = context->subjects;
		while (context->last_subj->next)
		    context->last_subj = context->last_subj->next;
	    }

	    (*sp)->prev = context->last_subj;
	    context->last_subj->next = *sp;
	    context->last_subj = *sp;
	} else {
	    context->subjects = *sp;
	    context->last_subj = context->subjects;
	}
    }

    return *sp;
}

/* replaces ascii 0-32 & 127 and iso-8859-x 128-159 & 255 with ' ' */
static void munge_nonprintable(char *str)
{
    unsigned char	*c = (unsigned char *)str;

    while (*c != '\0') {
	if ((*c & 0x60) == 0 || *c == '\b')
	    *c = ' ';
	c++;
    }
}

ARTICLE *parse_xover_line(THREAD_CONTEXT *context, char *buffer) {
    ARTICLE	*art;
    ARTICLE	**fake_p;
    char	*c1, *c2;
    long	art_no, lines /*, bytes*/;
    char	*subject, *from, *msgid;
    time_t	date;

    if (!(c1 = strchr(buffer, '\t')) || buffer[0] < '0' || buffer[0] > '9') {
	overview_error(0, "no article number.");
	return NULL;
    }
    *(c1++) = '\0';
    art_no = atol(buffer);

    if (!(c2 = strchr(c1, '\t'))) {
	overview_error(art_no, "no Subject: field.");
	return NULL;
    }
    *(c2++) = '\0';
    munge_nonprintable(c1);
    subject = eat_re(c1);

    if (!(c1 = strchr(c2, '\t'))) {
	overview_error(art_no, "no From: field.");
	return NULL;
    }
    *(c1++) = '\0';
    munge_nonprintable(c2);
    from = c2;

    if (!(c2 = strchr(c1, '\t'))) {
	overview_error(art_no, "no Date: field.");
	return NULL;
    }
    *(c2++) = '\0';
    date = parsedate(c1);

    if (!(c1 = strchr(c2, '\t'))) {
	overview_error(art_no, "no Message-Id: field.");
	return NULL;
    }
    if (*(c1-1) != '>') {
	overview_error(art_no, "syntax error in Message-Id: field.");
	return NULL;
    }
    *(c1-1) = '\0';
    if (*(c2++) != '<' || !valid_msgid(c2, c1 - 1)) {
	overview_error(art_no, "syntax error in Message-Id: field.");
	return NULL;
    }
    c1++;
    msgid = c2;

    if (!(c2 = strchr(c1, '\t'))) {
	context->refs = NULL;
	overview_error(art_no, "no References: field.");
	return NULL;
    }
    *(c2++) = '\0';
    context->refs = c1;

    if (!(c1 = strchr(c2, '\t'))) {
	overview_error(art_no, "no Bytes: field.");
	return NULL;
    }
    *(c1++) = '\0';
/*    if (*c2 < '0' || *c2 > '9')
	bytes = 0;
    else
	bytes = atol(c2);*/
    /* Bytes, not optional */

    if ( (c2 = strchr(c1, '\t')) )
	*(c2++) = '\0';
    if (*c1 < '0' || *c1 > '9')
	lines = 0;
    else
	lines = atol(c1);
    /* Lines, not optional */

    if ( (art = *(fake_p = art_hash_find(context->art_ht,
					 msgid, strlen(msgid)))) ) {
	if (art->from) {
	    overview_error(art_no, "duplicate Message-Id.");
	    return NULL; /* duplicate Message-ID */
	}
	context->fake_had_subject = art->subject;
    } else {
	art = *fake_p = (ARTICLE *)XtMalloc(sizeof(ARTICLE));
	art->tree_data.hook = NULL;
	art->tree_data.label = NULL;
	PARENT(art) = NULL;
	SIBLING(art) = NULL;
	CHILD1(art) = NULL;
	art->hash_next = NULL;
	art->next = NULL;
	art->xref = NULL;
	art->hot = 0;
	art->read = FALSE;
	art->killed = FALSE;
	art->msgid = XtNewString(msgid);
	art->hash_len = strlen(msgid);
	context->fake_had_subject = NULL;
    }

    art->no = art_no;
    art->subject = get_subject(context, subject);
    art->from = XtNewString(from);
    art->date = date;
    art->lines = lines;
/*    art->bytes = bytes;*/

    while (c2) {
	c1 = c2;
	if ( (c2 = strchr(c1, '\t')) )
	    *(c2++) = '\0';
	if (strncasecmp(c1, "Xref: ", 6) == 0) {
	    c1 = strchr(c1 + 6, ' ');
	    if (c1) {
		c1++;
		art->xref = XtNewString(c1);
	    }
	    break;
	}
    }

    return art;
}

void thread_article(THREAD_CONTEXT *context, ARTICLE *art)
{
    ARTICLE	*a, *last;

    if (context->fake_had_subject) {
	ARTICLE *stop;

	if (context->fake_had_subject->thread != art->subject->thread)
	    merge_threads(context, context->fake_had_subject, art->subject);
	a = A_PARENT(art);
	while (a && !a->from && !A_SIBLING(A_CHILD1(a)))
	    a = A_PARENT(a);
	stop = a;
	unlink_child(art);
	for (a = A_PARENT(art) ; a != stop ; a = last) {
	    unlink_child(a);
	    last = A_PARENT(a);
	    a->date = 0;
	    a->subject = NULL;
	    PARENT(a) = NULL;
	}
	PARENT(art) = NULL;
	SIBLING(art) = NULL;
    }

    if (context->refs) {
	ARTICLE	**ap;
	char	*c, *end;

	last = art;
	a = NULL;
	end = context->refs + strlen(context->refs) - 1;
	while (TRUE) {
	    for (c = end ; c >= context->refs && *c != '<' ; c--);
	    if (*c != '<')
		break;
	    while (end >= c && ((unsigned char)*end <= ' ' || *end == ','))
		end--;
	    if (end < c || *end != '>')
		break;
	    *end = '\0';
	    if (!valid_msgid(c+1, end-1))
		break;
	    ap = art_hash_find(context->art_ht, c+1, end-c-1);
	    a = *ap;
	    if (a) {
		if ((a->date != 0 && !a->subject) || a == art) {
		    if ((a = last) == art) a = NULL;
		    continue;
		}
	    } else {
		a = *ap = (ARTICLE *)XtMalloc(sizeof(ARTICLE));
		a->tree_data.hook = NULL;
		a->tree_data.label = NULL;
		PARENT(a) = NULL;
		SIBLING(a) = NULL;
		CHILD1(a) = NULL;
		a->hash_next = NULL;
		a->next = NULL;
		a->from = NULL;
		a->xref = NULL;
		a->subject = NULL;
		a->no = 0;
		a->hot = 0;
		a->lines = 0;
/*		a->bytes = 0;*/
		a->read = FALSE;
		a->killed = FALSE;
		a->hash_len = end-c-1;
		a->msgid = XtNewString(c+1);
	    }
	    PARENT(last) = (ART_TREE_NODE *)a;
	    link_child(last);
	    if (a->subject)
		break;
	    a->date = art->date;
	    last = a;
	    end = c - 1;
	}
	if (!a) { /* no references */
	    link_child(art);
	    return;
	}
	if (a->subject) {
	    if (art->subject->thread != a->subject->thread)
		merge_threads(context, art->subject, a->subject);
	} else {
	    a->subject = art->subject;
	    link_child(a);
	}

	for (a = A_PARENT(art) ; a && !a->subject ; a = A_PARENT(a))
	    a->subject = art->subject;

	while (a && art != A_PARENT(a))
	    a = A_PARENT(a);
	if (a) {
	    unlink_child(a);
	    PARENT(a) = NULL;
	    link_child(a);
	}
    } else {
	link_child(art);
    }
}

static void mark_read_articles(THREAD_CONTEXT *context)
{
    SUBJECT		*subj = context->subjects;
    ARTICLE		*art = context->articles;
    long		n = 0;
    ART_LIST_NODE	*loop = context->group->read_arts;

    while (subj) {
	subj->no_unread = 0;
	subj = subj->next;
    }

    while (art && loop) {
	while (art && art->no < loop->first) {
	    art->read = False;
	    if (art->from) {
		n++;
		art->subject->no_unread++;
	    }
	    art = art->next;
	}

	while (art && art->no >= loop->first && art->no <= loop->last) {
	    if (art->from)
		art->read = True;
	    art = art->next;
	}

	loop = loop->next;
    }

    while (art) {
	art->read = False;
	if (art->from) {
	    n++;
	    art->subject->no_unread++;
	}
	art = art->next;
    }

    context->group->no_unread = n;
}

static void fix_authors(THREAD_CONTEXT *context)
{
    ARTICLE	*art = context->articles;
    char	author[128];

    if (GroupObjectShowNumberLines(global.group_object)) {
	char	buffer[160];

	while (art) {
	    int		n;
	    char	*c;

	    n = strlen(art->from);
	    if (n > sizeof(author) - 1)
		n = sizeof(author) - 1;
	    memcpy(author, art->from, n);
	    author[n] = '\0';
	    c = parse_author(author);

	    sprintf(buffer, "%4ld  %s", art->lines, c);
	    art->tree_data.label = XtNewString(buffer);

	    art = art->next;
	}
    } else {
	while (art) {
	    int	n;
	    char	*c;

	    n = strlen(art->from);
	    if (n > sizeof(author) - 1)
		n = sizeof(author) - 1;
	    memcpy(author, art->from, n);
	    author[n] = '\0';
	    c = parse_author(author);

	    art->tree_data.label = XtNewString(c);

	    art = art->next;
	}
    }
}

static char *xover_articles()
{
    char	command[512];
    char	*buffer;
    ARTICLE	*art;
    long	i, j, m;

    j = m = global.thr.total / 32;
    i = global.thr.n_done * 100;

    for (buffer = server_read(&global.server) ;
	 buffer && (buffer[0] != '.' || buffer[1] != '\0') ;
	 i += 100, buffer = server_read(&global.server)) {
	art = parse_xover_line(&global.thr, buffer);
	if (art) {
	    if (global.thr.last_art)
		global.thr.last_art->next = art;
	    else
		global.thr.articles = art;
	    global.thr.last_art = art;

	    thread_article(&global.thr, art);
	}

	if (j-- <= 0) {
	    sprintf(command, "Reading group... %3ld%%",
		    i / global.thr.total);
	    set_message(command, False);
	    j = m;
	}
    }

    if (!buffer)
	return NULL;
    
    return CODE_TO_STR(OK_GROUP);
}

/*************************************************************************/

#define MAX_CHUNK_SIZE	64

static int put_chunk_head(long max)
{
    char	command[64 * 12 + 1];
    char	*c;
    int		n;

    if (max <= 0)
	return 0;

    if (max > global.chunk_size)
	max = global.chunk_size;
    if (max > MAX_CHUNK_SIZE)
	max = MAX_CHUNK_SIZE;

    for (n = 0, c = command ; n < max ; n++, c += 12)
	sprintf(c, "HEAD\r\nNEXT\r\n");

    *c = '\0';
    server_write(&global.server, command);

    return max;
}

ARTICLE *parse_head_reply(THREAD_CONTEXT *context, char *buffer)
{
    long	art_no;

    buffer += 3;
    while (*buffer == ' ' || *buffer == '\t')
	buffer++;
    if (*buffer < '0' || *buffer > '9')
	return NULL;

    art_no = 0;
    do {
	art_no *= 10;
	art_no += *buffer++ - '0';
    } while (*buffer >= '0' && *buffer <= '9');

    while (*buffer == ' ' || *buffer == '\t')
	buffer++;

    if (*buffer++ == '<') {
	char	*c;

	c = strchr(buffer, '>');
	if (c) {
	    *c = '\0';
	    if (valid_msgid(buffer, c)) {
		long	msgid_len = c - buffer;
		ARTICLE	**fake_p = art_hash_find(context->art_ht,
						 buffer, msgid_len);
		ARTICLE	*art = *fake_p;

		if (art) {
		    if (art->from) {
			overview_error(art_no, "duplicate Message-Id.");
			return NULL;
		    }
		    context->fake_had_subject = art->subject;
		} else {
		    *fake_p = (ARTICLE *)XtMalloc(sizeof(ARTICLE));
		    art = *fake_p;
		    art->tree_data.hook = NULL;
		    art->tree_data.label = NULL;
		    PARENT(art) = NULL;
		    SIBLING(art) = NULL;
		    CHILD1(art) = NULL;
		    art->hash_next = NULL;
		    art->next = NULL;
		    art->msgid = XtNewString(buffer);
		    art->hash_len = msgid_len;
		    art->from = NULL;
		    art->xref = NULL;
		    art->date = (time_t)(-1);
		    art->lines = 0;
/*		    art->bytes = 0;*/
		    art->hot = 0;
		    art->read = False;
		    art->killed = False;
		    context->fake_had_subject = NULL;
		}

		art->no = art_no;

		return art;
	    }
	}
    }

    return NULL;
}

int parse_head(THREAD_CONTEXT *context, ARTICLE *art, char *buffer)
{
    char	*from = NULL, *subject = NULL, *xref = NULL, *refs = NULL;
    long	/*bytes = -1,*/ lines = -1;
    time_t	date = -1;

    while (*buffer != '\0') {
	int	skip = True;

	switch (*buffer) {
	case 'F':
	case 'f':
	    if (!from && strncasecmp(buffer, "From:", 5) == 0) {
		buffer += 5;
		while (IS_SPACE(*buffer))
		    buffer++;
		from = buffer;
		skip = False;
	    }
	    break;
	case 'S':
	case 's':
	    if (!subject && strncasecmp(buffer, "Subject:", 8) == 0) {
		buffer += 8;
		while (IS_SPACE(*buffer))
		    buffer++;
		buffer = eat_re(buffer);
		subject = buffer;
		skip = False;
	    }
	    break;
	case 'X':
	case 'x':
	    if (!xref && strncasecmp(buffer, "Xref:", 5) == 0) {
		char	*c;

		buffer += 5;
		while (IS_SPACE(*buffer))
		    buffer++;
		c = strchr(buffer, ' ');
		if (c) {
		    xref = c + 1;
		    skip = False;
		}
	    }
	    break;
	case 'R':
	case 'r':
	    if (!refs && strncasecmp(buffer, "References:", 11) == 0) {
		buffer += 11;
		while (IS_SPACE(*buffer))
		    buffer++;
		refs = buffer;
		skip = False;
	    }
	    break;
	case 'D':
	case 'd':
	    if (strncasecmp(buffer, "Date:", 5) == 0) {
		buffer += 5;
		while (IS_SPACE(*buffer))
		    buffer++;
		date = parsedate(buffer);
	    }
	    break;
/*	case 'B':
	case 'b':
	    if (strncasecmp(buffer, "Bytes:", 6) == 0) {
		buffer += 6;
		bytes = atol(buffer);
	    }
	    break;*/
	case 'L':
	case 'l':
	    if (strncasecmp(buffer, "Lines:", 6) == 0) {
		buffer += 6;
		lines = atol(buffer);
	    }
	    break;
	}

	if (skip) {
	    buffer = strchr(buffer, '\n');
	    while (buffer && IS_SPACE(buffer[1]))
		buffer = strchr(buffer + 1, '\n');
	    if (!buffer)
		break;
	    buffer++;
	} else {
	    char	*src, *dest;

	    src = strchr(buffer, '\n');
	    if (!src)
		break;
	    dest = src;
	    if (*(dest - 1) == '\r')
		dest--;

	    for (;;) {
		if (*src == '\r')
		    src++;
		else if (*src == '\0')
		    break;
		else if (*src == '\n') {
		    src++;
		    if (!IS_SPACE(*src))
			break;
		} else
		    *dest++ = *src++;
	    }

	    *dest = '\0';
	    buffer = src;	    
	}
    }

    if (!from || !subject)
	return False;

    munge_nonprintable(subject);
    munge_nonprintable(from);
    art->subject = get_subject(context, subject);
    art->from = XtNewString(from);
    art->date = date;
    art->lines = lines;
/*    art->bytes = bytes;*/
    if (xref)
	art->xref = XtNewString(xref);
    context->refs = refs;

    return True;
}

static char *head_articles()
{
    char	command[512];
    char	*buffer;
    ARTICLE	*art = NULL;
    long	n_pending, n_to_put;
    long	max_count, count;

    n_pending = put_chunk_head(global.thr.total);
    n_to_put = global.thr.total - n_pending;

    count = max_count = global.thr.total / 32;

    for (;;) {
	n_pending--;
	if (n_pending <= 0) {
	    n_pending = 1;
	    n_to_put = 0;
	    server_write(&global.server, "HEAD\r\nNEXT\r\n");
	} else if (n_pending < global.chunk_size && n_to_put > 0) {
	    long	temp;

	    temp = put_chunk_head(n_to_put);
	    n_pending += temp;
	    n_to_put -= temp;
	}

	buffer = server_read(&global.server);
	if (!buffer)
	    return NULL;

	if (global.head_debug)
	    fprintf(stderr, "%s\n", buffer);

	if (atoi(buffer) == OK_HEAD) {
	    art = parse_head_reply(&global.thr, buffer);
	    buffer = server_read_chunk(&global.server);
	    if (art && parse_head(&global.thr, art, buffer)) {
		if (global.thr.last_art)
		    global.thr.last_art->next = art;
		else
		    global.thr.articles = art;
		global.thr.last_art = art;

		thread_article(&global.thr, art);
		global.thr.refs = NULL;
	    }
	}

	buffer = server_read(&global.server);
	if (!buffer)
	    break;

	if (global.head_debug)
	    fprintf(stderr, "%s\n", buffer);

	if (atoi(buffer) != OK_NOTEXT)
	    break;

	global.thr.n_done++;

	if (count-- <= 0) {
	    sprintf(command,
		    "Reading group slowly... %3ld%%",
		    (100 * global.thr.n_done) / global.thr.total);
	    set_message(command, False);
	    count = max_count;
	}
    }

    if (!buffer)
	return NULL;

    if (global.head_debug && n_pending > 0)
	fprintf(stderr, "%ld pending, snarfing them.\n", n_pending);

    while (n_pending-- > 0) { /* next returned no next article */
	buffer = server_read(&global.server);
	if (!buffer)
	    return NULL;

	if (global.head_debug)
	    fprintf(stderr, "%s\n", buffer);

	if (atoi(buffer) == OK_HEAD) {
	    buffer = server_read_chunk(&global.server);
	    if (!buffer)
		return NULL;
	}

	buffer = server_read(&global.server); /* reply to next */
	if (!buffer)
	    return NULL;

	if (global.head_debug)
	    fprintf(stderr, "%s\n", buffer);
    }
    
    return CODE_TO_STR(OK_GROUP);
}

static char *get_articles(char *buffer)
{
    char	command[512];
    long	total, first, last;

    if (!buffer) {
	sprintf(command, "GROUP %s\r\n", global.thr.group->name);
	buffer = server_comm(&global.server, command, True);
    }

    if (!buffer || atoi(buffer) != OK_GROUP)
	return buffer;
    (void)sscanf(buffer, "%*d%ld%ld%ld", &total, &first, &last);
    global.thr.group->first_art = first;
    global.thr.group->last_art = last;
    if (total <= 0)
	total = 1;
    global.thr.total = total;

    global.curr_art  = NULL;
    global.curr_subj = NULL;

    if (global.xover_supported) {
	if (global.thr.last_art)
	    first = global.thr.last_art->no + 1;
	if (first > last)
	    return CODE_TO_STR(OK_GROUP);

	sprintf(command, "XOVER %ld-%ld\r\n", first, last);
	buffer = server_comm(&global.server, command, True);
	if (!buffer)
	    return NULL;
	if (atoi(buffer) == OK_XOVER)
	    return xover_articles();
    }

    if (global.thr.last_art) { /* stat the first article we want */
	long	n;

	first = global.thr.last_art->no + 1;
	/* this should succeed on the first pass */
	for (n = first ; n <= last ; n++) {
	    sprintf(command, "STAT %ld\r\n", first);
	    buffer = server_read(&global.server);
	    if (!buffer)
		return NULL;
	    if (atoi(buffer) == OK_ARTICLE)
		break;
	    if (atoi(buffer) != ERR_NOART)
		return buffer;
	}

	if (n > last)
	    return CODE_TO_STR(OK_GROUP);

	if (n == first) { /* got the first, go to next */
	    sprintf(command, "NEXT\r\n");
	    (void)server_write(&global.server, command);
	    buffer = server_read(&global.server);
	    if (!buffer)
		return NULL;
	    if (atoi(buffer) == ERR_NONEXT) /* got'em all */
		return CODE_TO_STR(OK_GROUP);
	    if (atoi(buffer) != OK_NOTEXT)
		return buffer;
	}

    }

    return head_articles();
}

void read_group(char *reply)
{
    long	no[2];

    if (!global.thr.group ||
	thread_ahead_check(global.thr.group) < 0)
	return;

    if (global.group_object)
	XtDestroyWidget((Widget)global.group_object);
    global.group_object =
	GroupObjectCreate((Widget)global.nntp_object, global.thr.group->name);
    update_misc_menu();

    set_busy(True, True);
    set_message("Server contacted, waiting for response...", False);
    reply = get_articles(reply);

    if (!reply) {
	clear_thread_context(&global.thr);
	(void)reconnect_server(True);
	unset_busy();
    } else {
	char	message[256];
	int	status = atoi(reply);

	if (status == OK_GROUP) {
	    change_interruptible(False);
	    set_message("Killing...", False);
	    mark_read_articles(&global.thr);
	    fix_authors(&global.thr);
	    kill_articles(no);
	    unset_busy();

	    sprintf(message,
		    "Group %s with %ld unread articles, "
		    "%ld killed, %ld hot.",
		    global.thr.group->name,
		    global.thr.group->no_unread,
		    no[0], no[1]);
	    set_message(message, False);
	    setNewsModeGroup(False);
	} else {
	    unset_busy();

	    if (strlen(reply) > 200) reply[200] = '\0';
	    sprintf(message,
		    "Error: failed message from server is %s", reply);
	    set_message(message, True);
	}
    }
}
