/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "codes.h"
#include "connect.h"
#include "context.h"
#include "server.h"
#include "util.h"
#include "uudecode.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/Message.h"

#define UU_REGEXP \
	"([^ \t]+)" 			/* the file name		*/ \
	"([[({ \t]|part|-)*"		/* skip garbage			*/ \
	"([0-9]+)"			/* which part			*/ \
	"[ \t]*(/|of)[ \t]*"		/* skip separator & garbage	*/ \
	"([0-9]+)"			/* number of parts		*/
#define UU_NAME_OFFSET		1
#define UU_PART_OFFSET		3
#define UU_NOPARTS_OFFSET	5
#define UU_NO_OFFSETS		6

static int uu_parse_subject(char *subject, char **file_name,
			    int *no_parts, int *part)
{
    static regex_t	re;
    static int		inited = False;
    regmatch_t		pmatch[UU_NO_OFFSETS];

    if (!inited) {
	if (regcomp(&re, UU_REGEXP, REG_EXTENDED | REG_ICASE) != 0) {
	    fprintf(stderr,
		    "knews: internal error, failed to compile"
		    "regexp to use for uudecoding.");
	    regfree(&re);
	    return False;
	}
	inited = True;
    }

    if (regexec(&re, subject, UU_NO_OFFSETS, pmatch, 0) != 0)
	return False;

    if (no_parts) {
	int	i;

	*no_parts = 0;
	for (i = pmatch[UU_NOPARTS_OFFSET].rm_so ;
	     i < pmatch[UU_NOPARTS_OFFSET].rm_eo ; i++) {
	    *no_parts *= 10;
	    *no_parts += subject[i] - '0';
	}
    }

    if (part) {
	int	i;

	*part = 0;
	for (i = pmatch[UU_PART_OFFSET].rm_so ;
	     i < pmatch[UU_PART_OFFSET].rm_eo ; i++) {
	    *part *= 10;
	    *part += subject[i] - '0';
	}
    }

    if (file_name) {
	int	len;

	len = pmatch[UU_NAME_OFFSET].rm_eo -
	    pmatch[UU_NAME_OFFSET].rm_so;
	*file_name = XtMalloc(len + 1);
	(void)memcpy(*file_name,
		     &subject[pmatch[UU_NAME_OFFSET].rm_so], len);
	(*file_name)[len] = '\0';
    }    

    return True;
}

static int do_decode(ARTICLE **arts, int no_parts)
{
    int		fd[2];
    pid_t	child_pid;
    char	stderr_buf[STDERR_BUFFLEN];
    int		result = False;

    if (pipe(fd) < 0) {
	perror("knews: pipe");
	set_message("Failed to open pipe to 'uudecode' program!", True);
	return False;
    }

    child_pid = fork_nicely(NULL, NULL, global.stderr_timeout >= 0);
    if (child_pid < 0) {
	(void)close(fd[0]);
	(void)close(fd[1]);
	set_message("Failed to start 'uudecode' program!", True);
	return False;
    }

    if (child_pid == 0) { /* child */
	char	*uu_dir = GroupObjectUuDir(global.group_object);
	char	*uu_program = GroupObjectUuProgram(global.group_object);

	if (uu_dir)
	    (void)chdir(uu_dir);

	(void)close(fd[1]);

	if (fd[0] != STDIN_FILENO) {
	    fd[0] = dup2(fd[0], STDIN_FILENO);
	    if (fd[0] < 0) {
		perror("knews: dup2");
		_exit(127);
	    }

	}

	if (uu_program) {
	    execl("/bin/sh", "sh", "-c", uu_program, (char *)NULL);
	    perror("knews: execl");
	    _exit(127);
	} else {
	    pid_t	pid;

	    if (pipe(fd) < 0) {
		perror("knews: pipe");
		_exit(127);
	    }

	    pid = fork();
	    if (pid < 0) {
		perror("knews: fork");
		_exit(127);
	    } else if (pid == 0) {
		(void)close(fd[1]);
		(void)close(STDIN_FILENO);

		if (fd[0] != STDIN_FILENO) {
		    fd[0] = dup2(fd[0], STDIN_FILENO);
		    if (fd[0] < 0) {
			perror("knews:dup2");
			_exit(127);
		    }
		}

		execlp("uudecode", "uudecode", (char *)0);
		perror("knews: execlp");
		_exit(127);
	    } else {
		char	buffer[1024];
		FILE	*fin, *fout = NULL;
		enum {
		    UuStateSearching,
		    UuStateSkipping,
		    UuStateDoing
		} state = UuStateSearching;

		(void)close(fd[0]);
		if (!(fin = fdopen(STDIN_FILENO, "r")) ||
		    !(fout = fdopen(fd[1], "w"))) {
		    perror("knews: fdopen");
		    _exit(127);
		}

		while (fgets(buffer, sizeof(buffer), fin)) {
		    char	*c = strchr(buffer, '\n');

		    if (c)
			*c = '\0';
		    else {
			fprintf(stderr, "Buffer overflow.\n");
			_exit(1);
		    }
	    
		    switch (state) {
		    case UuStateSearching:
			if (strncmp(buffer, "BEGIN", 5) == 0)
			    state = UuStateDoing;
			else if (strncmp(buffer, "begin ", 6) == 0 &&
				 isdigit(buffer[6]) &&
				 isdigit(buffer[7]) &&
				 isdigit(buffer[8])) {
			    state = UuStateDoing;
			    if (fprintf(fout, "%s\n", buffer) == EOF)
				_exit(1);
			}
			break;
		    case UuStateSkipping:
			if (strncmp(buffer, "BEGIN", 5) == 0)
			    state = UuStateDoing;
			break;
		    case UuStateDoing:
			if (strncmp(buffer, "END", 3) == 0) {
			    state = UuStateSkipping;
			    break;
			}
			if (fprintf(fout, "%s\n", buffer) == EOF)
			    _exit(1);
			break;
		    }
		}

		if (state == UuStateSearching) {
		    fprintf(stderr, "No begin line.\n");
		    _exit(1);
		}

		(void)fclose(fin);
		if (fclose(fout) == EOF) {
		    perror("fclose");
		    _exit(1);
		}

		_exit(0);
	    }
	}
	_exit(127); /* paranoia */
    } else { /* parent */
	char	command[80];
	char	*buffer = NULL;
	FILE	*fp = NULL;
	int	status;
	int	part = 0;
	int	broken_pipe = False;

	(void)close(fd[0]);
	fp = fdopen(fd[1], "w");
	if (!fp)
	    (void)close(fd[1]);
	else {
	    int	n_pending = 0, n_put;

	    set_busy(True, True);
	    set_message("Uudecoding...", False);

	    n_pending = n_put =
		server_write_chunk_arts(&global.server, arts, False, True);

	    for (part = 0 ; !broken_pipe && part < no_parts ; part++) {
		n_pending--;
		if (n_pending < global.chunk_size &&
		    n_put < no_parts) {
		    int	temp;

		    temp =
			server_write_chunk_arts(&global.server,
						arts, False, True);
		    n_put += temp;
		    n_pending += temp;
		}

		buffer = server_read(&global.server);
		if (!buffer || atoi(buffer) != OK_BODY)
		    break;

		buffer = server_read(&global.server);
		while (buffer &&
		       (buffer[0] != '.' || buffer[1] != '\0')) {
		    if (!broken_pipe)
			if (fprintf(fp, "%s\n", buffer) == EOF)
			    broken_pipe = True;

		    buffer = server_read(&global.server);
		}

		sprintf(command,
			"Uudecoding...    %3d%%",
			(100 * (part + 1)) / no_parts);
		set_message(command, False);

		if (!buffer || broken_pipe)
		    break;
	    }

	    (void)fclose(fp);

	    if (buffer) {
		while (n_pending-- > 0) {
		    buffer = server_read(&global.server);
		    if (buffer && atoi(buffer) == OK_BODY)
			while (buffer &&
			       (buffer[0] != '.' || buffer[1] != '\0'))
			    buffer = server_read(&global.server);
		    if (!buffer)
			break;
		}
	    }
	}

	change_interruptible(False);
	child_pid = wait_for_pid(child_pid, &status, stderr_buf);
	if (!buffer) {
	    (void)reconnect_server(True);
	    unset_busy();
	}
	unset_busy();

	if (broken_pipe)
	    set_message("Broken pipe to 'uudecode' program!", True);
	else if (part < no_parts)
	    set_message("Error: Couldn't get all parts from server!", True);
	else {
	    if (WIFEXITED(status)) {
		status = WEXITSTATUS(status);

		if (status == 0) {
		    set_message("Uudecode done.", False);
		    result = True;
		} else if (status == 127)
		    set_message("Failed to start 'uudecode' program!", True);
		else
		    set_message("Program 'uudecode' exited abnormally!", True);
	    } else {
		set_message("Unknown error with 'uudecode' program!", True);
	    }
	}
    }

    if (!result)
	stderr_popup(stderr_buf, -1);

    return result;
}

void uudecode_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
    char	*file_name = NULL;
    int		is_uu, no_parts, part;
    ARTICLE	**arts;
    SUBJECT	*subj1, *subj2;

    if (global.busy) return;

    if (global.mode != NewsModeGroup &&
	global.mode != NewsModeThread) {
	if (global.bell)
	    XBell(XtDisplay(w), 0);
	return;
    }

    file_name = NULL;
    if (global.curr_art)
	is_uu = uu_parse_subject(global.curr_art->subject->subject,
				 &file_name, &no_parts, NULL);
    else if (global.curr_subj)
	is_uu = uu_parse_subject(global.curr_subj->subject,
				 &file_name, &no_parts, NULL);
    else {
	set_message("No selected subject!", True);
	return;
    }

    if (!is_uu || no_parts <= 0 || !file_name) {
	XtFree(file_name);
	set_message("This doesn't appear to be a uuencoded article.", True);
	return;
    }

    arts = (ARTICLE **)XtMalloc((no_parts + 2) * sizeof(ARTICLE *));
    for (part = 0 ; part < no_parts + 2 ; part++)
	arts[part] = NULL;

 fprintf(stderr, "%s\n", file_name);

#define FIND_PART(subj) \
    if (strstr(subj->subject, file_name) && \
	uu_parse_subject(subj->subject, NULL, NULL, &part)) { \
	if (part > no_parts) { \
	    set_message("Error: I'm confused by that subject...", True); \
	    XtFree((char *)arts); \
	    XtFree(file_name); \
	    return; \
	} \
	if (part >= 0 && !arts[part]) { \
	    ARTICLE	*art; \
	    for (art = subj->thread ; art ; \
		 art = next_in_thread_wrap(art)) \
		if (art->from && art->subject == subj) { \
		    arts[part] = art; \
		    break; \
		} \
	} \
    }

    subj1 = global.curr_subj;
    subj2 = subj1->prev;
    while (subj1 || subj2) {
	if (subj1) {
	    FIND_PART(subj1);
	    subj1 = subj1->next;
	}
	if (subj2) {
	    FIND_PART(subj2);
	    subj2 = subj2->prev;
	}
    }

    for (part = 1 ; part <= no_parts ; part++)
	if (!arts[part]) {
	    XtFree((char *)arts);
	    XtFree(file_name);
	    set_message("Error: Couldn't find all parts!", True);
	    return;
	}

    if (do_decode(arts + 1, no_parts)) {
	int	fixed_author = GroupObjectFixedAuthor(global.group_object);

	for (part = 0 ; part <= no_parts ; part++) {
	    if (arts[part] && arts[part]->from && !arts[part]->read) {
		arts[part]->read = True;
		arts[part]->subject->no_unread--;
		if (GroupObjectProcessXrefs(global.group_object))
		    process_xref(arts[part]);

		if (arts[part]->hot > 0)
		    update_subj_hot_value(arts[part]->subject);
		update_subj_entry(arts[part]->subject, fixed_author);
		if (global.mode == NewsModeThread) {
                    ArtTreeNodeSetInner(main_widgets.arttree,
                                        (ART_TREE_NODE *)arts[part], False);
                    ArtTreeNodeSetPixmap(main_widgets.arttree,
                                         (ART_TREE_NODE *)arts[part], None);
		}
	    }
	}
    }

    XtFree(file_name);
    XtFree((char *)arts);
}

void action_uudecode(Widget w, XEvent *event,
		     String *params, Cardinal *no_params)
{
    uudecode_callback(w, NULL, NULL);
}
