/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include "context.h"
#include "save.h"
#include "server.h"
#include "viewer.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Compat.h"
#include "../Widgets/Dialogue.h"
#include "../Widgets/Message.h"

extern void unref_viewable(VIEWABLE *v)
{
    if (v && --v->ref_count <= 0) {
	XtFree(v->data);
	v->data = NULL;
	v->len = 0;
	XtFree(v->viewer);
	v->viewer = NULL;
	v->ref_count = 0;
	XtFree((char *)v);
    }
}

static void mime_dialogue_callback(Widget w,
				   XtPointer client_data,
				   XtPointer call_data)
{
    DialogueReport	*report = (DialogueReport *)call_data;
    VIEWABLE		*viewable = (VIEWABLE *)client_data;

    switch (report->reply) {
    case DialogueReplyLeft:   /* save */
	if (report->buffer && report->buffer[0] != '\0') {
	    int		fd;
	    char	*data = viewable->data;
	    long	n = viewable->len;

	    fd = open_expand(report->buffer, O_WRONLY|O_CREAT|O_TRUNC, 0644);
	    if (fd < 0) {
		perror("knews: open");
		set_message("Error: Failed to open file!", True);
		return;
	    }

	    while (n > 0) {
		long	i = write(fd, data, n);

		if (i < 0) {
		    perror("knews: write");
		    set_message("Error: Failed to write file!", True);
		    (void)close(fd);
		    return;
		}

		n -= i;
		data += i;
	    }

	    if (close(fd) < 0) {
		perror("knews: close");
		set_message("Error: Failed to write file!", True);
		return;
	    }

	    set_message("Saved OK.", False);
	}
	break;
    case DialogueReplyMiddle: /* pipe */
	if (report->buffer && report->buffer[0] != '\0') {
	    pid_t	pid;
	    char	*cmd;

	    cmd = XtNewString(report->buffer);
	    pid = fork_nicely(cmd, pipe_context_callback,
			      global.stderr_timeout >= 0);

	    if (pid < 0) {
		set_message("Error: fork failed!", True);
		XtFree(cmd);
		return;
	    } else if (pid == 0) { /* child */
		char	*file_name = tmpnam(NULL);
		int	fd;

		if (!file_name) _exit(127);

		(void)unlink(file_name);
		fd = open(file_name, O_RDWR|O_CREAT|O_TRUNC, 0600L);
		if (fd < 0) {
		    perror("knews: open");
		    _exit(127);
		}

		while (viewable->len > 0) {
		    long	i;

		    i = write(fd, viewable->data, viewable->len);
		    if (i < 0) {
			perror("knews: write");
			if (errno == EINTR)
			    continue;
			else
			    _exit(127);
		    }
		    viewable->data += i;
		    viewable->len -= i;
		}

		(void)unlink(file_name);

		if (lseek(fd, SEEK_SET, 0) < 0) {
		    perror("knews: lseek");
		    _exit(127);
		}

		if (fd != STDIN_FILENO) {
		    if (dup2(fd, STDIN_FILENO) < 0) {
			perror("knews: dup2");
			_exit(127);
		    }
		    (void)close(fd);
		}

		(void)execl("/bin/sh", "sh", "-c",
			    report->buffer, (char *)NULL);
		perror("knews: execl");
		_exit(127);
	    }

	    set_message("Pipe started.", False);
	}
	break;
    case DialogueReplyEnter:  /* do nothing */
	/* maybe we should default to either save or pipe ? */
	return; /* don't fall through */
    case DialogueReplyTab:
	return; /* don't fall through */
    case DialogueReplyRight:
    case DialogueReplyClose:  /* cancel*/
	break;
    }

    unref_viewable(viewable);
    XtDestroyWidget(w);
}

void click_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
    VIEWABLE	*viewable = (VIEWABLE *)client_data;
    int		clicked = (int)call_data;

    if (!clicked)
	unref_viewable(viewable);
    else if (viewable->viewer) {
	char	*cmd;
	pid_t	pid;

	cmd = XtNewString(viewable->viewer);
	pid = fork_nicely(cmd, pipe_context_callback, True);

	if (pid < 0) {
	    perror("knews: fork");
	    set_message("Failed to start viewer!", True);
	    return;
	}

	if (pid == 0) { /* child */
	    char	*file_name = tmpnam(NULL);
	    char	*viewer;
	    int		fd;

	    if (!file_name)
		_exit(127);

	    (void)unlink(file_name);
	    fd = open(file_name, O_RDWR|O_CREAT|O_TRUNC, 0600L);
	    if (fd < 0) {
		perror("knews: open");
		_exit(127);
	    }

	    while (viewable->len > 0) {
		long	i;

		i = write(fd, viewable->data, viewable->len);
		if (i < 0) {
		    perror("knews: write");
		    if (errno == EINTR)
			continue;
		    else
			_exit(127);
		}
		viewable->data += i;
		viewable->len -= i;
	    }

	    viewer = viewable->viewer;
	    if (strstr(viewer, "%s")) {
		int	len, n;
		char	*c;

		n = 0;
		len = 1024;
		c = viewer;
		viewer = malloc(len);
		if (!viewer)
		    _exit(127);

		while (*c != '\0') {
		    if (c[0] == '%' && c[1] == 's') {
			int	i = strlen(file_name);

			if (n + i + 8 > len) {
			    len = n + i + 32;
			    viewer = malloc(len);
			    if (!viewer)
				_exit(127);
			}

			sprintf(viewer + n, "%s", file_name);
			n = strlen(viewer);
			c += 2;
		    } else {
			if (n + 8 > len) {
			    len *= 2;
			    viewer = malloc(len);
			    if (!viewer)
				_exit(127);
			}

			viewer[n++] = *c++;
		    }
		}
		viewer[n] = '\0';
	    } else {
		(void)unlink(file_name);

		if (lseek(fd, SEEK_SET, 0) < 0) {
		    perror("knews: lseek");
		    _exit(127);
		}

		if (fd != STDIN_FILENO) {
		    if (dup2(fd, STDIN_FILENO) < 0) {
			perror("knews: dup2");
			_exit(127);
		    }
		    (void)close(fd);
		}
	    }

	    (void)execl("/bin/sh", "sh", "-c", viewer, (char *)NULL);
	    perror("knews: execl");
	    _exit(127);
	}

	/* parent */
	set_message("Viewer started.", False);
    } else {
	viewable->ref_count++;
	set_message("No viewer registered for that Content-Type, "
		    "using internal viewer.", False);
	DialogueCreateAndPopup("mimedialogue", main_widgets.shell,
			       "Save to file or pipe to shell:",
			       "Save", "Pipe", "Cancel",
			       mime_dialogue_callback,
			       (XtPointer)viewable,
			       XtGrabNone);
    }
}

char *get_viewer(char *type, char *subtype, Arg *args)
{
    static XrmQuark	Knews = NULLQUARK;
    static XrmQuark	Mime, MimeType, MimeSubtype, Charset, AccessType;
    static XrmQuark	name, mime, text, message, extbody;
    static XrmDatabase	db;
    XrmName		name_list[6];
    XrmClass		class_list[6];
    XrmQuark		rep;
    XrmValue		val;
    char		*result, *src;
    long		i, n, len;

    if (Knews == NULLQUARK) {
	Knews       = XrmPermStringToQuark("Knews");
	Mime        = XrmPermStringToQuark("Mime");
	MimeType    = XrmPermStringToQuark("MimeType");
	MimeSubtype = XrmPermStringToQuark("MineSubtype");
	Charset     = XrmPermStringToQuark("Charset");
	AccessType  = XrmPermStringToQuark("AcessType");
	name        = XrmStringToQuark(XtName(main_widgets.shell));
	mime        = XrmPermStringToQuark("mime");
	text	    = XrmPermStringToQuark("text");
	message	    = XrmPermStringToQuark("message");
	extbody     = XrmPermStringToQuark("external-body");
#if XtSpecificationRelease > 4
	db = XtScreenDatabase(XtScreen(main_widgets.shell));
#else
	db = XtDatabase(XtDisplay(main_widgets.shell));
#endif
    }

    for (;;) { /* one pass loop to break out of */
	class_list[0] = Knews;
	class_list[1] = Mime;
	class_list[2] = MimeType;
	class_list[3] = MimeSubtype;

	name_list[0] = name;
	name_list[1] = mime;
	name_list[2] = XrmStringToQuark(type);
	name_list[3] = XrmStringToQuark(subtype);

	if (name_list[2] == text) {
	    for (i = 0 ; args[i].value ; i++)
		if (strcmp(args[i].name, "charset") == 0)
		    break;

	    if (args[i].value) {
		class_list[4] = Charset;
		class_list[5] = NULLQUARK;

		name_list[4] = XrmStringToQuark((char *)args[i].value);
		name_list[5] = NULLQUARK;

		if (XrmQGetResource(db, name_list, class_list, &rep, &val))
		    break;
	    }
	} else if (name_list[2] == message &&
		   name_list[3] == extbody) {
	    for (i = 0 ; args[i].value ; i++)
		if (strcmp(args[i].name, "access-type") == 0)
		    break;

	    if (args[i].value) {
		class_list[4] = AccessType;
		class_list[5] = NULLQUARK;

		name_list[4] = XrmStringToQuark((char *)args[i].value);
		name_list[5] = NULLQUARK;

		if (XrmQGetResource(db, name_list, class_list, &rep, &val))
		    break;
	    }
	}

	class_list[4] = NULLQUARK;
	name_list[4] = NULLQUARK;

	if (XrmQGetResource(db, name_list, class_list, &rep, &val))
	    break;

	class_list[3] = NULLQUARK;
	name_list[3] = NULLQUARK;

	if (XrmQGetResource(db, name_list, class_list, &rep, &val))
	    break;

	return NULL;
    }

    len = 128;
    result = XtMalloc(128);
    for (src = (char *)val.addr, n = 0 ; *src != '\0' ; ) {
	char	c = *src++;

	if (n + 8 > len) {
	    len += 32;
	    result = XtRealloc(result, len);
	}

	if (c == '%') {
	    char	*p = NULL;

	    c = *src++;
	    switch (c) {
	    case 't':
		i = n + strlen(type) + strlen(subtype) + 1;
		if (i + 8 > len) {
		    len = i + 8;
		    result = XtRealloc(result, len);
		}
		sprintf(result + n, "%s/%s", type, subtype);
		n = i;
		continue;
	    case '{':
		p = strchr(src, '}');
		break; /* fall through */
	    case '[':
		p = strchr(src, ']');
		break; /* fall through */
	    case '(':
		p = strchr(src, '}');
		break; /* fall through */
	    case '%':
		result[n++] = '%';
		continue;
	    case 's':
		result[n++] = '%';
		result[n++] = 's';
		continue;
	    default:
		XtFree(result);
		return NULL;
	    }

	    if (p) {
		int	j = p - src;

		for (i = 0 ; args[i].value ; i++)
		    if (strlen(args[i].name) == j &&
			strncasecmp(src, args[i].name, j) == 0)
			break;

		if (args[i].value) {
		    j = strlen((char *)args[i].value);
		    if (n + j + 8 > len) {
			len = n + j + 32;
			result = XtRealloc(result, len);
		    }
		    memcpy(result + n, (void *)args[i].value, j);
		    n += j;
		} else switch (c) {
		case '{': /* crucial arg */
		    XtFree(result);
		    return NULL;
		case '[': /* optional arg */
		    break;
		case '(': /* optional arg, use '' if not available */
		    result[n++] = '\'';
		    result[n++] = '\'';
		    break;
		}
		src = p + 1;
		continue;
	    } else {
		XtFree(result);
		return NULL;
	    }
	}

	result[n++] = c;
    }

    result[n] = '\0';

    return result;
}

char *get_server_command(char *res_name)
{
    XrmName	name_list[3];
    XrmClass	class_list[3];
    XrmQuark	rep;
    XrmDatabase	db;
    XrmValue	val;

    while (*res_name == ' ' ||
	   *res_name == '\t')
	res_name++;

    if (*res_name == '\0' ||
	strchr(res_name, '.') ||
	strchr(res_name, '*') ||
	strchr(res_name, ' ') ||
	strchr(res_name, '\t') ||
	strchr(res_name, '/'))
	return NULL;

    class_list[0] = XrmPermStringToQuark("Knews");
    class_list[1] = XrmPermStringToQuark("Server");
    class_list[2] = NULLQUARK;

    name_list[0] = XrmStringToQuark(XtName(main_widgets.shell));
    name_list[1] = XrmStringToQuark(res_name);
    name_list[2] = NULLQUARK;

#if XtSpecificationRelease > 4
    db = XtScreenDatabase(XtScreen(main_widgets.shell));
#else
    db = XtDatabase(XtDisplay(main_widgets.shell));
#endif

    if (XrmQGetResource(db, name_list, class_list, &rep, &val))
	return (char *)val.addr;
    else
	return NULL;
}
