/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xmu/Xmu.h>
#if !defined(__hpux) && XtSpecificationRelease > 4
#include <X11/Xmu/Editres.h>
#endif
#include "global.h"
#include "ahead.h"
#include "kill.h"
#include "misc.h"
#include "newsrc.h"
#include "pixmaps.h"
#include "pedit.h"
#include "procs.h"
#include "tag.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtText.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/Knapp.h"
#include "../Widgets/Layout.h"
#include "../Widgets/Manager.h"
#include "../Widgets/MenuKnapp.h"
#include "../Widgets/Message.h"
#include "../Widgets/Sash.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "../Widgets/Util.h"

#define DESCRIPTION(desc)	((desc) ? (desc) : "")
#define ARTTREE_STEPSIZE	16
#define LINE_LEN		128

struct _main_widgets main_widgets;

static void print_group_info_subscr(char *buffer, GROUP_NODE *group)
{
    sprintf(buffer, "%c %-42.42s %6ld %-64.64s",
	    group->ahead ? thread_ahead_char(group->ahead) : ' ',
	    group->name, group->no_unread,
	    DESCRIPTION(group->description));
}

static void print_group_info_all(char *buffer, GROUP_NODE *group)
{
    long	num;

    if (group->last_art == 0)
	num = 0;
    else {
	num = group->last_art - group->first_art + 1;
	if (num < 0)
	    num = 0;
    }

    sprintf(buffer, "%3s %c %-42.42s %6ld %-64.64s",
	    group->subscribed ? "[s]" : "   ",
	    group->ahead ? thread_ahead_char(group->ahead) : ' ',
	    group->name,
	    num < 0 ? 0 : num,
	    DESCRIPTION(group->description));
}

static void print_subj_info(char *buffer, SUBJECT *subj, int fixed_author)
{
    Boolean	indent = False;
    SUBJECT	*loop = subj->prev;
    char	*author;

    if (subj->no_unread == 0)
	sprintf(buffer, "     ");
    else
	sprintf(buffer, "%3ld  ", subj->no_unread);
    if (subj->has_tagged) buffer[3] = '*';

    while (loop && loop->thread == subj->thread) {
	if (loop->disp >= 0) {
	    indent = True;
	    break;
	}
	loop = loop->prev;
    }

    if (fixed_author || subj->no_unread == 0)
	author = first_author_with_subject(subj);
    else
	author = first_unread_author_with_subject(subj);
    if (!author)
	author = "";

    if (indent)
	sprintf(&buffer[5], "  %-54.54s %-64.64s", subj->subject, author);
    else
	sprintf(&buffer[5], "%-56.56s %-64.64s", subj->subject, author);
}

static void knapp_set_mode(NewsMode mode)
{
    Arg	arg;

    switch (mode) {
    case NewsModeDisconnected:
	XtSetMappedWhenManaged(main_widgets.knapp[2], False);
	XtSetMappedWhenManaged(main_widgets.knapp[6], False);
	XtSetMappedWhenManaged(main_widgets.knapp[7], False);
	XtSetMappedWhenManaged(main_widgets.knapp[8], False);

	KnappSetLabel(main_widgets.knapp[0], "Quit", True);
	KnappSetLabel(main_widgets.knapp[1], "Connect...", True);
	KnappSetLabel(main_widgets.knapp[3], NULL, False);
	KnappSetLabel(main_widgets.knapp[5], NULL, False);
	KnappSetLabel(main_widgets.knapp[10], NULL, False);
	KnappSetLabel(main_widgets.knapp[11], NULL, False);

	XtSetArg(arg, XtNmenuName, "miscshell2");
	break;
    case NewsModeConnected:
	XtSetMappedWhenManaged(main_widgets.knapp[2], True);
	XtSetMappedWhenManaged(main_widgets.knapp[6], True);
	XtSetMappedWhenManaged(main_widgets.knapp[7], True);
	XtSetMappedWhenManaged(main_widgets.knapp[8], True);

	KnappSetLabel(main_widgets.knapp[0], "Quit", True);
	KnappSetLabel(main_widgets.knapp[1], "Disconnect", True);
	KnappSetLabel(main_widgets.knapp[2], "All groups", True);
	KnappSetLabel(main_widgets.knapp[3], NULL, True);
	KnappSetLabel(main_widgets.knapp[5], NULL, True);
	KnappSetLabel(main_widgets.knapp[6], "Update", True);
	KnappSetLabel(main_widgets.knapp[7], "Read group", True);
	KnappSetLabel(main_widgets.knapp[8], "Rescan", True);
	KnappSetLabel(main_widgets.knapp[10], NULL, False);
	KnappSetLabel(main_widgets.knapp[11], NULL, False);

	XtSetArg(arg, XtNmenuName, "miscshell2");
	break;
    case NewsModeGroup:
	XtSetMappedWhenManaged(main_widgets.knapp[2], True);
	XtSetMappedWhenManaged(main_widgets.knapp[6], True);
	XtSetMappedWhenManaged(main_widgets.knapp[7], True);
	XtSetMappedWhenManaged(main_widgets.knapp[8], True);

	KnappSetLabel(main_widgets.knapp[0], "Done", True);
	KnappSetLabel(main_widgets.knapp[1], "View thread", True);
	KnappSetLabel(main_widgets.knapp[2], "All threads", True);
	KnappSetLabel(main_widgets.knapp[3], NULL, True);
	KnappSetLabel(main_widgets.knapp[5], NULL, True);
	KnappSetLabel(main_widgets.knapp[6], "Catchup", True);
	KnappSetLabel(main_widgets.knapp[7], "Next depth", True);
	KnappSetLabel(main_widgets.knapp[8], "Previous", True);
	KnappSetLabel(main_widgets.knapp[10], NULL, True);
	KnappSetLabel(main_widgets.knapp[11], NULL, True);

	XtSetArg(arg, XtNmenuName, "miscshell1");
	break;
    case NewsModeThread:
	XtSetMappedWhenManaged(main_widgets.knapp[2], True);
	XtSetMappedWhenManaged(main_widgets.knapp[6], True);
	XtSetMappedWhenManaged(main_widgets.knapp[7], True);
	XtSetMappedWhenManaged(main_widgets.knapp[8], True);

	KnappSetLabel(main_widgets.knapp[0], "Done", False);
	KnappSetLabel(main_widgets.knapp[1], "Back", True);
	KnappSetLabel(main_widgets.knapp[2], "All threads", False);
	KnappSetLabel(main_widgets.knapp[3], NULL, True);
	KnappSetLabel(main_widgets.knapp[5], NULL, True);
	KnappSetLabel(main_widgets.knapp[6], "Catchup", False);
	KnappSetLabel(main_widgets.knapp[7], "Next depth", True);
	KnappSetLabel(main_widgets.knapp[8], "Previous", True);
	KnappSetLabel(main_widgets.knapp[10], NULL, True);
	KnappSetLabel(main_widgets.knapp[11], NULL, True);

	XtSetArg(arg, XtNmenuName, "miscshell1");
	break;
    case NewsModeAllgroups:
	XtSetMappedWhenManaged(main_widgets.knapp[2], True);
	XtSetMappedWhenManaged(main_widgets.knapp[6], True);
	XtSetMappedWhenManaged(main_widgets.knapp[7], True);
	XtSetMappedWhenManaged(main_widgets.knapp[8], False);

	KnappSetLabel(main_widgets.knapp[0], "Done", True);
	KnappSetLabel(main_widgets.knapp[1], "Subscribe", True);
	KnappSetLabel(main_widgets.knapp[2], "Unsubscribe", True);
	KnappSetLabel(main_widgets.knapp[3], NULL, True);
	KnappSetLabel(main_widgets.knapp[5], NULL, True);
	KnappSetLabel(main_widgets.knapp[6], "Update", True);
	KnappSetLabel(main_widgets.knapp[7], "Goto group", True);
	KnappSetLabel(main_widgets.knapp[10], NULL, False);
	KnappSetLabel(main_widgets.knapp[11], NULL, True);

	XtSetArg(arg, XtNmenuName, "miscshell2");
	break;
    case NewsModeNewgroups:
	XtSetMappedWhenManaged(main_widgets.knapp[2], True);
	XtSetMappedWhenManaged(main_widgets.knapp[6], False);
	XtSetMappedWhenManaged(main_widgets.knapp[7], False);
	XtSetMappedWhenManaged(main_widgets.knapp[8], False);

	KnappSetLabel(main_widgets.knapp[0], "Done", True);
	KnappSetLabel(main_widgets.knapp[1], "Subscribe", True);
	KnappSetLabel(main_widgets.knapp[2], "Unsubscribe", True);
	KnappSetLabel(main_widgets.knapp[3], NULL, False);
	KnappSetLabel(main_widgets.knapp[5], NULL, True);
	KnappSetLabel(main_widgets.knapp[10], NULL, False);
	KnappSetLabel(main_widgets.knapp[11], NULL, False);

	XtSetArg(arg, XtNmenuName, "miscshell2");
	break;
    }

    XtSetValues(main_widgets.knapp[3], &arg, 1);
}

void setNewsModeDisconnected()
{
    global.mode = NewsModeDisconnected;
    XtUnmanageChild(main_widgets.arttree);
    XtUnmanageChild(main_widgets.thread_list);
    XtUnmanageChild(main_widgets.group_list);
    knapp_set_mode(NewsModeDisconnected);
    ArtTextClearLines(main_widgets.text);
    ScrBarSetLengthsAndPos(main_widgets.top_vbar, 0, 0, 0);
    ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);
    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);
}

void update_group_entry(GROUP_NODE *group)
{
    char	buffer[LINE_LEN];

    switch (global.mode) {
    case NewsModeConnected:
	if (group->disp >= 0) {
	    print_group_info_subscr(buffer, group);
	    ScrListSetLine(main_widgets.group_list, group->disp,
			   buffer, None);
	}
	break;
    case NewsModeAllgroups:
	if (group->disp >= 0) {
	    print_group_info_all(buffer, group);
	    ScrListSetLine(main_widgets.group_list, group->disp,
			   buffer, None);
	}
	break;
    case NewsModeDisconnected:
    case NewsModeNewgroups:
    case NewsModeGroup:
    case NewsModeThread:
	break;
    }
}

void setNewsModeConnected()
{
    Arg		args[4];
    long	i, n, sel = -1;
    long	shown, first;
    Boolean	flag = False;

    global.mode = NewsModeConnected;

    XtUnmanageChild(main_widgets.arttree);
    XtUnmanageChild(main_widgets.thread_list);
    ArtTextClearLines(main_widgets.text);
    ScrListClearLines(main_widgets.thread_list);
    ScrListClearLines(main_widgets.group_list);
    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);
    ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);
    XtSetArg(args[0], XtNstepSize, 1);
    XtSetValues(main_widgets.top_vbar, args, 1);
    ScrBarSetLengthsAndPos(main_widgets.top_vbar, 0, 0, 0);
    knapp_set_mode(NewsModeConnected);
    XtSetArg(args[0], XtNatMostOne, True);
    XtSetArg(args[1], XtNatLeastOne, True);
    XtSetArg(args[2], XtNallowDnd, False);
    XtSetValues(main_widgets.group_list, args, 3);
    XtManageChild(main_widgets.group_list);

    for (n = 0 ; n < global.no_groups ; n++) {
	GROUP_NODE	*group = global.groups[n];
	char		buffer[LINE_LEN];

	if (!group->subscribed)
	    break;

	calc_no_unread(group);
	if (group == global.thr.group) flag = True;
	if (group->no_unread > 0) {
	    print_group_info_subscr(buffer, group);
	    group->disp = ScrListAddLine(main_widgets.group_list,
					 buffer, None);
	    if (flag) {
		sel = group->disp;
		flag = False;
	    }
	} else {
	    group->disp = -1;
	}
    }

    if (sel < 0) {
	global.thr.group = NULL;
	for (n = 0 ; n < global.no_groups ; n++) {
	    if (!global.groups[n]->subscribed) {
		sel = 0;
		break;
	    } else if (global.groups[n]->disp == 0) {
		global.thr.group = global.groups[n];
		sel = 0;
		break;
	    }
	}
    }

    if (sel >= 0)
	ScrListSetSelected(main_widgets.group_list, sel, True);
    ScrListGetFirstShownSize(main_widgets.group_list, NULL, &shown, &i);
    first = 0;
    if (sel >= shown) {
	first = sel;
	if (first + shown >= n) {
	    first = i - shown;
	    if (first < 0) first = 0;
	}
	ScrListSetFirst(main_widgets.group_list, first);
    }
    ScrBarSetLengthsAndPos(main_widgets.top_vbar, i, shown, first);

    set_standard_message();
}

void update_subj_entry(SUBJECT *subj, int fixed_author)
{
    long	row = subj->disp;
    char	buffer[LINE_LEN];

    if (row >= 0) {
	print_subj_info(buffer, subj, fixed_author);
	ScrListSetLine(main_widgets.thread_list, row, buffer,
		       global.hot_pixmaps[subj->hot]);
    }
}

void setNewsModeGroup(int show_all)
{
    Arg		args[8];
    SUBJECT	*loop;
    long	n, shown;
    int		fixed_author = GroupObjectFixedAuthor(global.group_object);

    if (global.mode != NewsModeThread) {
	global.mode = NewsModeGroup;
	XtUnmanageChild(main_widgets.arttree);
	XtUnmanageChild(main_widgets.group_list);
	ScrListClearLines(main_widgets.group_list);
	ScrListClearLines(main_widgets.thread_list);
	XtManageChild(main_widgets.thread_list);
	XtSetArg(args[0], XtNtree, NULL);
	XtSetValues(main_widgets.arttree, args, 1);
	knapp_set_mode(NewsModeGroup);
	ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);

	for (n = 0, loop = global.thr.subjects ; loop ; loop = loop->next) {
	    char	buffer[LINE_LEN];

	    if (!show_all && loop->no_unread == 0) {
		loop->disp = -1;
		continue;
	    }

	    print_subj_info(buffer, loop, fixed_author);
	    loop->disp =
		ScrListAddLine(main_widgets.thread_list, buffer,
			       global.hot_pixmaps[loop->hot]);
	    if (n == 0)
		global.curr_subj = loop;

	    n++;
	}

	ScrListGetFirstShownSize(main_widgets.thread_list, NULL, &shown, &n);
	XtSetArg(args[0], XtNstepSize, 1);
	XtSetValues(main_widgets.top_vbar, args, 1);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar, n, shown, 0);

	/* message already set by get_articles */
    } else {
	long	first;

	global.mode = NewsModeGroup;
	XtUnmanageChild(main_widgets.arttree);
	XtSetArg(args[0], XtNtree, NULL);
	XtSetValues(main_widgets.arttree, args, 1);
	XtManageChild(main_widgets.thread_list);
	set_curr_subj(global.curr_subj);
	ScrListGetFirstShownSize(main_widgets.thread_list, &first, &shown, &n);
	ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar, n, shown, first);
	XtSetArg(args[0], XtNstepSize, 1);
	XtSetValues(main_widgets.top_vbar, args, 1);
	knapp_set_mode(NewsModeGroup);
	set_standard_message();
    }
}

void setNewsModeThread()
{
    Arg		args[8];
    Dimension	s_w, s_h, w, h;

    global.mode = NewsModeThread;
    XtUnmanageChild(main_widgets.group_list);
    XtUnmanageChild(main_widgets.thread_list);
    XtSetArg(args[0], XtNtree, NULL);
    XtSetValues(main_widgets.arttree, args, 1);
    XtSetArg(args[0], XtNtree, global.curr_subj->thread);
    XtSetArg(args[1], XtNsuspend, True);
    XtSetArg(args[2], XtNx, 0);
    XtSetArg(args[3], XtNy, 0);
    XtSetValues(main_widgets.arttree, args, 4);
    set_tree_stuff(global.curr_subj->thread);
    mark_tagged_articles(global.curr_subj->thread);
    XtSetArg(args[0], XtNwidth, &w);
    XtSetArg(args[1], XtNheight, &h);
    XtGetValues(main_widgets.arttree, args, 2);
    XtSetArg(args[0], XtNwidth, &s_w);
    XtSetArg(args[1], XtNheight, &s_h);
    XtGetValues(main_widgets.top_manager, args, 2);
    ScrBarSetLengthsAndPos(main_widgets.top_hbar, w, s_w, 0);
    ScrBarSetLengthsAndPos(main_widgets.top_vbar, h, s_h, 0);
    XtSetArg(args[0], XtNstepSize, ARTTREE_STEPSIZE);
    XtSetValues(main_widgets.top_hbar, args, 1);
    XtSetValues(main_widgets.top_vbar, args, 1);
    if (global.curr_art) {
	ArtTreeNodeMakeVisible(main_widgets.arttree,
			       (ART_TREE_NODE *)global.curr_art);
	ArtTreeNodeSetSelected(main_widgets.arttree,
			       (ART_TREE_NODE *)global.curr_art, True);
    }
    XtManageChild(main_widgets.arttree);
    XtSetArg(args[0], XtNsuspend, False);
    XtSetValues(main_widgets.arttree, args, 1);
    knapp_set_mode(NewsModeThread);
    set_standard_message();
}

void setNewsModeAllgroups()
{
    Arg		args[8];
    long	n = global.no_groups;
    long	shown, first = -1;

    global.mode = NewsModeAllgroups;

    XtSetArg(args[0], XtNnAlloc, global.no_groups + 4);
    XtSetValues(main_widgets.group_list, args, 1);

    XtUnmanageChild(main_widgets.arttree);
    XtUnmanageChild(main_widgets.thread_list);
    ScrListClearLines(main_widgets.thread_list);
    ArtTextClearLines(main_widgets.text);
    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);
    knapp_set_mode(NewsModeAllgroups);
    ScrListClearLines(main_widgets.group_list);
    XtSetArg(args[0], XtNatMostOne, False);
    XtSetArg(args[1], XtNatLeastOne, False);
    XtSetArg(args[2], XtNallowDnd, True);
    XtSetValues(main_widgets.group_list, args, 3);
    XtManageChild(main_widgets.group_list);
    XtSetArg(args[0], XtNstepSize, 1);
    XtSetValues(main_widgets.top_vbar, args, 1);
    ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);

    for (n = 0 ; n < global.no_groups ; n++) {
	GROUP_NODE	*group = global.groups[n];
	char		buffer[LINE_LEN];

	if (group == global.thr.group)
	    first = n;
	print_group_info_all(buffer, group);
	group->disp =
	    ScrListAddLine(main_widgets.group_list, buffer, None);
    }

    ScrListGetFirstShownSize(main_widgets.group_list, NULL, &shown, &n);
    if (first < 0)
	first = 0;
    else {
	ScrListSetSelected(main_widgets.group_list, first, True);
	if (first > shown && first > n - shown)
	    first = n - shown;
	ScrListSetFirst(main_widgets.group_list, first);
    }
    ScrBarSetLengthsAndPos(main_widgets.top_vbar, n, shown, first);

    set_standard_message();
}

void setNewsModeNewgroups()
{
    Arg		args[8];
    long	n;
    long	shown;

    global.mode = NewsModeNewgroups;

    for (n = 0 ; n < global.no_groups ; n++)
	global.groups[n]->disp = -1;

    XtSetArg(args[0], XtNnAlloc, global.no_new_groups + 4);
    XtSetValues(main_widgets.group_list, args, 1);

    XtUnmanageChild(main_widgets.arttree);
    XtUnmanageChild(main_widgets.thread_list);
    ScrListClearLines(main_widgets.thread_list);
    ScrListClearLines(main_widgets.group_list);
    ArtTextClearLines(main_widgets.text);
    ScrBarSetLengthsAndPos(main_widgets.text_scrbar, 0, 0, 0);
    knapp_set_mode(NewsModeNewgroups);
    XtSetArg(args[0], XtNatMostOne, False);
    XtSetArg(args[1], XtNatLeastOne, False);
    XtSetArg(args[2], XtNallowDnd, False);
    XtSetValues(main_widgets.group_list, args, 3);
    XtManageChild(main_widgets.group_list);
    XtSetArg(args[0], XtNstepSize, 1);
    XtSetValues(main_widgets.top_vbar, args, 1);
    ScrBarSetLengthsAndPos(main_widgets.top_hbar, 0, 0, 0);

    for (n = 0 ; n < global.no_new_groups ; n++) {
	GROUP_NODE	*group = global.new_groups[n];
	char		buffer[LINE_LEN];

	print_group_info_all(buffer, group);
	group->disp = ScrListAddLine(main_widgets.group_list, buffer, None);
    }

    ScrListGetFirstShownSize(main_widgets.group_list, NULL, &shown, NULL);
    ScrBarSetLengthsAndPos(main_widgets.top_vbar,
			   global.no_new_groups, shown, 0);

    set_message("New groups found.", False);
}

void create_main_widgets(Widget main_shell)
{
    Widget	main_layout, knapp_layout, text_layout;
    Widget	sash;
    Arg		args[10];
    Boolean	one_win = !global.separate_windows;

    main_widgets.shell = main_shell;
    main_widgets.save_shell = NULL;
    main_widgets.search_shell = NULL;
    main_widgets.kill_editor_shell = NULL;

    if (one_win) {
	main_widgets.second_shell = NULL;
	main_layout =
	    XtCreateManagedWidget("singlelayout", layoutWidgetClass,
				  main_widgets.shell, NULL, 0);
	text_layout =
	    XtCreateManagedWidget("textlayout", layoutWidgetClass,
				  main_layout, NULL, 0);
    } else {
	XtSetArg(args[0], XtNinput, True);
	main_widgets.second_shell =
	    XtCreatePopupShell("second", topLevelShellWidgetClass,
			       main_widgets.shell, args, 1);
	main_layout =
	    XtCreateManagedWidget("doublelayout", layoutWidgetClass,
				  main_widgets.shell, NULL, 0);
	text_layout =
	    XtCreateManagedWidget("textlayout", layoutWidgetClass,
				  main_widgets.second_shell, NULL, 0);
    }

    main_widgets.top_layout =
	XtCreateManagedWidget("toplayout", layoutWidgetClass,
			      main_layout, NULL, 0);
    XtSetArg(args[0], XtNcenter, False);
    main_widgets.message_line =
	XtCreateManagedWidget("message", messageWidgetClass,
			      main_layout, args, 1);
    knapp_layout =
	XtCreateManagedWidget("knapplayout", layoutWidgetClass,
			      main_layout, NULL, 0);

    main_widgets.top_manager =
	XtCreateManagedWidget("topmanager", managerWidgetClass,
			      main_widgets.top_layout, NULL, 0);

    sash = XtCreateManagedWidget("sash1", sashWidgetClass,
				 main_layout, NULL, 0);
    XtAddCallback(sash, XtNcallback, sash_callback, NULL);
    sash = XtCreateManagedWidget("sash2", sashWidgetClass,
				 main_layout, NULL, 0);
    XtAddCallback(sash, XtNcallback, sash_callback, NULL);
    if (one_win) {
	sash = XtCreateManagedWidget("sash3", sashWidgetClass,
				     main_layout, NULL, 0);
	XtAddCallback(sash, XtNcallback, sash_callback, NULL);
    }

    XtSetArg(args[0], XtNcoverHoriz, True);
    XtSetArg(args[1], XtNcoverVert, True);
    XtSetArg(args[2], XtNcontainHoriz, False);
    XtSetArg(args[3], XtNcontainVert, False);
    XtSetArg(args[4], XtNpixmapWidth, HOT_PIXMAP_SIZE);
    XtSetArg(args[5], XtNpixmapHeight, HOT_PIXMAP_SIZE);
    XtSetArg(args[6], XtNdepthOne, False);
    main_widgets.arttree =
	XtCreateWidget("arttree", artTreeWidgetClass,
		       main_widgets.top_manager, args, 7);
    args[2].value = (XtArgVal)True;
    args[3].value = (XtArgVal)True;
    XtSetArg(args[7], XtNusePixmaps, False);
    main_widgets.group_list =
	XtCreateManagedWidget("grouplist", scrListWidgetClass,
			      main_widgets.top_manager, args, 8);
    XtSetArg(args[7], XtNusePixmaps, True);
    XtSetArg(args[8], XtNatLeastOne, True);
    XtSetArg(args[9], XtNatMostOne, True);
    main_widgets.thread_list =
	XtCreateWidget("threadlist", scrListWidgetClass,
		       main_widgets.top_manager, args, 10);

    main_widgets.top_vbar =
	XtCreateManagedWidget("topvbar", scrBarWidgetClass,
			      main_widgets.top_layout, NULL, 0);
    XtSetArg(args[0], XtNorientation, XtorientHorizontal);
    XtSetArg(args[1], XtNstepSize, ARTTREE_STEPSIZE);
    main_widgets.top_hbar =
	XtCreateManagedWidget("tophbar", scrBarWidgetClass,
			      main_widgets.top_layout, args, 1);

    main_widgets.text =
	XtCreateManagedWidget("text", artTextWidgetClass,
			      text_layout, NULL, 0);
    main_widgets.text_scrbar =
	XtCreateManagedWidget("textscrbar", scrBarWidgetClass,
			      text_layout, NULL, 0);

    XtSetArg(args[0], XtNlabel, "      ");
    XtSetArg(args[1], XtNresizable, False);
    main_widgets.knapp[0] =
	XtCreateManagedWidget("knapp0", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "           ");
    main_widgets.knapp[1] =
	XtCreateManagedWidget("knapp1", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "          ");
    main_widgets.knapp[2] =
	XtCreateManagedWidget("knapp2", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "Misc");
    main_widgets.knapp[3] =
	XtCreateManagedWidget("knapp3", menuKnappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "Post");
    XtSetArg(args[2], XtNmenuName, "postshell");
    main_widgets.knapp[4] =
	XtCreateManagedWidget("knapp4", menuKnappWidgetClass,
			      knapp_layout, args, 3);
    XtSetArg(args[0], XtNlabel, "Kill...");
    main_widgets.knapp[5] =
	XtCreateManagedWidget("knapp5", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "      ");
    main_widgets.knapp[6] =
	XtCreateManagedWidget("knapp6", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "           ");
    main_widgets.knapp[7] =
	XtCreateManagedWidget("knapp7", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "            ");
    main_widgets.knapp[8] =
	XtCreateManagedWidget("knapp8", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "Abort");
    XtSetArg(args[2], XtNsensitive, False);
    main_widgets.knapp[9] =
	XtCreateManagedWidget("knapp9", knappWidgetClass,
			      knapp_layout, args, 3);
    XtSetArg(args[0], XtNlabel, "Save...");
    main_widgets.knapp[10] =
	XtCreateManagedWidget("knapp10", knappWidgetClass,
			      knapp_layout, args, 2);
    XtSetArg(args[0], XtNlabel, "Search...");
    main_widgets.knapp[11] =
	XtCreateManagedWidget("knapp11", knappWidgetClass,
			      knapp_layout, args, 2);

    XtRealizeWidget(main_shell);
    add_WM_DELETE_WINDOW_callback(main_shell, delete_window_callback, NULL);
    XtInstallAllAccelerators(main_shell, main_shell);

    if (main_widgets.second_shell) {
	XtRealizeWidget(main_widgets.second_shell);
	add_WM_DELETE_WINDOW_callback(main_widgets.second_shell,
				      delete_window_callback, NULL);
	XtInstallAllAccelerators(main_widgets.second_shell,
				 main_shell);
    }

    create_misc_menu1(main_shell);
    create_misc_menu2(main_shell);
    create_post_menu(main_shell);

    XtAddCallback(main_widgets.knapp[0], XtNcallback,
		  knapp0_callback, NULL);
    XtAddCallback(main_widgets.knapp[1], XtNcallback,
		  knapp1_callback, NULL);
    XtAddCallback(main_widgets.knapp[2], XtNcallback,
		  knapp2_callback, NULL);
    XtAddCallback(main_widgets.knapp[5], XtNcallback,
		  knapp5_callback, NULL);
    XtAddCallback(main_widgets.knapp[6], XtNcallback,
		  knapp6_callback, NULL);
    XtAddCallback(main_widgets.knapp[7], XtNcallback,
		  knapp7_callback, NULL);
    XtAddCallback(main_widgets.knapp[8], XtNcallback,
		  knapp8_callback, NULL);
    XtAddCallback(main_widgets.knapp[9], XtNcallback,
		  knapp9_callback, NULL);
    XtAddCallback(main_widgets.knapp[10], XtNcallback,
		  knapp10_callback, NULL);
    XtAddCallback(main_widgets.knapp[11], XtNcallback,
		  knapp11_callback, NULL);
    XtAddCallback(main_widgets.top_vbar, XtNscrollCallback,
		  top_vbar_callback, NULL);
    XtAddCallback(main_widgets.top_hbar, XtNscrollCallback,
		  top_hbar_callback, NULL);
    XtAddCallback(main_widgets.text_scrbar, XtNscrollCallback,
		  text_scrbar_callback, NULL);
    XtAddCallback(main_widgets.top_manager, XtNresizeCallback,
		  top_manager_callback, NULL);
    XtAddCallback(main_widgets.text, XtNresizeCallback,
		  text_resize_callback, NULL);
    XtAddCallback(main_widgets.arttree, XtNselectCallback,
		  arttree_sel_callback, NULL);
    XtAddCallback(main_widgets.arttree, XtNscrollCallback,
		  arttree_scroll_callback, NULL);
    XtAddCallback(main_widgets.arttree, XtNouterCallback,
		  arttree_tag_callback, NULL);
    XtAddCallback(main_widgets.thread_list, XtNselectCallback,
		  thread_list_sel_callback, NULL);
    XtAddCallback(main_widgets.thread_list, XtNcallback,
		  thread_list_callback, NULL);
    XtAddCallback(main_widgets.group_list, XtNcallback,
		  group_list_callback, NULL);
    XtAddCallback(main_widgets.group_list, XtNselectCallback,
		  group_list_sel_callback, NULL);
    XtAddCallback(main_widgets.group_list, XtNdndCallback,
		  group_list_dnd_callback, NULL);

#if !defined(__hpux) && (XtSpecificationRelease > 4)
    XtAddEventHandler(main_shell, (EventMask)0, True,
		      _XEditResCheckMessages, NULL);
    if (main_widgets.second_shell)
	XtAddEventHandler(main_widgets.second_shell, (EventMask)0,
			  True, _XEditResCheckMessages, NULL);

#endif
}

void purge_hot(short hot)
{
    if (global.mode == NewsModeGroup || global.mode == NewsModeThread) {
	Pixmap	pixmap = global.hot_pixmaps[hot];
	SUBJECT	*subj;
	ARTICLE	*art;

	ScrListPurgePixmap(main_widgets.thread_list, pixmap);

	for (art = global.thr.articles ; art ; art = art->next)
	    if (art->hot == hot) art->hot = 0;

	for (subj = global.thr.subjects ; subj ; subj = subj->next)
	    if (subj->hot == hot) update_subj_hot_value(subj);

	if (global.mode == NewsModeThread && global.curr_subj) {
	    for (art = global.curr_subj->thread ; art ;
		 art = next_in_thread_preorder(art))
		if (ArtTreeNodeGetPixmap(main_widgets.arttree,
					 (ART_TREE_NODE *)art) == pixmap)
		    ArtTreeNodeSetPixmap(main_widgets.arttree,
					 (ART_TREE_NODE *)art, None);

	    update_subj_hot_value(global.curr_subj);
	}
    }
}
