/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include "../src/codes.h"
#include "knewsd.h"
#include "overview.h"

static long	curr_art = -1;
static long	*arts = NULL;
static long	n_arts = 0;
static char	*group_path = NULL;
static int	updating_overview = False;

static char	*spool_dir	= SPOOL_DIR;
static char	*active		= ACTIVE_FILE;
#ifdef NEWSGROUPS_FILE
static char	*newsgroups	= NEWSGROUPS_FILE;
#else
static char	*newsgroups	= NULL;
#endif
#ifdef OVERVIEW_DIR
static char	*overview_dir	= OVERVIEW_DIR;
#else
static char	*overview_dir	= NULL;
#endif
static int	update_overview	= False;

static void sigpipe_handler(int sig)
{
    if (updating_overview &&
	(!overview_dir || chdir(overview_dir) == 0) &&
	chdir(group_path) == 0)
	(void)unlink(".overview");

    _exit(0);
}

static void *xrealloc(void *ptr, size_t n)
{
    if (ptr) /* hack for broken C libraries */
	ptr = realloc(ptr, n);
    else
	ptr = malloc(n);

    if (!ptr) {
	printf(CODE_TO_STR(ERR_FAULT) " Malloc failed.\r\n");
	exit(0);
    }

    return ptr;
}

static int only_digits(char *c)
{
    while (*c != '\0') {
	if (*c < '0' || *c > '9')
	    return False;
	c++;
    }

    return True;
}

static int long_cmp(const void *c1, const void *c2)
{
    return *(long *)c1 - *(long *)c2;
}

static long readn(int fd, char *buffer, int len)
{
    char	*c = buffer;
    int		i, n = 0;

    while (len > 0) {
	i = read(fd, c, len);

	if (i == 0)
	    break;
	if (i < 0)
	    return -1;
	c += i;
	n += i;
	len -= i;
    }

    return n;
}

static long art_search(long *arts, long n_arts, long no)
{
    long	i;

    for (i = 0 ; i < n_arts ; i++)
	if (arts[i] == no)
	    return i;

    return -1;
}

static char *extract_messageid(char *head)
{
    static char	result[256];

    for (;;) {
	if (strncasecmp(head, "Message-Id: ", 12) == 0) {
	    char	*c, *p;
	    int		len;

	    head += 12;
	    if (*head != '<')
		return NULL;

	    c = strchr(head, '>');
	    if (!c)
		return NULL;

	    p = strchr(head, '\n');
	    if (p && p < c)
		return NULL;

	    len = c - head + 1;
	    if (len > sizeof(result) - 2)
		return NULL;

	    (void)memcpy(result, head, len);
	    result[len] = '\0';

	    return result;
	}

	head = strchr(head, '\n');
	if (head)
	    head++;
	else
	    return NULL;

    }

    /* not reached */
}

static void get_article(char *args, int head, int body)
{
    if (!group_path)
	printf(CODE_TO_STR(ERR_NCING) " Not in a newsgroup.\r\n");
    else if (!args && curr_art < 0)
	printf(CODE_TO_STR(ERR_NOCRNT) " No current article.\r\n");
    else if (args && !only_digits(args)) {
	if (args[0] == '<')
	    printf(CODE_TO_STR(ERR_FAULT)
		   " Message-id lookup not implemented.\r\n");
	else
	    printf(CODE_TO_STR(ERR_CMDSYN) " Syntax error.\r\n");
    } else {
	char	buffer[1024];
	int	fd;

	if (args) {
	    long	no = atol(args);

	    no = art_search(arts, n_arts, no);
	    if (no < 0) {
		printf(CODE_TO_STR(ERR_NOART) " No such article.\r\n");
		return;
	    }
	    curr_art = no;
	}
	    
	sprintf(buffer, "%ld", arts[curr_art]);

	if ((fd = open(buffer, O_RDONLY)) < 0)
	    printf(CODE_TO_STR(ERR_NOART) " No such article.\r\n");
	else {
	    long	len;
	    char	*c;
	    int		bol, in_head;

	    len = readn(fd, buffer, sizeof(buffer) - 1);

	    if (len < 0) {
		printf(CODE_TO_STR(ERR_FAULT)
		       " Error while reading article.\r\n");
		(void)close(fd);
		return;
	    }

	    buffer[len] = '\0';
	    c = extract_messageid(buffer);

	    if (!c) {
		printf(CODE_TO_STR(ERR_FAULT)
		       " Couldn't extract message id.\r\n");
		(void)close(fd);
		return;
	    }

	    if (head && body)
		printf(CODE_TO_STR(OK_ARTICLE) " %ld %s article.\r\n",
		       arts[curr_art], c);
	    else if (head)
		printf(CODE_TO_STR(OK_HEAD) " %ld %s head.\r\n",
		       arts[curr_art], c);
	    else
		printf(CODE_TO_STR(OK_BODY) " %ld %s body.\r\n",
		       arts[curr_art], c);

	    bol = True;
	    in_head = True;
	    do {
		int	i;

		for (i = 0 ; i < len ; i++) {
		    if (in_head && bol && buffer[i] == '\n') {
			in_head = False;
			if (!body)
			    break;
			if (!head)
			    continue; /* no separating empty line */
		    }

		    if (bol && buffer[i] == '.')
			putchar('.');

		    if (in_head ? head : body) {
			if (buffer[i] == '\n')
			    putchar('\r');
			putchar(buffer[i]);
		    }

		    bol = (buffer[i] == '\n');
		}

		if (i < len || /* only wanted head */
		    len < sizeof(buffer) - 1) /* EOF */
		    break;
	    } while ((len = readn(fd, buffer, sizeof(buffer) - 1)) > 0);

	    if (bol)
		printf(".\r\n");
	    else
		printf("\r\n.\r\n");

	    (void)close(fd);
	}
    }
}

/************************************************************************/

static void do_article(char *args)
{
    get_article(args, True, True);
}

static void do_body(char *args)
{
    get_article(args, False, True);
}

static void do_group(char *args)
{
    if (!args)
	printf(CODE_TO_STR(ERR_FAULT) " No newsgroup specified.\r\n");
    else if (strlen(args) > 512)
	printf(CODE_TO_STR(ERR_FAULT) " Group name too long.\r\n");
    else {
	char	path[1024];
	DIR	*dir;
	char	*p;

	strcpy(path, spool_dir);
	p = path + strlen(path);
	*p++ = '/';
	strcpy(p, args);
	while (*p != '\0')
	    if (*p == '.')
		*p++ = '/';
	    else
		p++;

	if (chdir(path) < 0 ||
	    !(dir = opendir(".")))
	    printf(CODE_TO_STR(ERR_NOGROUP) " No such newsgroup.\r\n");
	else {
	    long		n_alloc = n_arts;
	    struct dirent	*dp;

	    group_path = xrealloc(group_path, strlen(args) + 1);
	    strcpy(group_path, args);
	    for (p = strchr(group_path, '.') ; p ; p = strchr(p + 1, '.'))
		*p = '/';

	    n_arts = 0;
	    curr_art = -1;
	    while ( (dp = readdir(dir)) )
		if (only_digits(dp->d_name)) {
		    if (n_arts > n_alloc - 2) {
			n_alloc = 2 * (n_alloc + 1);
			arts = xrealloc(arts, n_alloc * sizeof(long));
		    }

		    arts[n_arts++] = atol(dp->d_name);
		}

	    (void)closedir(dir);

	    if (n_arts == 0)
		curr_art = -1;
	    else {
		curr_art = 0;
		qsort(arts, n_arts, sizeof(long), long_cmp);
	    }

	    printf(CODE_TO_STR(OK_GROUP) " %ld %ld %ld.\r\n",
		   n_arts,
		   n_arts > 0 ? arts[0] : 0l,
		   n_arts > 0 ? arts[n_arts - 1] : 0l);
	}
    }
}

static void do_head(char *args)
{
    get_article(args, True, False);
}

static void do_help(char *args)
{
    printf(CODE_TO_STR(INF_HELP) " Legal commands\r\n"
	   "  article [Number]\r\n"
	   "  body [Number]\r\n"
	   "  group newsgroup\r\n"
	   "  head [Number]\r\n"
	   "  help\r\n"
	   "  list [active|newsgroups]\r\n"
	   "  next\r\n"
	   "  post\r\n"
	   "  quit\r\n"
	   "  xover [range]\r\n"
	   ".\r\n");
}

static void do_list(char *args)
{
    if (!args ||
	strcasecmp(args, "active") == 0) {
	FILE	*fp;

	if (!active)
	    printf(CODE_TO_STR(ERR_FAULT)
		   " Active file not available.\r\n");
	else if (!(fp = fopen(active, "r")))
	    printf(CODE_TO_STR(ERR_FAULT)
		   " Couldn't open active file.\r\n");
	else {
	    int	c, prev = 0;

	    printf(CODE_TO_STR(OK_GROUPS) " Active file follows.\r\n");

	    while ((c = getc(fp)) != EOF) {
		prev = c;
		if (c == '\n')
		    putchar('\r');
		putchar(c);
	    }

	    if (prev == '\n')
		printf(".\r\n");
	    else
		printf("\r\n.\r\n");

	    (void)fclose(fp);
	}
    } else if (strcasecmp(args, "newsgroups") == 0) {
	FILE	*fp;

	if (!newsgroups)
	    printf(CODE_TO_STR(ERR_FAULT)
		   " Newsgroups file not available.\r\n");
	else if (!(fp = fopen(newsgroups, "r")))
	    printf(CODE_TO_STR(ERR_FAULT)
		   " Couldn't open newsgroups file.\r\n");
	else {
	    int	c, prev = 0;

	    printf(CODE_TO_STR(OK_GROUPS) " Descriptions follow.\r\n");

	    while ((c = getc(fp)) != EOF) {
		prev = c;
		if (c == '\n')
		    putchar('\r');
		putchar(c);
	    }

	    if (prev == '\n')
		printf(".\r\n");
	    else
		printf("\r\n.\r\n");

	    (void)fclose(fp);
	}
    } else {
	printf(CODE_TO_STR(ERR_CMDSYN) " Syntax error.\r\n");
    }
}

static void do_next(char *args)
{
    if (args)
	printf(CODE_TO_STR(ERR_CMDSYN) " Syntax error.\r\n");
    else if (!group_path)
	printf(CODE_TO_STR(ERR_NCING) " Not in a newsgroup.\r\n");
    else if (curr_art < 0)
	printf(CODE_TO_STR(ERR_NOCRNT) " No current article.\r\n");
    else if (curr_art >= n_arts - 1)
	printf(CODE_TO_STR(ERR_NONEXT) " No next article.\r\n");
    else /* not conforming to the rfc */
	printf(CODE_TO_STR(OK_NOTEXT) " %ld\r\n", arts[++curr_art]);
}

static void do_not_impl(char *args)
{
    printf(CODE_TO_STR(ERR_FAULT) " Command not implemented.\r\n");
}

static void do_post(char *args)
{
    printf(CODE_TO_STR(ERR_NOPOST) " You can't post to a fake server.\r\n");
}

static void do_quit(char *args)
{
    printf(CODE_TO_STR(OK_GOODBYE) " Goodbye.\r\n");
    exit(0);
}

static void do_xover(char *args)
{
    char	buffer[1024];
    FILE	*fp = NULL;
    long	first, last, no;
    char	*c;

    if (!group_path) {
	printf(CODE_TO_STR(ERR_NCING) " Not in a newsgroup.\r\n");
	return;
    }

    if (!args || args[0] == '\0') {
	if (curr_art >= 0)
	    first = last = arts[curr_art];
	else {
	    printf(CODE_TO_STR(ERR_NOCRNT) " No selected article.\r\n");
	    return;
	}
    } else {
	c = strchr(args, '-');

	if (c) {
	    *c++ = '\0';
	    if (sscanf(args, "%ld", &first) != 1 ||
		sscanf(c, "%ld", &last) != 1) {
		printf(CODE_TO_STR(ERR_CMDSYN) " Syntax error.\r\n");
		return;
	    }
	} else {
	    if (sscanf(args, "%ld", &first) != 1) {
		printf(CODE_TO_STR(ERR_CMDSYN) " Syntax error.\r\n");
		return;
	    }
	    last = first;
	}
    }

    if (overview_dir)
	sprintf(buffer, "%s/%s/.overview",  overview_dir, group_path);
    else
	sprintf(buffer, ".overview");

    fp = fopen(buffer, "r");
    if (!fp)
	perror(buffer);

    printf(CODE_TO_STR(OK_XOVER) " Overview file follows.\r\n");

    no = 0;
    if (fp) {
	while (fgets(buffer, sizeof(buffer), fp)) {
	    if (buffer[0] < '0' || buffer[0] > '9' ||
		(no = atol(buffer)) < first) {
		if (!strchr(buffer, '\n')) {
		    int	ch;

		    do {
			ch = getc(fp);
		    } while (ch != EOF && ch != '\n');

		    if (ch == EOF)
			break;
		}
		continue;
	    }
	    if (no > last)
		break;

	    fputs(buffer, stdout);
	    if (!strchr(buffer, '\n')) {
		int	ch;

		while ((ch = getc(fp)) != EOF && ch != '\n')
		    putchar(ch);

		printf("\r\n");
	    }
	}

	(void)fclose(fp);
	fp = NULL;
    }

    if (no < last) {
	long	n = 0;

	if (update_overview) {
	    if (overview_dir)
		sprintf(buffer, "%s/%s/.overview", overview_dir, group_path);
	    else
		sprintf(buffer, ".overview");

	    fp = fopen(buffer, "a");
	    if (fp)
		updating_overview = True;
	    else
		perror(".overview");
	}

	while (n < n_arts && arts[n] <= no)
	    n++;

	while (n < n_arts && arts[n] <= last) {
	    int	fd;

	    sprintf(buffer, "%ld", arts[n]);
	    fd = open(buffer, O_RDONLY);
	    if (fd < 0)
		perror("open");
	    else {
		make_over_rec(fd, arts[n], fp);
		(void)close(fd);
	    }

	    n++;
	}

	updating_overview = False;
	if (fp)
	    (void)fclose(fp);
    }

    printf(".\r\n");
}

/*************************************************************************/

static char *get_stdin_line()
{
    static char	buffer[512] = "";
    static int	curr_pos = 0;
    char	*temp, *c;

    temp = strchr(buffer + curr_pos, '\n');

    if (!temp) {
	int	n;

	fflush(stdout);
	n = strlen(buffer + curr_pos);
	if (curr_pos > 0)
	    (void)memmove(buffer, buffer + curr_pos, n + 1);
	curr_pos = 0;
	c = buffer + n;
	n = sizeof(buffer) - n - 2;

	for (;;) {
	    int	i;

	    if (n <= 0) {
		printf(CODE_TO_STR(ERR_FAULT) " Too long line, exiting.\r\n");
		exit(0);
	    }

	    i = read(STDIN_FILENO, c, n);

	    if (i <= 0) {
		if (i < 0)
		    perror("read");
		exit(0);
	    }

	    c[i] = '\0';
	    if ( (temp = strchr(c, '\n')) )
		break;
	    c += i;
	    n -= i;
	}
    }

    if (temp > buffer && temp[-1] == '\r')
	temp[-1] = '\0';
    *temp++ = '\0';
    c = buffer + curr_pos;
    curr_pos = temp - buffer;

    return c;
}

int main(int argc, char **argv)
{
#define N_COMMANDS	(sizeof(commands)/sizeof(commands[0]))
    static struct {
	char	*name;
	void	(*proc)(char*);
    } commands[] = {
	{"article",	do_article},
	{"authinfo",	do_not_impl},
	{"body",	do_body},
	{"date",	do_not_impl},
	{"group",	do_group},
	{"head",	do_head},
	{"help",	do_help},
	{"ihave",	do_not_impl},
	{"list",	do_list},
	{"listgroup",	do_not_impl},
	{"next",	do_next},
	{"newgroups",	do_not_impl},
	{"newnews",	do_not_impl},
	{"post",	do_post},
	{"quit",	do_quit},
	{"sendme",	do_not_impl},
	{"slave",	do_not_impl},
	{"stat",	do_not_impl},
	{"xgtitle",	do_not_impl},
	{"xhdr",	do_not_impl},
	{"xover",	do_xover},
	{"xpat",	do_not_impl},
	{"xpath",	do_not_impl},
    };
#ifdef _POSIX_SOURCE
    struct sigaction	sig_act;

    sig_act.sa_handler = sigpipe_handler;
    sigemptyset(&sig_act.sa_mask);
    sig_act.sa_flags = 0;
    if (sigaction(SIGPIPE, &sig_act, NULL) < 0)
	perror("sigaction");
#else
    if (signal(SIGPIPE, sigpipe_handler) == SIG_ERR)
	perror("signal");
#endif

    while (--argc > 0 &&
	   **++argv == '-') {
	int	n = strlen(*argv) - 1;

#define IS_OPTION(o, has_arg) \
	(strncmp(*argv, o, n) == 0 && \
	 n < sizeof(o) && (!has_arg || argc > 0))

	switch (argv[0][1]) {
	case 's':
	    if (IS_OPTION("-spool", True)) {
		spool_dir = *++argv;
		argc--;
		continue;
	    }
	    break;
	case 'a':
	    if (IS_OPTION("-active", True)) {
		active = *++argv;
		argc--;
		continue;
	    }
	    break;
	case 'n':
	    if (IS_OPTION("-newsgroups", True)) {
		newsgroups = *++argv;
		argc--;
		continue;
	    }
	    break;
	case 'o':
	    if (IS_OPTION("-overview", True)) {
		overview_dir = *++argv;
		argc--;
		continue;
	    }
	    break;
	case 'u':
	    if (IS_OPTION("-update", False)) {
		update_overview = True;
		continue;
	    }
	    break;
	default:
	    break;
	}
#undef IS_OPTION
	break;
    }

    if (argc != 0) {
	printf(CODE_TO_STR(ERR_FAULT) " Bad command line arguments.\r\n");
	exit(0);
    } else if (strlen(spool_dir) > 480) {
	printf(CODE_TO_STR(ERR_FAULT) " Spool dir name to long.\r\n");
	exit(0);
    } else if (overview_dir && strlen(overview_dir) > 480) {
	printf(CODE_TO_STR(ERR_FAULT) " Overview dir name to long.\r\n");
	exit(0);
    } else if (chdir(spool_dir) < 0) {
	perror(spool_dir);
	printf(CODE_TO_STR(ERR_FAULT) " Couldn't chdir to spool dir.\r\n");
	exit(0);
    }

    printf(CODE_TO_STR(OK_NOPOST) " knewsd, no posting.\r\n");

    for (;;) {
	char	*command;
	char	*c, *p;
	int	i;

	command = get_stdin_line();
	c = strchr(command, ' ');
	if (c)
	    *c++ = '\0';
	for (p = command ; *p != '\0' ; p++)
	    if (*p >= 'A' && *p <= 'Z')
		*p -= 'A' - 'a';

	for (i = 0 ; i < N_COMMANDS ; i++)
	    if (command[0] == commands[i].name[0] &&
		strcmp(command, commands[i].name) == 0) {
		commands[i].proc(c);
		break;
	    }

	if (i == N_COMMANDS)
	    printf(CODE_TO_STR(ERR_CMDSYN) " Command not recognized.\r\n");
    }

    /* not reached */
}
