/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include "overview.h"

static char *find_lflf(char *c)
{
    for (c = strchr(c, '\n') ; c ; c = strchr(c, '\n'))
	if (*++c == '\n')
	    return c;

    return NULL;
}

static char *get_header(int fd)
{
    static char	buffer[16384 + 3];
    char	*c, *p;
    long	n;

    n = sizeof(buffer);
    c = buffer;

    do {
	long	i = read(fd, c, 1024);

	if (i <= 0) {
	    if (i < 0)
		perror("read");
	    return NULL;
	}

	c[i] = '\0';
	c += i;
	n -= i;

	if ( (p = find_lflf(buffer)) ) {
	    *p = '\0';
	    return buffer;
	}
    } while (n > 0);

    return NULL;
}

static char *hack_inreplyto(char *inreplyto)
{
    char	*c = strchr(inreplyto, '<');
    char	*p = c;

    if (!c)
	return NULL;

    for (;;)
	switch (*p++) {
	case '>':
	    *p = '\0';
	    return c;
	case '\0':
	case '\t':
	case ' ':
	    return NULL;
	}

    /* not reached */
}

static char *stat_lines(int fd)
{
    static char	result[32];
    struct stat	st;

    if (fstat(fd, &st) < 0)
	return "";

    sprintf(result, "%ld", (long)st.st_size);

    return result;
}

int make_over_rec(int fd, long no, FILE *fp)
{
    register char	*header;
    char	*subject = "";
    char	*from = "";
    char	*date = "";
    char	*messageid = NULL;
    char	*refs = NULL;
    char	*bytes = NULL;
    char	*lines = "";
    char	*xref = NULL;
    char	*inreplyto = NULL;

    header = get_header(fd);
    if (!header)
	return -1;

    while (*header != '\0') {
	int	skip = True;

	switch (*header) {
	case 'b':
	case 'B':
	    if (strncasecmp(header, "Bytes: ", 7) == 0) {
		bytes = header + 7;
		skip = False;
	    }
	    break;
	case 'd':
	case 'D':
	    if (strncasecmp(header, "Date: ", 6) == 0) {
		date = header + 6;
		skip = False;
	    }
	    break;
	case 'f':
	case 'F':
	    if (strncasecmp(header, "From: ", 6) == 0) {
		from = header + 6;
		skip = False;
	    }
	    break;
	case 'i':
	case 'I':
	    if (strncasecmp(header, "In-Reply-To: ", 13) == 0) {
		inreplyto = header + 13;
		skip = False;
	    }
	    break;
	case 'l':
	case 'L':
	    if (strncasecmp(header, "Lines: ", 7) == 0) {
		lines = header + 7;
		skip = False;
	    }
	    break;
	case 'm':
	case 'M':
	    if (strncasecmp(header, "Message-Id: ", 12) == 0) {
		messageid = header + 12;
		skip =  False;
	    }
	    break;
	case 'r':
	case 'R':
	    if (strncasecmp(header, "References: ", 12) == 0) {
		refs = header + 12;
		skip = False;
	    }
	    break;
	case 's':
	case 'S':
	    if (strncasecmp(header, "Subject: ", 9) == 0) {
		subject = header + 9;
		skip = False;
	    }
	    break;
	case 'x':
	case 'X':
	    if (strncasecmp(header, "Xref: ", 6) == 0) {
		xref = header;
		skip = False;
	    }
	    break;
	}

	if (skip) {
	    header = strchr(header, '\n');
	    while (header && IS_SPACE(header[1]))
		header = strchr(header + 1, '\n');
	    if (!header)
		break;
	    header++;
	} else {
	    char	*src, *dest;

	    src = strchr(header, '\n');
	    if (!src)
		break;
	    dest = src;

	    for (;;) {
		if (*src == '\0')
		    break;
		else if (*src == '\n') {
		    src++;
		    if (!IS_SPACE(*src))
			break;
		    *dest++ = ' ';
		} else if (*src == '\t') {
		    src++;
		    *dest++ = ' ';
		} else {
		    *dest++ = *src++;
		}
	    }

	    *dest = '\0';
	    header = src;
	}
    }

    if (!messageid) {
	fprintf(stderr, "%ld: no message-id, skipping.\n", no);
	return -1;
    }
    if (!refs) {
	if (inreplyto)
	    refs = hack_inreplyto(inreplyto);
	else
	    refs = "";
    }
    if (!bytes)
	bytes = stat_lines(fd);

    printf("%ld\t%s\t%s\t%s\t%s\t%s\t%s\t%s",
	   no, subject, from, date, messageid, refs, bytes, lines);
    if (xref)
	printf("\t%s", xref);
    putchar('\n');

    if (fp) {
	fprintf(fp, "%ld\t%s\t%s\t%s\t%s\t%s\t%s\t%s",
		no, subject, from, date, messageid, refs, bytes, lines);
	if (xref)
	    fprintf(fp, "\t%s", xref);
	putc('\n', fp);
    }

    return 0;
}
